Package: GeneSelectMMD
Type: Package
Title: Gene selection based on the marginal distributions of gene
        profiles that characterized by a mixture of three-component
        multivariate distributions
Version: 2.56.0
Date: 2020-04-03
Author: Jarrett Morrow <remdj@channing.harvard.edu>, Weiliang Qiu
        <weiliang.qiu@gmail.com>, Wenqing He <whe@stats.uwo.ca>,
        Xiaogang Wang <stevenw@mathstat.yorku.ca>, Ross Lazarus
        <ross.lazarus@channing.harvard.edu>.
Maintainer: Weiliang Qiu <weiliang.qiu@gmail.com>
Depends: R (>= 2.13.2), Biobase
Imports: MASS, graphics, stats, limma
Suggests: ALL
Description: Gene selection based on a mixture of marginal
        distributions.
License: GPL (>= 2)
Packaged: 2026-04-29 21:26:28 UTC; root
biocViews: DifferentialExpression
Repository: https://bioc-release.r-universe.dev
Date/Publication: 2026-04-28 12:33:47 UTC
RemoteUrl: https://github.com/bioc/GeneSelectMMD
RemoteRef: RELEASE_3_23
RemoteSha: 6a28f02754efedd96ed960cef22ec6c2b404a8f5
NeedsCompilation: yes
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-29 22:22:33 UTC; unix

export(
  gsMMD,
  gsMMD.default,
  gsMMD2,
  gsMMD2.default,
  errRates,
  plotHistDensity,
  obtainResi
)

useDynLib(GeneSelectMMD, .registration = TRUE)

import(Biobase)

importFrom(MASS,
          ginv)

importFrom(graphics,
          hist,
          legend,
          lines,
          plot,
          title)

importFrom(stats,
          cor,
          dnorm,
          qqnorm,
          sd,
          t.test,
          var,
          wilcox.test,
          model.matrix)

importFrom("stats", "optimize")

importFrom(limma,
          lmFit,
          eBayes)

### R code from vignette source 'gsMMD.Rnw'

###################################################
### code chunk number 1: gsMMD.Rnw:494-510
###################################################
  library(GeneSelectMMD)
  library(ALL)
  data(ALL)
  eSet1 <- ALL[1:100, ALL$BT == "B3" | ALL$BT == "T2"]
  
  mem.str <- as.character(eSet1$BT)
  nSubjects <- length(mem.str)
  memSubjects <- rep(0,nSubjects)
  # B3 coded as 0, T2 coded as 1
  memSubjects[mem.str == "T2"] <- 1
  
  obj.gsMMD <- gsMMD(eSet1, memSubjects, transformFlag = TRUE, 
    transformMethod = "boxcox", scaleFlag = TRUE, quiet = TRUE)
  para <- obj.gsMMD$para
  print(round(para, 3))



###################################################
### code chunk number 2: gsMMD.Rnw:513-531 (eval = FALSE)
###################################################
##   library(GeneSelectMMD)
##   library(ALL)
##   data(ALL)
##   eSet1 <- ALL[1:100, ALL$BT == "B3" | ALL$BT == "T2"]
##   mat <- exprs(eSet1)
##   
##   mem.str <- as.character(eSet1$BT)
##   nSubjects <- length(mem.str)
##   memSubjects <- rep(0,nSubjects)
##   # B3 coded as 0, T2 coded as 1
##   memSubjects[mem.str == "T2"] <- 1
##  
##   obj.gsMMD <- gsMMD.default(mat, memSubjects, iniGeneMethod = "Ttest",
##           transformFlag = TRUE, transformMethod = "boxcox", scaleFlag = TRUE,
##           quiet=TRUE)
##   para <- obj.gsMMD$para
##   print(round(para, 3))
## 


###################################################
### code chunk number 3: gsMMD.Rnw:537-576 (eval = FALSE)
###################################################
##   library(GeneSelectMMD)
##   library(ALL)
##   data(ALL)
##   eSet1 <- ALL[1:100, ALL$BT == "B3" | ALL$BT == "T2"]
##   
##   mem.str <- as.character(eSet1$BT)
##   nSubjects <- length(mem.str)
##   memSubjects <- rep(0,nSubjects)
##   # B3 coded as 0, T2 coded as 1
##   memSubjects[mem.str == "T2"] <- 1
##   
##   myWilcox <-
##   function(x, memSubjects, alpha = 0.05)
##   {
##     xc <- x[memSubjects == 1]
##     xn <- x[memSubjects == 0]
##   
##     m <- sum(memSubjects == 1)
##     res <- wilcox.test(x = xc, y = xn, conf.level = 1 - alpha)
##     res2 <- c(res$p.value, res$statistic - m * (m + 1) / 2)
##     names(res2) <- c("p.value", "statistic")
##   
##     return(res2)
##   }
##   
##   mat <- exprs(eSet1)
##   tmp <- t(apply(mat, 1, myWilcox, memSubjects = memSubjects))
##   colnames(tmp) <- c("p.value", "statistic")
##   memIni <- rep(2, nrow(mat))
##   memIni[tmp[, 1] < 0.05 & tmp[, 2] > 0] <- 1
##   memIni[tmp[, 1] < 0.05 & tmp[, 2] < 0] <- 3
##   
##   print(table(memIni))
## 
##   obj.gsMMD <- gsMMD2(eSet1, memSubjects, memIni, transformFlag = TRUE, 
##        transformMethod = "boxcox", scaleFlag = TRUE, quiet = TRUE)
##   para <- obj.gsMMD$para
##   print(round(para, 3))
## 


###################################################
### code chunk number 4: gsMMD.Rnw:579-620 (eval = FALSE)
###################################################
##   library(GeneSelectMMD)
##   library(ALL)
##   data(ALL)
##   eSet1 <- ALL[1:100, ALL$BT == "B3" | ALL$BT == "T2"]
##   mat <- exprs(eSet1)
##   
##   mem.str <- as.character(eSet1$BT)
##   nSubjects <- length(mem.str)
##   memSubjects <- rep(0,nSubjects)
##   # B3 coded as 0, T2 coded as 1
##   memSubjects[mem.str == "T2"] <- 1
##  
##   myWilcox <-
##   function(x, memSubjects, alpha = 0.05)
##   {
##     xc <- x[memSubjects == 1]
##     xn <- x[memSubjects == 0]
##   
##     m <- sum(memSubjects == 1)
##     res <- wilcox.test(x = xc, y = xn, conf.level = 1 - alpha)
##     res2 <- c(res$p.value, res$statistic - m * (m + 1) / 2)
##     names(res2) <- c("p.value", "statistic")
##   
##     return(res2)
##   }
##   
##   tmp <- t(apply(mat, 1, myWilcox, memSubjects = memSubjects))
##   colnames(tmp) <- c("p.value", "statistic")
##   memIni <- rep(2, nrow(mat))
##   memIni[tmp[, 1] < 0.05 & tmp[, 2] > 0] <- 1
##   memIni[tmp[, 1] < 0.05 & tmp[, 2] < 0] <- 3
##   
##   print(table(memIni))
##   
##   obj.gsMMD <- gsMMD2.default(mat, memSubjects, memIni = memIni,
##           transformFlag = TRUE, transformMethod = "boxcox", scaleFlag = TRUE,
##           quiet=TRUE)
##   para <- obj.gsMMD$para
##   print(round(para, 3))
## 
## 


###################################################
### code chunk number 5: gsMMD.Rnw:659-661
###################################################
  print(round(errRates(obj.gsMMD), 3))



###################################################
### code chunk number 6: gsMMD.Rnw:677-686
###################################################
  plotHistDensity(obj.gsMMD, plotFlag = "case", 
      mytitle = "Histogram of gene expression levels for T2\nimposed with estimated density (case)", 
      plotComponent = TRUE, 
      x.legend = c(0.8, 3), 
      y.legend = c(0.3, 0.4), 
      numPoints = 500,
      cex.main = 1,
      cex = 1)

# modified on Sept. 28, 2011
#  (1) added 'na.rm=TRUE' to function 'sum'
#
# v0.1.2
#  (1) obj.gsMMD is changed to contain 'memSubjects'
#  (2) fixed a bug relating to the 'ylim' of the plot.
#       new ylim is obtained from the ranges of both the estimated density
#       obtained from the model and the density obtained by 'hist' function
plotHistDensity<-function(obj.gsMMD,
                          plotFlag="case",
                          plotComponent=FALSE,
                          myxlab="expression level",
                          myylab="density",
                          mytitle="Histogram with estimated density (case)",
                          x.legend=NULL, y.legend=NULL,
                          numPoints=500,
                          mycol=1:4, mylty=1:4, mylwd=rep(3,4),
                          cex.main=2, cex.lab=1.5, cex.axis=1.5, cex=2,
                          bty="n")
{
  plotFlag<-match.arg(plotFlag, choices=c("case", "control"))

  dat<-obj.gsMMD$dat
  para<-as.numeric(obj.gsMMD$para)
  names(para)<-paraNames
  memSubjects<-obj.gsMMD$memSubjects

  nCases<-sum(memSubjects==1, na.rm=TRUE)
  nControls<-sum(memSubjects==0, na.rm=TRUE)
  n<-nCases+nControls

  pi.1<-para[1]
  pi.2<-para[2]
  pi.3<-para[3]
  mu.c1<-para[4]
  sigma2.c1<-para[5]
  rho.c1<-para[6]

  mu.n1<-para[7]
  sigma2.n1<-para[8]
  rho.n1<-para[9]

  mu.2<-para[10]
  sigma2.2<-para[11]
  rho.2<-para[12]

  mu.c3<-para[13]
  sigma2.c3<-para[14]
  rho.c3<-para[15]

  mu.n3<-para[16]
  sigma2.n3<-para[17]
  rho.n3<-para[18]

  if(plotFlag=="case")
  { x<-as.vector(dat[,memSubjects==1, drop=FALSE])
  }
  else {
    x<-as.vector(dat[,memSubjects==0, drop=FALSE])
  }
  x<-sort(x)
  len.x<-length(x)
  delta<-floor(len.x/numPoints)
  x2<-x[seq(from=1,to=len.x, by=delta)]

  if(plotFlag=="case")
  { y1<-pi.1*dnorm(x2, mean=mu.c1, sd=sqrt(sigma2.c1))
    y2<-pi.2*dnorm(x2, mean=mu.2, sd=sqrt(sigma2.2))
    y3<-pi.3*dnorm(x2, mean=mu.c3, sd=sqrt(sigma2.c3))
    y<-y1+y2+y3

  } else {
    y1<-pi.1*dnorm(x2, mean=mu.n1, sd=sqrt(sigma2.n1))
    y2<-pi.2*dnorm(x2, mean=mu.2, sd=sqrt(sigma2.2))
    y3<-pi.3*dnorm(x2, mean=mu.n3, sd=sqrt(sigma2.n3))
    y<-y1+y2+y3
  }

  tmp<-hist(x, plot=FALSE)
  myylim<-range(c(y, tmp$density), na.rm=TRUE)
  hist(x, freq=FALSE, main=mytitle,xlab=myxlab, ylab=myylab, ylim=myylim,
    cex.main=cex.main, cex.lab=cex.lab)
  lines(x2, y, col=mycol[1], lty=mylty[1], lwd=mylwd[1])

  if(plotComponent)
  {
    lines(x2, y1, col=mycol[2], lty=mylty[2], lwd=mylwd[2])
    lines(x2, y2, col=mycol[3], lty=mylty[3], lwd=mylwd[2])
    lines(x2, y3, col=mycol[4], lty=mylty[4], lwd=mylwd[2])
  }
  if(is.null(x.legend))
  {
    x.max<-max(x, na.rm=TRUE)
    d<-max(x, na.rm=TRUE)-min(x, na.rm=TRUE)
    tmp1<-x.max-d/3
    x.legend<-c(tmp1, x.max)
  }
  if(is.null(y.legend))
  {
    y.max<-max(y, na.rm=TRUE)
    d<-max(y, na.rm=TRUE)-min(y, na.rm=TRUE)
    tmp1<-y.max-d/4
    y.legend<-c(tmp1, y.max)

  }

  if(plotComponent)
  { legend(x=x.legend, y=y.legend, legend=c("overall","component1", "component2"
, "component3"),
    lty=mylty, col=mycol, lwd=mylwd, cex=cex, bty=bty)
  }
  invisible(list(x=x,x2=x2, y=y, y1=y1, y2=y2, y3=y3))
}

# modified on Sept. 28, 2011
#  (1) added 'na.rm=TRUE' to function 'sum'
#
# define some constant
PI<-3.1415926
TNumPara<-18
paraNames<- c("pi.1", "pi.2", "pi.3",
              "mu.c1", "sigma2.c1", "rho.c1", 
              "mu.n1", "sigma2.n1", "rho.n1",
              "mu.2", "sigma2.2", "rho.2",
              "mu.c3", "sigma2.c3", "rho.c3", 
              "mu.n3", "sigma2.n3", "rho.n3")

# re-parametrization
TNumParaRP<-17
paraNamesRP<- c("pi.1", "pi.2", 
              "mu.c1", "tau.c1", "r.c1", 
              "delta.n1", "tau.n1", "r.n1",
              "mu.2", "tau.2", "r.2",
              "mu.c3", "tau.c3", "r.c3", 
              "delta.n3", "tau.n3", "r.n3")


"gsMMD"<-
function(obj.eSet, 
         memSubjects, 
         maxFlag=TRUE, 
         thrshPostProb=0.50, 
         geneNames=NULL, 
         alpha=0.05, 
         iniGeneMethod= "Ttest", 
         transformFlag=FALSE, 
         transformMethod="boxcox", 
         scaleFlag=TRUE, 
         criterion=c("cor", "skewness", "kurtosis"),
         minL=-10, 
         maxL=10, 
         stepL=0.1, 
         eps=1.0e-3, 
         ITMAX=100, 
         plotFlag=FALSE,
         quiet=TRUE)
{
  # get expression level matrix
  X<-exprs(obj.eSet)

  res<-gsMMD.default(X, 
         memSubjects, 
         maxFlag, 
         thrshPostProb, 
         geneNames, 
         alpha, 
         iniGeneMethod,
         transformFlag,
         transformMethod,
         scaleFlag,
         criterion,
         minL,
         maxL,
         stepL,
         eps,
         ITMAX,
         plotFlag,
         quiet)

  invisible(res)

}


"gsMMD.default"<-
function(X, 
         memSubjects, 
         maxFlag=TRUE, 
         thrshPostProb=0.50, 
         geneNames=NULL, 
         alpha=0.05, 
         iniGeneMethod= "Ttest", 
         transformFlag=FALSE, 
         transformMethod="boxcox", 
         scaleFlag=TRUE, 
         criterion=c("cor", "skewness", "kurtosis"),
         minL=-10, 
         maxL=10, 
         stepL=0.1, 
         eps=1.0e-3, 
         ITMAX=100, 
         plotFlag=FALSE,
         quiet=TRUE)
{
  transformMethod<-match.arg(transformMethod, choices=c("boxcox", "log2", "log10", "log", "none"))
  criterion<-match.arg(criterion, c("cor", "skewness", "kurtosis"))

  posMethod<-match(iniGeneMethod, c("Ttest","Wilcox"))
  tmppos<-which(is.na(posMethod)==TRUE)
  if(length(tmppos)>0)
  { msg<-paste("The initial gene partition method(s):", iniGeneMethod[tmppos], " not available!\n")
   stop(msg)
  }

  X<-as.matrix(X)
  nGenes<-nrow(X)
  nSubjects<-ncol(X)
  nMethods<-length(iniGeneMethod)
  nCases<-sum(memSubjects==1, na.rm=TRUE)
  nControls<-sum(memSubjects==0, na.rm=TRUE)

  if(sum(is.null(geneNames), na.rm=TRUE))
  { geneNames<-paste("gene", 1:nGenes, sep="") }

  cat("Programming is running. Please be patient...\n")
  lambda<-NA
  if(transformFlag)
  { 
    if(transformMethod!="none")
    {
      vec<-as.numeric(X)
      min.vec<-min(vec, na.rm=TRUE)
      if(min.vec<0)
      {
        cat("****** Begin Warning ******** \n")
        cat("Warning: Data contains non-positive values! To continue ",
          transformMethod, " transformation,\n")
        cat("We first perform the following transformation:\n")
        cat("x<-x+abs(min(x, na.rm=TRUE))+1\n")
        cat("****** End Warning ******** \n")
    
        X<-X+abs(min.vec)+1
      }
    }
    tmp<-transFunc(X, transformMethod, criterion, 
                   minL, maxL, stepL, eps, plotFlag, ITMAX=0) 
    if(transformMethod=="boxcox")
    { X<-tmp$dat 
      lambda<-tmp$lambda.avg
    }
    else {
     X<-tmp
    }
    if(!quiet)
    { cat(paste("Data transformation (", transformMethod, ") performed\n")) }
  }

  if(scaleFlag)
  {
    if(!quiet)
    { cat("Gene profiles are scaled so that they have mean zero and variance one!\n") }
    X<-t(apply(X, 1, scale, center=TRUE, scale=TRUE))

    # to avoid linear dependence of tissue samples after scaling
    # gene profiles, we delete a tissue sample.
    # We arbitrarily select the tissue sample, which has the biggest label number, 
    # from the tissue sample group that has larger size than the other 
    # tissue sample group. For example, if there are 6 cancer tissue samples 
    # and 10 normal tissue samples, we delete the 10-th normal tissue sample after scaling.

    if(nCases>nControls)
    { 
      pos<-which(memSubjects==1)
      pos2<-pos[nCases]
      X<-X[,-pos2]
      memSubjects<-memSubjects[-pos2]
    } else {
      pos<-which(memSubjects==0)
      pos2<-pos[nControls]
      X<-X[,-pos2]
      memSubjects<-memSubjects[-pos2]
    }
    nCases<-sum(memSubjects==1, na.rm=TRUE)
    nControls<-sum(memSubjects==0, na.rm=TRUE)
    nSubjects<-nCases+nControls
  }

  # records initial parameter estimates
  paraIniMatRP<-matrix(0, nrow=TNumParaRP, ncol=nMethods)
  # records initial gene-membership estimates
  memIniMat<-matrix(0, nrow=nGenes, ncol=nMethods)

  # records initial log-likelihood estimates
  llkhIniVec<-rep(0, nMethods)

  # records parameter estimates
  paraMatRP<-matrix(0, nrow=TNumParaRP, ncol=nMethods)
  # records gene-membership estimates
  memMat<-matrix(0, nrow=nGenes, ncol=nMethods)

  # records log-likelihood estimates
  llkhVec<-rep(0, nMethods)

  cat("Programming is running. Please be patient...\n")
  # records E(z_{ij} | x_i, Psi^{(m)})
  wiArray<-array(0, c(nGenes, 3, nMethods))
  for(i in 1:nMethods)
  {
    if(!quiet)
    { cat("******** initial parameter estimates method>>", 
        iniGeneMethod[i], " *******\n") }
    tmpIni<-getIniMemGenes(X, memSubjects, geneNames, 
                              iniGeneMethod[i], alpha, eps=eps)
    iniGeneMethod[i]<-tmpIni$iniGeneMethod
    memIniMat[,i]<-tmpIni$memGenes
    paraIniMatRP[,i]<-tmpIni$para
    llkhIniVec[i]<-tmpIni$llkh
    ttt<-paraIniMatRP[,i]
    names(ttt)<-paraNamesRP
    if(!quiet)
    { cat("paraIniMatRP[,i]>>\n"); print(round(ttt,3)); cat("\n"); }

    # Gene Selection based on EM algorithm
    res<- paraEst(X, tmpIni$para, memSubjects=memSubjects, 
                   maxFlag=maxFlag, thrshPostProb, geneNames=geneNames,
                   ITMAX=ITMAX, eps=eps, quiet=quiet)

    if(res$loop==0)
    {
      paraMatRP[,i]<-paraIniMatRP[,i]
      llkhVec[i]<-llkhIniVec[i]
      memMat[,i]<-memIniMat[,i]
      wiArray[,,i]<-res$wiMat
    } else {
      paraMatRP[,i]<-res$para
      llkhVec[i]<-res$llkh
      memMat[,i]<-res$memGenes
      wiArray[,,i]<-res$wiMat
    }
  }
  rownames(paraIniMatRP)<-paraNamesRP
  colnames(paraIniMatRP)<-iniGeneMethod

  rownames(memIniMat)<-geneNames
  colnames(memIniMat)<-iniGeneMethod
  names(llkhIniVec)<-iniGeneMethod
  rownames(paraMatRP)<-paraNamesRP

  colnames(paraMatRP)<-iniGeneMethod
  rownames(memMat)<-geneNames
  colnames(memMat)<-iniGeneMethod
  names(llkhVec)<-iniGeneMethod
  dimnames(wiArray)<-list(geneNames,
                          paste("cluster", 1:3, sep=""),
                          iniGeneMethod)                          


  # final results
  flagPi<-rep(0, nMethods)
  paraIniMat<-matrix(0, nrow=TNumPara, ncol=nMethods)
  rownames(paraIniMat)<-paraNames
  colnames(paraIniMat)<-iniGeneMethod

  paraMat<-matrix(0, nrow=TNumPara, ncol=nMethods)
  rownames(paraMat)<-paraNames
  colnames(paraMat)<-iniGeneMethod

  for(i in 1:nMethods)
  { 
    paraIniMat[,i]<-paraRPConverter(paraIniMatRP[,i], nCases, nControls)
    paraMat[,i]<-paraRPConverter(paraMatRP[,i], nCases, nControls)
    flagPi[i]<-sum(paraMat[2,i]>paraMat[1,i] & paraMat[2,i]>paraMat[3,i], na.rm=TRUE)
  }
  tmppos<-which(flagPi==0)
  if(length(tmppos))
  { llkhVec[tmppos]<- -Inf }
  if(!quiet)
  { cat("llkhVec>>\n"); print(llkhVec); cat("\n"); }
  pos<-which(llkhVec==max(llkhVec, na.rm=TRUE))
  len<-length(pos)
  tt<-sample(1:len, 1, replace=FALSE)
  pos<-pos[tt]

  memGenes<-as.vector(memMat[,pos])
  para<-paraMat[,pos]
  paraRP<-paraMatRP[,pos]
  llkh<-llkhVec[pos]
  wiMat<-wiArray[,,pos]

  memGenes2<-rep(1, nGenes)
  memGenes2[memGenes==2]<-0 # non-differentially expressed genes

  if(!quiet)
  { cat("*******************************************************\n\n") 
    cat("Initial parameter estimates>>\n"); print(round(paraIniMat,3)); cat("\n");
    cat("Initial loglikelihood>>\n"); print(round(llkhIniVec,3)); cat("\n");
    #tmpMat<-matrix(0, nrow=TNumPara, ncol=nMethods)
    #rownames(tmpMat)<-rownames(paraMat)
    #colnames(tmpMat)<-colnames(paraMat)
    #for(i in 1:nMethods)
    #{
    #  #tmpMat[,i]<-paraMat[[i]][,1]
    #  tmpMat[,i]<-paraMat[,1]
    #}
    #cat("Final parameter estimates based on initial estimates>>\n"); print(round(tmpMat,3)); cat("\n");
    cat("Final loglikelihood based on initial estimates>>\n"); print(round(llkhVec,3)); cat("\n");
    cat("Final parameter estimates>>\n"); print(round(para,3)); cat("\n");
    cat("Final loglikelihood>>\n"); print(round(llkh,3)); cat("\n");
    cat("*******************************************************\n\n")
  }

  res<-list(dat=X, memSubjects=memSubjects, 
            memGenes=memGenes, memGenes2=memGenes2, 
            para=para, 
            llkh=llkh, wiMat=wiMat, wiArray=wiArray,
            memIniMat=memIniMat, paraIniMat=paraIniMat, llkhIniVec=llkhIniVec,
            memMat=memMat, paraMat=paraMat, llkhVec=llkhVec, lambda=lambda)
  invisible(res) 
}

getIniMemGenes<-function(X, memSubjects, geneNames, iniGeneMethod="Ttest",
                         alpha=0.05, eps=1e-6)
{
  iniGeneMethod<-match.arg(iniGeneMethod,  choices=c("Ttest","Wilcox"))

  if(iniGeneMethod=="Ttest")
  {
    # (1) two-sample t-test
#    tmp<-iniMemGenesTestFunc(X, memSubjects=memSubjects, testFun=myTtest, 
#   string "myTtest" is passed to enable the Fortran myTtest routine
    tmp<-iniMemGenesTestFunc(X, memSubjects=memSubjects, testFun="myTtest", 
                        geneNames=geneNames, alpha = alpha, eps=eps)
  } else { #iniGeneMethod=="Wilcox"
    # (2) two-sample wilcoxon test
    tmp<-iniMemGenesTestFunc(X, memSubjects=memSubjects, testFun=myWilcox, 
                        geneNames=geneNames, alpha = alpha, eps=eps)
  }
   
  res<-list(para=tmp$para, llkh=tmp$llkh, memGenes=tmp$memGenes, 
            memGenes2=tmp$memGenes2, iniGeneMethod=iniGeneMethod)
  return(res)
}

paraRPConverter<-function(paraRP, nCases, nControls)
{
  nc<-nCases
  nn<-nControls
  n<-nc+nn

  # mixture proportions
  pi.1<-paraRP[1]; pi.2<-paraRP[2]; 
  pi.3<-1-pi.1-pi.2

  # mean expression level for cluster 1 for diseased subjects
  mu.c1<-paraRP[3]; 
  # variance of expression levels for cluster 1 for diseased subjects
  tau.c1<-paraRP[4]; 
  sigma2.c1<-exp(tau.c1)
  # modified logit of correlation among expression levels for cluster 1 for diseased subjects
  r.c1<-paraRP[5]; 
  rho.c1<-(exp(r.c1)-1/(nc-1))/(1+exp(r.c1))

  # mean expression level for cluster 1 for normal subjects
  delta.n1<-paraRP[6]; 
  mu.n1<-mu.c1-exp(delta.n1)
  # variance of expression levels for cluster 1 for normal subjects
  tau.n1<-paraRP[7]; 
  sigma2.n1<-exp(tau.n1)
  # modified logit of correlation among expression levels for cluster 1 for normal subjects
  r.n1<-paraRP[8]; 
  rho.n1<-(exp(r.n1)-1/(nn-1))/(1+exp(r.n1))

  # mean expression level for cluster 2
  mu.2<-paraRP[9]; 
  # variance of expression levels for cluster 2
  tau.2<-paraRP[10]; 
  sigma2.2<-exp(tau.2)
  # modified logit of correlation among expression levels for cluster 2
  r.2<-paraRP[11]; 
  rho.2<-(exp(r.2)-1/(n-1))/(1+exp(r.2))

  # mean expression level for cluster 3 for diseased subjects
  mu.c3<-paraRP[12]; 
  # variance of expression levels for cluster 3 for diseased subjects
  tau.c3<-paraRP[13]; 
  sigma2.c3<-exp(tau.c3)
  # modified logit of correlation among expression levels for cluster 3 for diseased subjects
  r.c3<-paraRP[14]; 
  rho.c3<-(exp(r.c3)-1/(nc-1))/(1+exp(r.c3))

  # mean expression level for cluster 3 for normal subjects
  delta.n3<-paraRP[15]; 
  mu.n3<-mu.c3+exp(delta.n3)
  # variance of expression levels for cluster 3 for normal subjects
  tau.n3<-paraRP[16]; 
  sigma2.n3<-exp(tau.n3)
  # modified logit of correlation among expression levels for cluster 3 for normal subjects
  r.n3<-paraRP[17]; 
  rho.n3<-(exp(r.n3)-1/(nn-1))/(1+exp(r.n3))

  ##############################
  para<-c(pi.1, pi.2, pi.3,
               mu.c1, sigma2.c1, rho.c1,
               mu.n1, sigma2.n1, rho.n1,
               mu.2, sigma2.2, rho.2,
               mu.c3, sigma2.c3, rho.c3,
               mu.n3, sigma2.n3, rho.n3
  )
  names(para)<-paraNames

  return(para)
}

errRates<-function(obj.gsMMD)
{
  # number of genes
  p<-length(obj.gsMMD$memGenes2)
  # number of selected genes
  pr<-sum(obj.gsMMD$memGenes2==1, na.rm=TRUE)

  wiMat<-obj.gsMMD$wiMat
  # W2
  W2<-wiMat[,2]

  # calculate FDR
  flag<-wiMat[,2] < (apply(wiMat[,-2], 1, max, na.rm=TRUE))
  if(pr>0)
  { FDR<-sum(W2*flag, na.rm=TRUE)/pr }
  else
  { FDR<-0 }

  # calculate FNDR
  flag2<- wiMat[,2] == (apply(wiMat, 1, max, na.rm=TRUE))  
  pdiff<-p-pr 
  if(pdiff>0)
  { FNDR<-sum((1-W2)*flag2, na.rm=TRUE)/pdiff }
  else
  { FNDR<-0 }

  # calculate FPR
  denom<-sum(W2, na.rm=TRUE)
  if(denom>0)
  { FPR<-sum(W2*flag, na.rm=TRUE)/denom }
  else
  { FPR<-0 }

  # calculate FNR
  denom2<-sum(1-W2, na.rm=TRUE)
  if(denom2>0)
  { FNR<-sum((1-W2)*flag2, na.rm=TRUE)/denom2 }
  else
  { FNR<-0 }

  estRates<-c(FDR, FNDR, FPR, FNR)
  names(estRates)<-c("FDR", "FNDR", "FPR", "FNR")

  return(estRates)
}

!     the code in the loop from paraEst.R converted to Fortran

      subroutine paraestloop(w1, w2, w3, x, psi_m, memsubjects,
     +                  xctxc, xct1,
     +                  xntxn, xnt1,
     +                  xtx, xt1,
     +                  ngenes, n, nc, nn,
     +                  itmax, eps, quiet, loop)

      implicit none

      integer itmax, loop
      integer ngenes, n, nc, nn
      integer memsubjects(n)

      double precision x(ngenes,n), psi_m(17)
      double precision xctxc(ngenes), xct1(ngenes)
      double precision xntxn(ngenes), xnt1(ngenes)
      double precision xtx(ngenes), xt1(ngenes)
      double precision eps
      logical quiet

      double precision w1(ngenes), w2(ngenes), w3(ngenes)
      double precision sumw1, sumw2, sumw3

      double precision sumw1xctxc, sumw1xct1, sumw1xct1_sq
      double precision sumw1xntxn, sumw1xnt1, sumw1xnt1_sq
      double precision sumw2xtx, sumw2xt1, sumw2xt1_sq
      double precision sumw3xctxc, sumw3xct1, sumw3xct1_sq
      double precision sumw3xntxn, sumw3xnt1, sumw3xnt1_sq

      double precision const_b1, const_b2, const_b3
      double precision sumb
      double precision pivec_new(2)
      double precision psi_m_new(17)

      double precision xct1_sq(ngenes), xnt1_sq(ngenes), xt1_sq(ngenes)
      double precision sumtwo, sumd

      double precision thetaini(15)

      character*60 task
      integer isave(44)
      double precision dsave(29)

      integer i
      logical done

      done = .TRUE.

      const_b1 = 2d0
      const_b2 = 2d0
      const_b3 = 2d0
      sumb = const_b1 + const_b2 + const_b3


      do 10 i=1,ngenes
        xct1_sq(i) = xct1(i)**2
        xnt1_sq(i) = xnt1(i)**2
        xt1_sq(i) = xt1(i)**2
 10   continue


!     loop until convergence and stop if max number of loops is reached
      do 100 loop=0,itmax

!       check for user interrupt
        call rchkusr()

        if(.not. quiet) then
          call intpr('********** loop=', -1, loop, 1)
!          write(6,*)'************'
!          write(6,*)'loop=',loop
        end if


!       E-step
!       obtain weights w_{ij}(\Psi^{(m)}), i.e. E(Z_{ij} | x_i, Psi^{(m)})

!       checkpara was called within wiFun.R, however this will all be handled
!       in Fortran, so checking the dimension of Psi.m will not be included
!       in the future

        call wifun(w1, w2, w3, x, psi_m, memsubjects, 
     +            1.0d-6, ngenes, n, nc, nn       )


        sumw1 = sumd(w1,ngenes)
        sumw2 = sumd(w2,ngenes)
        sumw3 = sumd(w3,ngenes)

!        if (.not. quiet) then
!          write(6,*)'sumw1=',sumw1,' sumw2=',sumw2,' sumw3=',sumw3
!        end if


        sumw1xctxc = sumtwo(w1, xctxc, ngenes)
        sumw1xct1 = sumtwo(w1, xct1, ngenes)
        sumw1xct1_sq = sumtwo(w1, xct1_sq, ngenes)

        sumw1xntxn = sumtwo(w1, xntxn, ngenes)
        sumw1xnt1 = sumtwo(w1, xnt1, ngenes)
        sumw1xnt1_sq = sumtwo(w1, xnt1_sq, ngenes)

        sumw2xtx = sumtwo(w2, xtx, ngenes)
        sumw2xt1 = sumtwo(w2, xt1, ngenes)
        sumw2xt1_sq = sumtwo(w2, xt1_sq, ngenes)

        sumw3xctxc = sumtwo(w3, xctxc, ngenes)
        sumw3xct1 = sumtwo(w3, xct1, ngenes)
        sumw3xct1_sq = sumtwo(w3, xct1_sq, ngenes)
 
        sumw3xntxn = sumtwo(w3, xntxn, ngenes)
        sumw3xnt1 = sumtwo(w3, xnt1, ngenes)
        sumw3xnt1_sq = sumtwo(w3, xnt1_sq, ngenes)


!       M-step
!       mixture proportions
        pivec_new(1) = (sumw1+const_b1-1d0)/(ngenes+sumb-3d0)
        pivec_new(2) = (sumw2+const_b2-1d0)/(ngenes+sumb-3d0)


        do 20 i=1,15
          thetaini(i) = psi_m(i+2)
 20     continue

!        if(.not. quiet) then
!          write(6,*)
!          write(6,*)'thetaIni>>'
!          write(6,30)'     muc1= ',thetaini(1),
!     +               '   tau.c1= ',thetaini(2),
!     +               '     r.c1= ',thetaini(3),
!     +               ' delta.n1= ',thetaini(4),
!     +               '   tau.n1= ',thetaini(5)
!          write(6,30)'     r.n1= ',thetaini(6),
!     +               '     mu.2= ',thetaini(7),
!     +               '    tau.2= ',thetaini(8),
!     +               '      r.2= ',thetaini(9),
!     +               '    mu.c3= ',thetaini(10)
!          write(6,30)'   tau.c3= ',thetaini(11),
!     +               '     r.c3= ',thetaini(12),
!     +               ' delta.n3= ',thetaini(13),
!     +               '   tau.n3= ',thetaini(14),
!     +               '     r.n3= ',thetaini(15)
!        end if
! 30     format(A,D10.4,A,D10.4,A,D10.4,A,D10.4,A,D10.4)

        
        call lbfgsbdriver(task, isave, dsave, thetaini,
     +                  nc, nn, n,
     +                  sumw1, sumw2, sumw3,
     +                  sumw1xctxc, sumw1xct1, sumw1xct1_sq,
     +                  sumw1xntxn, sumw1xnt1, sumw1xnt1_sq,
     +                  sumw2xtx, sumw2xt1, sumw2xt1_sq,
     +                  sumw3xctxc, sumw3xct1, sumw3xct1_sq,
     +                  sumw3xntxn, sumw3xnt1, sumw3xnt1_sq       )

!        if(.not. quiet) then
!          write(6,*)
!          if(task(1:4) .eq. 'CONV') then
!            write(6,*)'message= ',task
!            write(6,*)'number of calls to fn and gr>> ',isave(34)
!          end if
!          write(6,*)
!        end if
!!        negQ<-res$dsave[2]


!       # update parameters
        psi_m_new(1) = pivec_new(1)
        psi_m_new(2) = pivec_new(2)
        do 40 i=1,15
          psi_m_new(i+2) = thetaini(i)
 40     continue

!       names(Psi.m.new)<-paraNamesRP

!        if(.not. quiet) then
!          write(6,*)'Psi.m.new>>'
!          write(6,50)'      pi1= ',psi_m_new(1),
!     +               '      pi2= ',psi_m_new(2)
!          write(6,55)'     muc1= ',psi_m_new(3),
!     +               '   tau.c1= ',psi_m_new(4),
!     +               '     r.c1= ',psi_m_new(5),
!     +               ' delta.n1= ',psi_m_new(6),
!     +               '   tau.n1= ',psi_m_new(7)
!          write(6,55)'     r.n1= ',psi_m_new(8),
!     +               '     mu.2= ',psi_m_new(9),
!     +               '    tau.2= ',psi_m_new(10),
!     +               '      r.2= ',psi_m_new(11),
!     +               '    mu.c3= ',psi_m_new(12)
!          write(6,55)'   tau.c3= ',psi_m_new(13),
!     +               '     r.c3= ',psi_m_new(14),
!     +               ' delta.n3= ',psi_m_new(15),
!     +               '   tau.n3= ',psi_m_new(16),
!     +               '     r.n3= ',psi_m_new(17)
!          write(6,*)
!        end if
! 50     format(A,D10.4,A,D10.4)
! 55     format(A,D10.4,A,D10.4,A,D10.4,A,D10.4,A,D10.4)


        done = .TRUE.

        do 60 i=1,17
          if(abs(psi_m_new(i)-psi_m(i)) .gt. eps) done = .FALSE.
 60     continue

        if(done) then
!         # the following statement is to 
!         # make sure eta.new is an update from current wMat
!         # so that d Q / d eta = 0
          do 70 i=1,17
            psi_m(i) = psi_m_new(i)
 70       continue
          return
        end if


        do 80 i=1,17
          psi_m(i) = psi_m_new(i)
 80     continue


 100  continue

      if(loop .ge. itmax) then
        call rwarn('***** Warning! ITMAX exceeded *****')
        call rwarn('EM algorithm did not converge!')
      end if

      return
      end




! ************************************************************************************
!     function to sum an array of size n

      function sumd(arr1, n)

      implicit none

      integer n
      double precision sumd
      double precision arr1(n)
      integer i

      sumd = 0d0

      do 100 i=1,n
        sumd = sumd + arr1(i)
 100  continue

      return
      end



! ************************************************************************************
!     function to sum an array that is the element by element
!     product of two arrays (arr1, arr2) 
      function sumtwo(arr1, arr2, n)

      implicit none

      integer n
      double precision sumtwo, sumd
      double precision arr1(n), arr2(n), arr3(n)
      integer i

      do 100 i=1,n
        arr3(i) = arr1(i)*arr2(i)
 100  continue

      sumtwo=sumd(arr3,n)

      return
      end



! ************************************************************************************
! find the position of the maximum element of memmat

      subroutine maxposfun(memgenes, memmat, 
     +                  ngenes, maxflag, thrspostprob)

      implicit none

      integer ngenes, memgenes(ngenes)
      double precision memmat(ngenes,3)
      logical maxflag
      double precision thrspostprob

      integer i

      if (maxflag) then
        do 100 i=1,ngenes
          if ((memmat(i,1) .gt. memmat(i,2)) .and. 
     +                   (memmat(i,1) .gt. memmat(i,3))) then
            memgenes(i) = 1
          else if ((memmat(i,2) .gt. memmat(i,1)) .and. 
     +                   (memmat(i,2) .gt. memmat(i,3))) then
            memgenes(i) = 2
          else if ((memmat(i,3) .gt. memmat(i,1)) .and. 
     +                   (memmat(i,3) .gt. memmat(i,2))) then
            memgenes(i) = 3
          end if
 100    continue
        return
      end if

      do 200 i=1,ngenes
          if ((memmat(i,1) .le. thrspostprob) .and.
     +      (memmat(i,2) .le. thrspostprob) .and.
     +      (memmat(i,3) .le. thrspostprob)   ) then
!           non-differentially expressed
            memgenes(i) = 2
          end if
 200  continue
      return

      end

# modified on Sept. 28, 2011
#  (1) added 'na.rm=TRUE' to functions 'sum' and 'mean'
#
transFunc<-function(x, transformMethod="boxcox", 
                    criterion=c("cor", "skewness", "kurtosis"), 
                    minL=-10, maxL=10, stepL=0.1, 
                    eps=1.0e-6, plotFlag=FALSE, ITMAX=0)
{
  transformMethod<-match.arg(transformMethod, choices=c("boxcox", "log2", "log10", "log", "none"))
  criterion<-match.arg(criterion, c("cor", "skewness", "kurtosis"))
  tmpx<-as.vector(x)
  minx<-min(tmpx, na.rm=TRUE)
  if(minx<0)
  { cat("****** Begin Warning ******** \n")
    cat("Warning: Data contains non-positive values! To continue box-cox transformation,\n")
    cat("We perform the following transformation:\n")
    cat("x<-x+abs(min(x, na.rm=TRUE))+1\n")
    cat("****** End Warning ******** \n")
    x<-x+abs(minx)+1 
  }

 
  if(transformMethod=="boxcox")
  { x2<-transBoxCoxMat(x, criterion, minL, maxL, stepL, eps, plotFlag, ITMAX)
    return(x2)
  }
  if(transformMethod=="log2")
  { 
    return(log2(x))
  }
  if(transformMethod=="log10")
  { 
    return(log10(x))
  }
  if(transformMethod=="log")
  { 
    return(log(x))
  }
  return(x)
}

transBoxCoxMat<-function(mat,
                      criterion=c("cor", "skewness", "kurtosis"),
                      minL=-10, 
                      maxL=10, 
                      stepL=0.1, 
                      eps=1.0e-3, 
                      plotFlag=FALSE, 
                      ITMAX=0)
{
  lambda.vec<-apply(mat, 2, transBoxCox2, criterion=criterion,
        minL=minL, maxL=maxL, stepL=stepL,
        eps=eps, plotFlag=plotFlag, ITMAX=ITMAX) 
  lambda<-mean(lambda.vec, na.rm=TRUE)
  #print(paste("******** average lambda=", lambda))


  mat.new<-apply(mat, 2, transBoxCox.default, lambda=lambda, eps=eps) 


  return(list(dat=mat.new, lambda.avg=lambda, lambda.vec=lambda.vec))
}

# box-cox transformation
transBoxCox2<-function(x, 
                      criterion=c("cor", "skewness", "kurtosis"),
                      minL=-10, 
                      maxL=10, 
                      stepL=0.1, 
                      eps=1.0e-3, 
                      plotFlag=FALSE, 
                      ITMAX=0)
{
  criterion<-match.arg(criterion, c("cor", "skewness", "kurtosis"))

  xold<-x 
  mycor<- -qqnormCorFunc(xold)
  skewness<-abs(skewnessFunc(xold)[2])
  kurtosis<-abs(kurtosisFunc(xold)[2])
  
  best.criVec<-c(mycor, skewness, kurtosis)
  names(best.criVec)<-c("cor", "skewness", "kurtosis")
  lambda.optim<-0

  loop<-0
  while(loop<=ITMAX)
  {
    loop<-loop+1
    #print(paste("******loop=", loop))
    tmp<-optimalLambda(xold, criterion, minL, maxL, stepL, eps, plotFlag)
    lambda<-tmp$lambda
    if(loop==1)
    { lambda.optim<-lambda }
    #print(paste("optimal lambda=", lambda))
    xnew<-transBoxCox.default(xold, lambda, eps)
    mycor<- -qqnormCorFunc(xnew)
    skewness<-abs(skewnessFunc(xnew)[2])
    kurtosis<-abs(kurtosisFunc(xnew)[2])
    criVec<-c(mycor, skewness, kurtosis)
    names(criVec)<-c("cor", "skewness", "kurtosis")

    diff<-max(abs(best.criVec-criVec), na.rm=TRUE)
    tmp<-sum(best.criVec<criVec, na.rm=TRUE)
    if(tmp>1 || diff<eps)
    { 
      break
    } 
    best.criVec<-criVec
    xold<-xnew
    lambda.optim<-lambda
  }
  return(lambda.optim)
}

# box-cox transformation
transBoxCox.default<-function(x, lambda, eps=1.0e-3)
{
  minx<-min(x, na.rm=TRUE)
  if(minx<0)
  { cat("****** Begin Warning ******** \n")
    cat("Warning: Data contains non-positive values! To continue box-cox transformation,\n")
    cat("We perform the following transformation:\n")
    cat("x<-x+abs(min(x, na.rm=TRUE))+1\n")
    cat("****** End Warning ******** \n")
    x<-x+abs(minx)+1 
  }

  if(abs(lambda)<eps)
  { return(log(x)) }
  res<-(x^lambda-1)/lambda
  return(res)
}


statVecFunc<-function(lambdai,x,eps,criterion=c("cor", "skewness", "kurtosis"))
{
    x2<-transBoxCox.default(x, lambdai, eps)
    if(criterion=="cor")
    { tmp<-qqnormCorFunc(x2) 
      statVecFunc<- -tmp
    }
    else if (criterion=="skewness")
    { tmp<-skewnessFunc(x2)
      statVecFunc<-abs(tmp[2])
    } else # criterion=="kurtosis"
    { tmp<-kurtosisFunc(x2)
      statVecFunc<-abs(tmp[2])
    }
    return(statVecFunc)
}

optimalLambda<-function(x, 
                        criterion=c("cor", "skewness", "kurtosis"), 
                        minL=-10, 
                        maxL=10, 
                        stepL=0.1, 
                        eps=1.0e-3, 
                        plotFlag=FALSE)
{
  criterion<-match.arg(criterion, c("cor", "skewness", "kurtosis"))


# search for the optimal lambda
  res <- optimize(   statVecFunc,c(minL,maxL),
                        x=x,eps=eps,criterion=criterion,tol=stepL,maximum=FALSE )

  lambda <- res$minimum

# statVec and lambdaVec are not calculated and put into arrays when using the lambda optimization
# unless a plot is required (for statVec, the value at optimal lambda is stored by default)
  statVec<- res$objective
  lambdaVec<-lambda

  if(plotFlag)
  { 
    lambdaVec<-seq(from=minL, to=maxL, by=stepL)
    statVec<-lapply(lambdaVec,statVecFunc,x=x,eps=eps,criterion=criterion)

    plot(lambdaVec, statVec, xlab="lambda", ylab="correlation")
    title(main="Box-Cox normality plot", sub=paste("optimal lambda=", lambda, sep=""))
  }

  res<-list(lambda=lambda, lambdaVec=lambdaVec, statVec=statVec)
  return(res)
}

qqnormCorFunc<-function(x)
{
  tmp<-qqnorm(x, plot.it=FALSE) 
  x<-as.vector(tmp$x)
  y<-as.vector(tmp$y)
  sd.x<-sd(x, na.rm=TRUE)
  sd.y<-sd(y, na.rm=TRUE)
  if(sd.x>0 && sd.y>0)
  { res<-cor(x, y, use="na.or.complete") }
  else {
    res<-0
  }

  return(res)
}

skewnessFunc<-function(x)
{
  n<-length(x)
  xbar<-mean(x, na.rm=TRUE)
  numer<-sum((x-xbar)^3, na.rm=TRUE)*sqrt(n)
  denom<-(sum((x-xbar)^2, na.rm=TRUE))^(3/2)
  g1<-numer/denom
  G1<-g1*sqrt(n*(n-1))/(n-2)
  res<-c(g1, G1)
  names(res)<-c("g1", "G1")
  return(res)
}

kurtosisFunc<-function(x)
{
  n<-length(x)
  xbar<-mean(x, na.rm=TRUE)
  numer<-sum((x-xbar)^4, na.rm=TRUE)*n
  denom<-(sum((x-xbar)^2, na.rm=TRUE))^2
  g2<-numer/denom-3

  k2<-sum((x-xbar)^2, na.rm=TRUE)/(n-1)
  part1numer<-(n+1)*n
  part1denom<-(n-1)*(n-2)*(n-3)
  part1<-part1numer/part1denom

  part2numer<-sum((x-xbar)^4, na.rm=TRUE)
  part2<-part2numer/k2^2

  part3<-3*(n-1)^2/((n-2)*(n-3))

  G2<-part1*part2-part3

  res<-c(g2, G2)
  names(res)<-c("g2", "G2")

  return(res)
}

"gsMMD2"<-
function(obj.eSet, 
         memSubjects, 
         memIni,
         maxFlag=TRUE, 
         thrshPostProb=0.50, 
         geneNames=NULL, 
         alpha=0.05, 
         transformFlag=FALSE, 
         transformMethod="boxcox", 
         scaleFlag=TRUE, 
         criterion=c("cor", "skewness", "kurtosis"),
         minL=-10, 
         maxL=10, 
         stepL=0.1, 
         eps=1.0e-3, 
         ITMAX=100, 
         plotFlag=FALSE,
         quiet=TRUE)
{
  # get expression level matrix
  X<-exprs(obj.eSet)

  res<-gsMMD2.default(X, 
         memSubjects, 
         memIni,
         maxFlag, 
         thrshPostProb, 
         geneNames, 
         alpha, 
         transformFlag,
         transformMethod,
         scaleFlag,
         criterion,
         minL,
         maxL,
         stepL,
         eps,
         ITMAX,
         plotFlag,
         quiet)

  invisible(res)

}


"gsMMD2.default"<-
function(X, 
         memSubjects, 
         memIni,
         maxFlag=TRUE, 
         thrshPostProb=0.50, 
         geneNames=NULL, 
         alpha=0.05, 
         transformFlag=FALSE, 
         transformMethod="boxcox", 
         scaleFlag=TRUE, 
         criterion=c("cor", "skewness", "kurtosis"),
         minL=-10, 
         maxL=10, 
         stepL=0.1, 
         eps=1.0e-3, 
         ITMAX=100, 
         plotFlag=FALSE,
         quiet=TRUE)
{

  transformMethod<-match.arg(transformMethod, choices=c("boxcox", "log2", "log10", "log", "none"))
  criterion<-match.arg(criterion, c("cor", "skewness", "kurtosis"))

  X<-as.matrix(X)
  nGenes<-nrow(X)
  nSubjects<-ncol(X)
  nCases<-sum(memSubjects==1, na.rm=TRUE)
  nControls<-sum(memSubjects==0, na.rm=TRUE)

  if(sum(is.null(geneNames), na.rm=TRUE))
  { geneNames<-paste("gene", 1:nGenes, sep="") }

  cat("Programming is running. Please be patient...\n")
  lambda<-NA
  if(transformFlag)
  { 
    if(transformMethod!="none")
    {
      vec<-as.numeric(X)
      min.vec<-min(vec, na.rm=TRUE)
      if(min.vec<0)
      {
        cat("****** Begin Warning ******** \n")
        cat("Warning: Data contains non-positive values! To continue ",
           transformMethod, " transformation,\n")
        cat("We first perform the following transformation:\n")
        cat("x<-x+abs(min(x, na.rm=TRUE))+1\n")
        cat("****** End Warning ******** \n")
    
        X<-X+abs(min.vec)+1
      }
    }
    tmp<-transFunc(X, transformMethod, criterion, 
                   minL, maxL, stepL, eps, plotFlag, ITMAX=0) 
    if(transformMethod=="boxcox")
    { X<-tmp$dat 
      lambda<-tmp$lambda.avg
    }
    else {
     X<-tmp
    }
    if(!quiet)
    { cat(paste("Data transformation (", transformMethod, ") performed\n")) }
  }

  if(scaleFlag)
  {
    if(!quiet)
    { cat("Gene profiles are scaled so that they have mean zero and variance one!\n") }
    X<-t(apply(X, 1, scale, center=TRUE, scale=TRUE))

    # to avoid linear dependence of tissue samples after scaling
    # gene profiles, we delete a tissue sample.
    # We arbitrarily select the tissue sample, which has the biggest label number, 
    # from the tissue sample group that has larger size than the other 
    # tissue sample group. For example, if there are 6 cancer tissue samples 
    # and 10 normal tissue samples, we delete the 10-th normal tissue sample after scaling.

    if(nCases>nControls)
    { 
      pos<-which(memSubjects==1)
      pos2<-pos[nCases]
      X<-X[,-pos2]
      memSubjects<-memSubjects[-pos2]
    } else {
      pos<-which(memSubjects==0)
      pos2<-pos[nControls]
      X<-X[,-pos2]
      memSubjects<-memSubjects[-pos2]
    }
    nCases<-sum(memSubjects==1, na.rm=TRUE)
    nControls<-sum(memSubjects==0, na.rm=TRUE)
    nSubjects<-nCases+nControls
  }

  #cat("Programming is running. Please be patient...\n")
  # records initial parameter estimates
  tmpIni<-getPara(X, memSubjects, memIni, eps)
  paraIniRP<-tmpIni$para

  # records initial log-likelihood estimates
  llkhIni<-tmpIni$llkh

  # records E(z_{ij} | x_i, Psi^{(m)})
  wiMat<-matrix(0, nrow=nGenes, ncol=3)
  # Gene Selection based on EM algorithm
  res<- paraEst(X, paraIniRP, memSubjects=memSubjects, 
                 maxFlag=maxFlag, thrshPostProb, geneNames=geneNames,
                 ITMAX=ITMAX, eps=eps, quiet=quiet)

  if(res$loop==0)
  {
    paraRP<-paraIniRP
    llkh<-llkhIni
    memGenes<-memIni
    wiMat<-res$memMat
  } else {
    paraRP<-res$para
    llkh<-res$llkh
    memGenes<-res$memGenes
    wiMat<-res$memMat
  }
  para<-paraRPConverter(paraRP, nCases, nControls)
  names(para)<-paraNames
  names(paraRP)<-paraNamesRP

  names(memGenes)<-geneNames
  rownames(wiMat)<-geneNames
  colnames(wiMat)<-paste("cluster", 1:3, sep="")

  memGenes2<-rep(1, nGenes)
  memGenes2[memGenes==2]<-0 # non-differentially expressed genes

  paraIni<-paraRPConverter(paraIniRP, nCases, nControls)
  names(paraIni)<-paraNames
  names(paraIniRP)<-paraNamesRP

  if(!quiet)
  {
    cat("*******************************************************\n\n")
    cat("Initial parameter estimates>>\n"); print(round(paraIni,3)); cat("\n");
    cat("Initial loglikelihood>>\n"); print(round(llkhIni,3)); cat("\n");
    cat("Final parameter estimates>>\n"); print(round(para,3)); cat("\n");
    cat("Final loglikelihood>>\n"); print(round(llkh,3)); cat("\n");
    cat("*******************************************************\n\n")
  }
  
  res<-list(dat=X, memSubjects=memSubjects, 
            memGenes=memGenes, memGenes2=memGenes2, 
            para=para, 
            llkh=llkh, wiMat=wiMat, 
            memIni=memIni, paraIni=paraIni, llkhIni=llkhIni,
            lambda=lambda)
  invisible(res) 
}

      subroutine timer(ttime)
      double precision ttime
c
      real temp
c
c     This routine computes cpu time in double precision; it makes use of 
c     the intrinsic f90 cpu_time therefore a conversion type is
c     needed.
c
c           J.L Morales  Departamento de Matematicas, 
c                        Instituto Tecnologico Autonomo de Mexico
c                        Mexico D.F.
c
c           J.L Nocedal  Department of Electrical Engineering and
c                        Computer Science.
c                        Northwestern University. Evanston, IL. USA
c                         
c                        January 21, 2011
c
      temp = sngl(ttime)
      call cpu_time(temp)
      ttime = dble(temp) 

      return

      end
      

!     converted to Fortran 77, June 2012
!
!# modified on Sept. 28, 2011
!#  (1) added 'na.rm=TRUE' to function 'sum'
!#
!# sum_{i=1}^{nGenes} log10(pi.1*f1(xi)+pi.2*f2(xi)+pi.3*f3(xi))

!# negative of the M-step's objective function - the part relating to pi's
!# Q(Psi)=sum_{i=1}^{nGenes} wi*V(pi)+\sum_{i=1}^{nGenes) wi*U(theta)


!     NOTE: removed dat argument
!      function negqfunc(theta, dat, nc, nn, n, &

      SUBROUTINE negqfunc(negq, theta, nc, nn, n,
     +      sumw1, sumw2, sumw3,
     +      sumw1xcTxc, sumw1xcT1, sumw1xcT1_sq,
     +      sumw1xnTxn, sumw1xnT1, sumw1xnT1_sq,
     +      sumw2xTx, sumw2xT1, sumw2xT1_sq,
     +      sumw3xcTxc, sumw3xcT1, sumw3xcT1_sq,
     +      sumw3xnTxn, sumw3xnT1, sumw3xnT1_sq  )

      implicit none

      double precision theta(15)
      double precision negq

      double precision nc, nn, n

      double precision sumw1, sumw2, sumw3
      double precision sumw1xcTxc, sumw1xcT1, sumw1xcT1_sq
      double precision sumw1xnTxn, sumw1xnT1, sumw1xnT1_sq
      double precision sumw2xTx, sumw2xT1, sumw2xT1_sq
      double precision sumw3xcTxc, sumw3xcT1, sumw3xcT1_sq
      double precision sumw3xnTxn, sumw3xnT1, sumw3xnT1_sq

      double precision mu_c1, tau_c1, r_c1, rho_c1, delta_n1, mu_n1
      double precision tau_n1, r_n1, rho_n1, mu_2, tau_2, r_2, rho_2
      double precision mu_c3, tau_c3, r_c3, rho_c3, delta_n3, mu_n3
      double precision tau_n3, r_n3, rho_n3

      double precision sumw1acTac,sumw1acT1_sq,sumw1anTan,sumw1anT1_sq
      double precision sumw2aTa,sumw2aT1_sq,sumw3acTac,sumw3acT1_sq
      double precision sumw3anTan,sumw3anT1_sq

      double precision tt, part1, part2, part3
      double precision part1_1, part1_2, part1_3
      double precision part1_4, part1_5, part1_6
      double precision part2_1, part2_2, part2_3
      double precision part3_1, part3_2, part3_3, part3_4
      double precision part3_5, part3_6

      logical risnan

      double precision pi

      pi = 2d0*asin(1d0)


!     mean expression level for cluster 1 for diseased subjects
      mu_c1 = theta(1)
!     log(variance) of expression levels for cluster 1 for diseased subjects
      tau_c1 = theta(2) 
!     modified logit of correlation among expression levels for cluster 1 for diseased subjects
      r_c1 = theta(3)
 
      rho_c1 = (exp(r_c1)-1d0/(nc-1d0))/(1d0+exp(r_c1))

!     mean expression level for cluster 1 for normal subjects
      delta_n1 = theta(4)
      mu_n1 =  mu_c1-exp(delta_n1)
!     log(variance) of expression levels for cluster 1 for normal subjects
      tau_n1 = theta(5)
!     modified logit of correlation among expression levels for cluster 1 for normal subjects
      r_n1 =  theta(6)
      rho_n1 = (exp(r_n1)-1d0/(nn-1d0))/(1d0+exp(r_n1))
!     mean expression level for cluster 2
      mu_2 = theta(7)
!     log(variance) of expression levels for cluster 2
      tau_2 = theta(8)
!     modified logit of correlation among expression levels for cluster 2
      r_2 = theta(9)
      rho_2 = (exp(r_2)-1d0/(n-1d0))/(1d0+exp(r_2))

!     mean expression level for cluster 3 for diseased subjects
      mu_c3 =  theta(10)
!     log(variance) of expression levels for cluster 3 for diseased subjects
      tau_c3 = theta(11)
!     modified logit of correlation among expression levels for cluster 3 for diseased subjects
      r_c3 = theta(12)

      rho_c3 = (exp(r_c3)-1d0/(nc-1d0))/(1d0+exp(r_c3))

!     mean expression level for cluster 3 for normal subjects
      delta_n3 = theta(13)
      mu_n3 = mu_c3+exp(delta_n3)
!     log(variance) of expression levels for cluster 3 for normal subjects
      tau_n3 = theta(14)
!     modified logit of correlation among expression levels for cluster 3 for normal subjects
      r_n3 = theta(15)
      rho_n3 = (exp(r_n3)-1d0/(nn-1d0))/(1d0+exp(r_n3))

!     ###########
      sumw1acTac = sumw1xcTxc-2d0*mu_c1*sumw1xcT1+nc*mu_c1**2*sumw1
      sumw1acT1_sq = sumw1xcT1_sq-2d0*nc*mu_c1*sumw1xcT1+
     +        nc**2*mu_c1**2*sumw1

      sumw1anTan = sumw1xnTxn-2d0*mu_n1*sumw1xnT1+nn*mu_n1**2*sumw1
      sumw1anT1_sq = sumw1xnT1_sq-2d0*nn*mu_n1*sumw1xnT1+
     +        nn**2*mu_n1**2*sumw1
 
      sumw2aTa = sumw2xTx-2d0*mu_2*sumw2xT1+n*mu_2**2*sumw2
      sumw2aT1_sq = sumw2xT1_sq-2d0*n*mu_2*sumw2xT1+
     +        n**2*mu_2**2*sumw2

      sumw3acTac = sumw3xcTxc-2d0*mu_c3*sumw3xcT1+nc*mu_c3**2*sumw3
      sumw3acT1_sq = sumw3xcT1_sq-2d0*nc*mu_c3*sumw3xcT1+
     +        nc**2*mu_c3**2*sumw3
 
      sumw3anTan = sumw3xnTxn-2d0*mu_n3*sumw3xnT1+nn*mu_n3**2*sumw3
      sumw3anT1_sq = sumw3xnT1_sq-2d0*nn*mu_n3*sumw3xnT1+
     +        nn**2*mu_n3**2*sumw3

!     ##########

      part1_1 = -nc*log(2d0*pi)/2d0-nc*tau_c1/2d0-nc*log(nc)/2d0+
     +        (nc-1d0)*log(nc-1d0)/2d0

      tt = exp(r_c1)
      if (tt .lt. 1.0d308) then
            part1_1 = part1_1+nc*log(1d0+exp(r_c1))/2d0-r_c1/2d0
      else
            part1_1 = part1_1+nc*r_c1/2d0-r_c1/2d0
      endif
      part1_1 = part1_1 * sumw1

      part1_2 = -(nc-1d0)*(exp(-tau_c1)+
     +        exp(r_c1-tau_c1))/(2d0*nc)*sumw1acTac

      part1_3 = ((nc-2d0)*exp(-tau_c1)+exp(r_c1-tau_c1)*(nc-1d0)-
     +        exp(-r_c1-tau_c1))/(2d0*nc**2)*sumw1acT1_sq

      part1_4 = -nn*log(2d0*pi)/2d0-nn*tau_n1/2d0-nn*log(nn)/2d0+
     +        (nn-1d0)*log(nn-1d0)/2d0

      tt = exp(r_n1)
      if (tt .lt. 1.0d308) then
            part1_4 = part1_4+nn*log(1d0+exp(r_n1))/2d0-r_n1/2d0
      else
            part1_4 = part1_4+nn*r_n1/2d0-r_n1/2d0
      endif
      part1_4 = part1_4 * sumw1

      tt = exp(r_n1-tau_n1)
      if (tt .lt. 1.0d308) then
            part1_5 = -(nn-1d0)*(exp(-tau_n1)+
     +        exp(r_n1-tau_n1))/(2d0*nn)*sumw1anTan

            part1_6 = ((nn-2d0)*exp(-tau_n1)+exp(r_n1-tau_n1)*(nn-1d0)-
     +        exp(-r_n1-tau_n1))/(2d0*nn**2)*sumw1anT1_sq

            part1 = part1_1+part1_2+part1_3+part1_4+part1_5+part1_6
      else
            part1 = part1_1+part1_2+part1_3+part1_4
      endif

!     #
!     #  part1<-part1.1+part1.2+part1.3+part1.4+part1.5+part1.6
!     #
!     ##########
      part2_1 = -n*log(2d0*pi)/2d0-n*tau_2/2d0-n*log(n)/2d0+
     +            (n-1d0)*log(n-1d0)/2d0
      tt = exp(r_2)
      if (tt .lt. 1.0d308) then
            part2_1 = part2_1+n*log(1d0+exp(r_2))/2d0-r_2/2d0
      else
            part2_1 = part2_1+n*r_2/2d0-r_2/2d0
      endif
      part2_1 = part2_1 * sumw2

      part2_2 = -(n-1d0)*(exp(-tau_2)+exp(r_2-tau_2))/(2d0*n)*sumw2aTa
      part2_3 = ((n-2d0)*exp(-tau_2)+exp(r_2-tau_2)*(n-1d0)-
     +        exp(-r_2-tau_2))/(2d0*n**2)*sumw2aT1_sq

      if (.not. risnan(part2_2)) then
            part2 = part2_1+part2_2+part2_3
      else
            part2 = part2_1+part2_3
      endif


!     ##########

      part3_1 = -nc*log(2d0*pi)/2d0-nc*tau_c3/2d0-
     +        nc*log(nc)/2d0+(nc-1d0)*log(nc-1d0)/2d0

      tt = exp(r_c3)
      if (tt .lt. 1.0d308) then
            part3_1 = part3_1+nc*log(1d0+exp(r_c3))/2d0-r_c3/2d0
      else
            part3_1 = part3_1+nc*r_c3/2d0-r_c3/2d0
      endif
      part3_1 = part3_1 * sumw3

      part3_2 = -(nc-1d0)*(exp(-tau_c3)+
     +        exp(r_c3-tau_c3))/(2d0*nc)*sumw3acTac
      part3_3 = ((nc-2d0)*exp(-tau_c3)+exp(r_c3-tau_c3)*(nc-1d0)-
     +        exp(-r_c3-tau_c3))/(2d0*nc**2)*sumw3acT1_sq

      part3_4 = -nn*log(2d0*pi)/2d0-nn*tau_n3/2d0-
     +        nn*log(nn)/2d0+(nn-1d0)*log(nn-1d0)/2d0

      tt = exp(r_n3)
      if (tt .lt. 1.0d308) then
            part3_4 = part3_4+nn*log(1d0+exp(r_n3))/2d0-r_n3/2d0
      else
            part3_4 = part3_4+nn*r_n3/2d0-r_n3/2d0
      endif
      part3_4 = part3_4 * sumw3

      part3_5 = -(nn-1d0)*(exp(-tau_n3)+
     +        exp(r_n3-tau_n3))/(2d0*nn)*sumw3anTan
      part3_6 = ((nn-2d0)*exp(-tau_n3)+exp(r_n3-tau_n3)*(nn-1d0)-
     +        exp(-r_n3-tau_n3))/(2d0*nn**2)*sumw3anT1_sq

      part3 = part3_1+part3_2+part3_3+part3_4+part3_5+part3_6


!     #Qfunc<-part0+part1+part2+part3
      negq = part1+part2+part3

      if (.not. risnan(negq)) then
            negq = -negq
      else
            negq = (1.0d308)
      endif


      return
      end




!#    first derivatives of the negative of the objective function
!#    Q(Psi)=sum_{i=1}^{nGenes} wi*V(pi)+\sum_{i=1}^{nGenes) wi*U(theta)


!     NOTE: removed dat argument
!      function dnegqfunc(theta, dat, nc, nn, n, &

      SUBROUTINE dnegqfunc(dnegq, theta, nc, nn, n, 
     +      sumw1, sumw2, sumw3, 
     +      sumw1xcTxc, sumw1xcT1, sumw1xcT1_sq, 
     +      sumw1xnTxn, sumw1xnT1, sumw1xnT1_sq, 
     +      sumw2xTx, sumw2xT1, sumw2xT1_sq, 
     +      sumw3xcTxc, sumw3xcT1, sumw3xcT1_sq, 
     +      sumw3xnTxn, sumw3xnT1, sumw3xnT1_sq  )

      implicit none

      double precision theta(15)
      double precision dnegq(15)

      double precision nc, nn, n

      double precision sumw1, sumw2, sumw3
      double precision sumw1xcTxc, sumw1xcT1, sumw1xcT1_sq
      double precision sumw1xnTxn, sumw1xnT1, sumw1xnT1_sq
      double precision sumw2xTx, sumw2xT1, sumw2xT1_sq
      double precision sumw3xcTxc, sumw3xcT1, sumw3xcT1_sq
      double precision sumw3xnTxn, sumw3xnT1, sumw3xnT1_sq

      double precision mu_c1, tau_c1, r_c1, rho_c1, delta_n1, mu_n1
      double precision tau_n1, r_n1, rho_n1, mu_2, tau_2, r_2, rho_2
      double precision mu_c3, tau_c3, r_c3, rho_c3, delta_n3, mu_n3
      double precision tau_n3, r_n3, rho_n3

      double precision sumw1acTac,sumw1acT1_sq,sumw1anTan,sumw1anT1_sq
      double precision sumw2aTa,sumw2aT1_sq,sumw3acTac,sumw3acT1_sq
      double precision sumw3anTan,sumw3anT1_sq

      double precision dqdmu_c1,dqddelta_n1,dqdmu_2,dqdmu_c3,dqddelta_n3
      double precision dqdtau_c1,dqdtau_n1,dqdtau_2,dqdtau_c3,dqdtau_n3
      double precision dqdr_c1,dqdr_n1,dqdr_2,dqdr_c3,dqdr_n3
      double precision tt

      logical risnan

      integer i


!     mean expression level for cluster 1 for diseased subjects
      mu_c1 = theta(1)
!     log(variance) of expression levels for cluster 1 for diseased subjects
      tau_c1 = theta(2)
!     modified logit of correlation among expression levels for cluster 1 for diseased subjects
      r_c1 = theta(3)
      rho_c1 = (exp(r_c1)-1d0/(nc-1d0))/(1d0+exp(r_c1))

!     mean expression level for cluster 1 for normal subjects
      delta_n1 = theta(4)
      mu_n1 = mu_c1-exp(delta_n1)
!     log(variance) of expression levels for cluster 1 for normal subjects
      tau_n1 = theta(5)
!     modified logit of correlation among expression levels for cluster 1 for normal subjects
      r_n1 = theta(6)
      rho_n1 = (exp(r_n1)-1d0/(nn-1d0))/(1d0+exp(r_n1))

!     mean expression level for cluster 2
      mu_2 = theta(7)
!     log(variance) of expression levels for cluster 2
      tau_2 = theta(8)
!     modified logit of correlation among expression levels for cluster 2
      r_2 = theta(9)
      rho_2 = (exp(r_2)-1d0/(n-1d0))/(1d0+exp(r_2))

!     mean expression level for cluster 3 for diseased subjects
      mu_c3 = theta(10)
!     log(variance) of expression levels for cluster 3 for diseased subjects
      tau_c3 = theta(11)
!     modified logit of correlation among expression levels for cluster 3 for diseased subjects
      r_c3 = theta(12)
      rho_c3 = (exp(r_c3)-1d0/(nc-1d0))/(1d0+exp(r_c3))

!     mean expression level for cluster 3 for normal subjects
      delta_n3 = theta(13)
      mu_n3 = mu_c3+exp(delta_n3)
!     log(variance) of expression levels for cluster 3 for normal subjects
      tau_n3 = theta(14)
!     modified logit of correlation among expression levels for cluster 3 for normal subjects
      r_n3 = theta(15)
      rho_n3 = (exp(r_n3)-1d0/(nn-1d0))/(1d0+exp(r_n3))

!     ###########
      sumw1acTac = sumw1xcTxc-2d0*mu_c1*sumw1xcT1+nc*mu_c1**2*sumw1
      sumw1acT1_sq = sumw1xcT1_sq-2d0*nc*mu_c1*sumw1xcT1+
     +        nc**2*mu_c1**2*sumw1
 
      sumw1anTan = sumw1xnTxn-2d0*mu_n1*sumw1xnT1+nn*mu_n1**2*sumw1
      sumw1anT1_sq = sumw1xnT1_sq-2d0*nn*mu_n1*sumw1xnT1+
     +        nn**2*mu_n1**2*sumw1
 
      sumw2aTa = sumw2xTx-2d0*mu_2*sumw2xT1+n*mu_2**2*sumw2
      sumw2aT1_sq = sumw2xT1_sq-2d0*n*mu_2*sumw2xT1+
     +        n**2*mu_2**2*sumw2
 
      sumw3acTac = sumw3xcTxc-2d0*mu_c3*sumw3xcT1+nc*mu_c3**2*sumw3
      sumw3acT1_sq = sumw3xcT1_sq-2d0*nc*mu_c3*sumw3xcT1+
     +        nc**2*mu_c3**2*sumw3
 
      sumw3anTan = sumw3xnTxn-2d0*mu_n3*sumw3xnT1+nn*mu_n3**2*sumw3
      sumw3anT1_sq = sumw3xnT1_sq-2d0*nn*mu_n3*sumw3xnT1+
     +        nn**2*mu_n3**2*sumw3


!     ##########
      dqdmu_c1 = (exp(-tau_c1)+exp(-r_c1-tau_c1))*
     +        (sumw1xcT1-nc*mu_c1*sumw1)/nc
      dqdmu_c1 = dqdmu_c1+(exp(-tau_n1)+exp(-r_n1-tau_n1))*
     +        (sumw1xnT1-nn*mu_n1*sumw1)/nn

      dqddelta_n1 = -exp(delta_n1)*(exp(-tau_n1)+exp(-r_n1-tau_n1))*
     +        (sumw1xnT1-nn*mu_n1*sumw1)/nn

      dqdmu_2 = (exp(-tau_2)+exp(-r_2-tau_2))*(sumw2xT1-n*mu_2*sumw2)/n

      dqdmu_c3 = (exp(-tau_c3)+exp(-r_c3-tau_c3))*
     +        (sumw3xcT1-nc*mu_c3*sumw3)/nc

      dqdmu_c3 = dqdmu_c3+(exp(-tau_n3)+exp(-r_n3-tau_n3))*
     +        (sumw3xnT1-nn*mu_n3*sumw3)/nn

      dqddelta_n3 = exp(delta_n3)*(exp(-tau_n3)+exp(-r_n3-tau_n3))*
     +        (sumw3xnT1-nn*mu_n3*sumw3)/nn

!     #####
      dqdtau_c1 = -nc*sumw1/2d0+(nc-1d0)*
     +            (exp(-tau_c1)+exp(r_c1-tau_c1))*
     +            sumw1acTac/(2d0*nc)

      dqdtau_c1 = dqdtau_c1-((nc-2d0)*exp(-tau_c1)+(nc-1d0)*
     +         exp(r_c1-tau_c1)-exp(-tau_c1-r_c1))*
     +         sumw1acT1_sq/(2d0*nc**2)

      dqdtau_n1 = -nn*sumw1/2d0+(nn-1d0)*
     +        (exp(-tau_n1)+exp(r_n1-tau_n1))*sumw1anTan/(2d0*nn)

      dqdtau_n1 = dqdtau_n1-((nn-2d0)*exp(-tau_n1)+(nn-1d0)*
     +            exp(r_n1-tau_n1)-exp(-tau_n1-r_n1))*
     +            sumw1anT1_sq/(2d0*nn**2)

      dqdtau_2 = -n*sumw2/2d0+(n-1d0)*(exp(-tau_2)+exp(r_2-tau_2))*
     +        sumw2aTa/(2d0*n)

      dqdtau_2 = dqdtau_2 - ((n-2d0)*exp(-tau_2)+(n-1d0)*exp(r_2-tau_2)-
     +        exp(-tau_2-r_2))*sumw2aT1_sq/(2d0*n**2)


      dqdtau_c3 = -nc*sumw3/2d0+(nc-1d0)*(exp(-tau_c3)+
     +        exp(r_c3-tau_c3))*sumw3acTac/(2d0*nc)

      dqdtau_c3 = dqdtau_c3 - ((nc-2d0)*exp(-tau_c3)+(nc-1d0)*
     +            exp(r_c3-tau_c3)-exp(-tau_c3-r_c3))*
     +            sumw3acT1_sq/(2d0*nc**2)

      dqdtau_n3 = -nn*sumw3/2d0+(nn-1d0)*
     +            (exp(-tau_n3)+exp(r_n3-tau_n3))*
     +            sumw3anTan/(2d0*nn)

      dqdtau_n3 = dqdtau_n3 - ((nn-2d0)*exp(-tau_n3)+(nn-1d0)*
     +            exp(r_n3-tau_n3)-exp(-tau_n3-r_n3))*
     +            sumw3anT1_sq/(2d0*nn**2)


!     #####
      tt = exp(r_c1)
      if (tt .lt. 1.0d308) then
            dqdr_c1 = (nc*exp(r_c1)/(2d0*(1d0+exp(r_c1)))-0.5d0)*sumw1
      else
            dqdr_c1 = (nc/(2d0*(1d0+exp(-r_c1)))-0.5d0)*sumw1
      endif
      dqdr_c1 = dqdr_c1-(nc-1d0)*exp(r_c1-tau_c1)*sumw1acTac/(2d0*nc)
      dqdr_c1 = dqdr_c1+(exp(r_c1-tau_c1)*(nc-1d0)+exp(-r_c1-tau_c1))*
     +        sumw1acT1_sq/(2d0*nc**2)

!     #####
      tt = exp(r_n1)
      if (tt .lt. 1.0d308) then
            dqdr_n1 = (nn*exp(r_n1)/(2d0*(1+exp(r_n1)))-0.5d0)*sumw1
      else
            dqdr_n1 = (nn/(2d0*(exp(-r_n1)+1d0))-0.5d0)*sumw1
      endif
      dqdr_n1 = dqdr_n1-(nn-1d0)*exp(r_n1-tau_n1)*sumw1anTan/(2d0*nn)

      dqdr_n1 = dqdr_n1+(exp(r_n1-tau_n1)*(nn-1d0)+exp(-r_n1-tau_n1))*
     +        sumw1anT1_sq/(2d0*nn**2)


!     #####
      tt = exp(r_2)
      if (tt .lt. 1.0d308) then
            dqdr_2 = (n*exp(r_2)/(2d0*(1d0+exp(r_2)))-0.5d0)*sumw2
      else
            dqdr_2 = (n/(2d0*(exp(-r_2)+1d0))-0.5d0)*sumw2
      endif
      dqdr_2 = dqdr_2-(n-1d0)*exp(r_2-tau_2)*sumw2aTa/(2d0*n)
      dqdr_2 = dqdr_2+(exp(r_2-tau_2)*(n-1d0)+exp(-r_2-tau_2))*
     +        sumw2aT1_sq/(2d0*n**2)

!     #####
      tt = exp(r_c3)
      if (tt .lt. 1.0d308) then
            dqdr_c3 = (nc*exp(r_c3)/(2d0*(1d0+exp(r_c3)))-0.5d0)*sumw3
      else
            dqdr_c3 = (nc/(2d0*(1d0+exp(-r_c3)))-0.5d0)*sumw3
      endif
      dqdr_c3 = dqdr_c3-(nc-1d0)*exp(r_c3-tau_c3)*sumw3acTac/(2d0*nc)
      dqdr_c3 = dqdr_c3+(exp(r_c3-tau_c3)*(nc-1d0)+exp(-r_c3-tau_c3))*
     +        sumw3acT1_sq/(2d0*nc**2)

!     #####
      tt = exp(r_n3)
      if (tt .lt. 1.0d308) then
            dqdr_n3 = (nn*exp(r_n3)/(2d0*(1d0+exp(r_n3)))-0.5d0)*sumw3
      else
            dqdr_n3 = (nn/(2d0*(1d0+exp(-r_n3)))-0.5d0)*sumw3
      endif
      dqdr_n3 = dqdr_n3-(nn-1d0)*exp(r_n3-tau_n3)*sumw3anTan/(2*nn)
      dqdr_n3 = dqdr_n3+(exp(r_n3-tau_n3)*(nn-1d0)+exp(-r_n3-tau_n3))*
     +        sumw3anT1_sq/(2d0*nn**2)

!     #####

      dnegq(1) = dqdmu_c1
      dnegq(2) = dqdtau_c1
      dnegq(3) = dqdr_c1
      dnegq(4) = dqddelta_n1
      dnegq(5) = dqdtau_n1
      dnegq(6) = dqdr_n1
      dnegq(7) = dqdmu_2
      dnegq(8) = dqdtau_2
      dnegq(9) = dqdr_2
      dnegq(10)= dqdmu_c3
      dnegq(11)= dqdtau_c3
      dnegq(12)= dqdr_c3
      dnegq(13)= dqddelta_n3
      dnegq(14)= dqdtau_n3
      dnegq(15)= dqdr_n3

!     no NaN or infinity values
      do 100 i=1,15
            if ((abs(dnegq(i)) .gt. 1.0d308) .or. risnan(dnegq(i))) then
                 dnegq(i) = 0d0
            else
                 dnegq(i) = -dnegq(i)
            endif
  100 continue


      return
      end

!     provides the t-test data for two matrices of data
!     results are provided for each row (gene) in the matrices
      subroutine myttest(xc, xn, nc, nn, ngenes, pvals, tstats)

      implicit none

      integer nc, nn, ngenes
      integer i, j
      double precision xc(ngenes,nc), xn(ngenes,nn)
      double precision pvals(ngenes), tstats(ngenes)

      double precision tmpxc(nc), tmpxn(nn)
      double precision sumxc, meanxc, sumxn, meanxn
      double precision varxc, varxn
      double precision sumd
      double precision tmpt, df, rpt

      integer nnc, nnn
      logical risnan


!     loop through the genes
!     exclude the NA values in the xc and xn matrices
      do 100 i=1,ngenes
         nnc = 0
         do 20 j=1,nc
            if (.not. risnan(xc(i,j))) then
              tmpxc(j) = xc(i,j)
              nnc = nnc + 1
            end if
 20      continue

         nnn = 0
         do 25 j=1,nn
            if (.not. risnan(xn(i,j))) then
              tmpxn(j) = xn(i,j)
              nnn = nnn + 1
            end if
 25      continue

!        use the nc and nn values for non-NA values (nnc, nnn)

!        find the mean and sample variance
         sumxc = sumd(tmpxc,nnc)
         meanxc = sumxc/nnc
         varxc = 0d0
         do 30 j=1,nnc
            varxc = varxc + (tmpxc(j) - meanxc)**2
 30      continue
         varxc = varxc/(nnc-1d0)

!        find the mean and sample variance
         sumxn = sumd(tmpxn,nnn)
         meanxn = sumxn/nnn
         varxn = 0d0
         do 40 j=1,nnn
            varxn = varxn + (tmpxn(j) - meanxn)**2
 40      continue
         varxn = varxn/(nnn-1d0)

!        calculate test statistic
         tmpt = (meanxc-meanxn)/sqrt((varxc/nnc)+(varxn/nnn))

!        calculate the degrees of freedom
         df = ((varxc/nnc)+(varxn/nnn))**2
     +           /((varxc/nnc)**2/(nnc-1d0)+(varxn/nnn)**2/(nnn-1d0))

!        calculate the p-value using call to c wrapper for R function pt()
         pvals(i) = 2d0*rpt(-abs(tmpt),df,1,0)
         tstats(i) = tmpt

 100   continue

      return
      end

# this was in wiFun.R - the wiFun function has been converted to Fortran
# so checkPara is in its own file

checkPara<-function(para, eps=1.0e-6)
{
  if(length(para) !=TNumParaRP)
  { stop("Number of parameters is not correct!\n") }

  # mixture proportions
  pi.1<-para[1]; pi.2<-para[2]; 
  pi.3<- 1-pi.1-pi.2

  if(pi.1>=1 || pi.1<=0)
  { cat("pi.1=", pi.1, " pi.2=", pi.2, " pi.3=",pi.3,"\n") 
    stop("pi.1 should be in (0, 1)!\n") }
  if(pi.2>=1 || pi.2<=0)
  { cat("pi.1=", pi.1, " pi.2=", pi.2, " pi.3=",pi.3,"\n") 
    stop("pi.2 should be in (0, 1)!\n") }
  if(pi.3>=1 || pi.3<=0)
  { cat("pi.1=", pi.1, " pi.2=", pi.2, " pi.3=",pi.3,"\n") 
    stop("pi.3 should be in (0, 1)!\n") }
  if(abs((pi.1+pi.2+pi.3)-1)>eps)
  { cat("pi.1=", pi.1, " pi.2=", pi.2, " pi.3=",pi.3,"\n") 
    stop("pi.1+pi.2+pi.3 should be equal to 1!\n") }

}

      subroutine llkhfun(llkh, x, psi_m, memsubjects, eps,
     +                  ngenes, n, nc, nn)

      implicit none


      double precision llkh
      double precision f1arr(ngenes), f2arr(ngenes), f3arr(ngenes)
      double precision x(ngenes,n), psi_m(17)
      double precision eps

      integer memsubjects(n), ngenes, n, nc, nn
      integer i, j, ci, ni

      character*85 errmess

      double precision pi_1, pi_2, pi_3

      double precision mu_c1, tau_c1, r_c1, rho_c1, delta_n1, mu_n1
      double precision tau_n1, r_n1, rho_n1, mu_2, tau_2, r_2, rho_2
      double precision mu_c3, tau_c3, r_c3, rho_c3, delta_n3, mu_n3
      double precision tau_n3, r_n3, rho_n3
      double precision sigma2_c1, sigma2_n1, sigma2_2
      double precision sigma2_c3, sigma2_n3

      double precision xci(nc), xni(nn), xi(n)
      double precision xci_sq(nc), xni_sq(nn), xi_sq(n)

      double precision xiTxi_c, xiT1_c, aiTai_c1, aiT12_c1
      double precision xiTxi_n, xiT1_n, aiTai_n1, aiT12_n1
      double precision part_c1, part_n1, delta1, log_detSigma1
      double precision xiTxi, xiT1, aiTai_2, aiT12_2
      double precision delta2, log_detSigma2

      double precision aiTai_c3, aiT12_c3, aiTai_n3, aiT12_n3
      double precision part_c3, part_n3, delta3, log_detSigma3


!      eps = 1.0d-6

!     check if parameters are in appropriate ranges
!     not checking dimension of psi_m though
!     this was previously handled by the R routine checkPara

!     mixture proportions
      pi_1 = psi_m(1)
      pi_2 = psi_m(2)
      pi_3 = 1d0 - pi_1 - pi_2

!     if parameters are out of the allowed range then exit with error message
      if((pi_1 .ge. 1d0) .or. (pi_1 .le. 0d0)) then
            write(errmess,'(A,D10.4,A,D10.4,A,D10.4,A)')
     +                        'pi.1=',pi_1,' pi.2=',pi_2,' pi.3=',pi_3,
     +                        '    pi.1 should be in (0, 1)          '
            call rexit(errmess)
      else if((pi_2 .ge. 1d0) .or. (pi_2 .le. 0d0)) then
            write(errmess,'(A,D10.4,A,D10.4,A,D10.4,A)')
     +                        'pi.1=',pi_1,' pi.2=',pi_2,' pi.3=',pi_3,
     +                        '    pi.2 should be in (0, 1)          '
            call rexit(errmess)
      else if((pi_3 .ge. 1d0) .or. (pi_3 .le. 0d0)) then
            write(errmess,'(A,D10.4,A,D10.4,A,D10.4,A)')
     +                        'pi.1=',pi_1,' pi.2=',pi_2,' pi.3=',pi_3,
     +                        '    pi.3 should be in (0, 1)          '
            call rexit(errmess)
      else if(abs((pi_1+pi_2+pi_3)-1d0) .gt. eps) then
            write(errmess,'(A,D10.4,A,D10.4,A,D10.4,A)')
     +                        'pi.1=',pi_1,' pi.2=',pi_2,' pi.3=',pi_3,
     +                        '   pi.1+pi.2+pi.3 should be equal to 1'
            call rexit(errmess)
      end if


!     mixture proportions calculated above

!     mean expression level for cluster 1 for diseased subjects
      mu_c1 = psi_m(3)
!     variance of expression levels for cluster 1 for diseased subjects
      tau_c1 = psi_m(4)
      sigma2_c1 = exp(tau_c1)
!     modified logit of correlation among expression levels for cluster 1 for diseased subjects
      r_c1 = psi_m(5)
      rho_c1 = (exp(r_c1)-1d0/(nc-1d0))/(1d0+exp(r_c1))

!     mu.n1=mu.c1-exp(delta.n1)
!      =mean expression level for cluster 1 for normal subjects
      delta_n1 = psi_m(6)
      mu_n1 = mu_c1-exp(delta_n1)
!     variance of expression levels for cluster 1 for normal subjects
      tau_n1 = psi_m(7)
      sigma2_n1 = exp(tau_n1)
!     modified logit of correlation among expression levels for cluster 1 for normal subjects
      r_n1 = psi_m(8)
      rho_n1 = (exp(r_n1)-1d0/(nn-1d0))/(1d0+exp(r_n1))

!     mean expression level for cluster 2
      mu_2 = psi_m(9)
!     variance of expression levels for cluster 2
      tau_2 = psi_m(10)
      sigma2_2 = exp(tau_2)
!     modified logit of correlation among expression levels for cluster 2
      r_2 = psi_m(11)
      rho_2 = (exp(r_2)-1d0/(n-1d0))/(1d0+exp(r_2))

!     mean expression level for cluster 3 for diseased subjects
      mu_c3 = psi_m(12)
!     variance of expression levels for cluster 3 for diseased subjects
      tau_c3 = psi_m(13)
      sigma2_c3 = exp(tau_c3)
!     modified logit of correlation among expression levels for cluster 3 for diseased subjects
      r_c3 = psi_m(14)
      rho_c3 = (exp(r_c3)-1d0/(nc-1d0))/(1d0+exp(r_c3))

!     mu.n3=mu.c3+exp(delta.n3)
!     =mean expression level for cluster 3 for normal subjects
      delta_n3 = psi_m(15)
      mu_n3 = mu_c3 + exp(delta_n3)
!     variance of expression levels for cluster 3 for normal subjects
      tau_n3 = psi_m(16)
      sigma2_n3 = exp(tau_n3)
!     modified logit of correlation among expression levels for cluster 3 for normal subjects
      r_n3 = psi_m(17)
      rho_n3 = (exp(r_n3)-1d0/(nn-1d0))/(1d0+exp(r_n3))


!     if values are out of the allowed range then exit with error message
      if(nc .ge. n) then
            errmess = 'Number of cases >= Total number of patients!'
            call rexit(errmess)
      else if(nn .ge. n) then
            errmess = 'Number of controls >= Total number of patients!'
            call rexit(errmess)
      end if



!     loop over ngenes to find each wi value
      do 100 j=1,ngenes

            ci = 1
            ni = 1
            do 20 i=1,n
                  xi(i) = x(j,i)
                  xi_sq(i) = x(j,i)**2
!                 expression levels of the gene for diseased subjects
                  if (memsubjects(i) .eq. 1) then
                        xci(ci) = x(j,i)
                        xci_sq(ci) = x(j,i)**2
                        ci = ci + 1
!                 expression levels of the gene for non-diseased subjects
                  else if (memsubjects(i) .eq. 0) then
                        xni(ni) = x(j,i)
                        xni_sq(ni) = x(j,i)**2
                        ni = ni + 1
                  end if
 20         continue



!           ###
!           # density for genes in cluster 1 (over-expressed)
!           ###

            xiTxi_c = sum(xci_sq)
            xiT1_c = sum(xci)
            aiTai_c1 = xiTxi_c-2d0*mu_c1*xiT1_c+nc*mu_c1**2
            aiT12_c1 = xiT1_c**2-2d0*nc*mu_c1*xiT1_c+nc**2*mu_c1**2

            xiTxi_n = sum(xni_sq)
            xiT1_n = sum(xni)
            aiTai_n1 = xiTxi_n-2d0*mu_n1*xiT1_n+nn*mu_n1**2
            aiT12_n1 = xiT1_n**2-2d0*nn*mu_n1*xiT1_n+nn**2*mu_n1**2

            part_c1 = (aiTai_c1-rho_c1*aiT12_c1/(1d0+(nc-1d0)*rho_c1))
     +                  / (sigma2_c1 * (1d0-rho_c1)) 
            part_n1 = (aiTai_n1-rho_n1*aiT12_n1/(1d0+(nn-1d0)*rho_n1))
     +                  / (sigma2_n1 * (1d0-rho_n1)) 
            delta1 = (part_c1+part_n1)

            log_detSigma1 = nc*log(sigma2_c1)+
     +            ((nc-1d0)*log(1d0-rho_c1)+log(1d0+(nc-1d0)*rho_c1))+
     +                  nn*log(sigma2_n1)+
     +            ((nn-1d0)*log(1d0-rho_n1)+log(1d0+(nn-1d0)*rho_n1))


!           ###
!           # density for genes in cluster 2 (non-expressed)
!           ###
            xiTxi = sum(xi_sq)
            xiT1 = sum(xi)
            aiTai_2 = xiTxi-2d0*mu_2*xiT1+n*mu_2**2
            aiT12_2 = xiT1**2-2d0*n*mu_2*xiT1+n**2*mu_2**2

            delta2 = ( aiTai_2 - rho_2* aiT12_2 / (1d0+(n-1d0)*rho_2) )
     +                  / (sigma2_2 * (1d0-rho_2)) 

            log_detSigma2 = n*log(sigma2_2)+
     +                  ((n-1d0)*log(1d0-rho_2)+log(1d0+(n-1d0)*rho_2))


!           ###
!           # density for genes in cluster 3 (under-expressed)
!           ###

            aiTai_c3 = xiTxi_c-2d0*mu_c3*xiT1_c+nc*mu_c3**2
            aiT12_c3 = xiT1_c**2-2d0*nc*mu_c3*xiT1_c+nc**2*mu_c3**2

            aiTai_n3 = xiTxi_n-2d0*mu_n3*xiT1_n+nn*mu_n3**2
            aiT12_n3 = xiT1_n**2-2d0*nn*mu_n3*xiT1_n+nn**2*mu_n3**2

            part_c3 = (aiTai_c3-rho_c3*aiT12_c3/(1d0+(nc-1d0)*rho_c3))
     +                  / (sigma2_c3 * (1d0-rho_c3)) 
            part_n3 = (aiTai_n3-rho_n3*aiT12_n3/(1d0+(nn-1d0)*rho_n3))
     +                  / (sigma2_n3 * (1d0-rho_n3)) 

            delta3 = (part_c3+part_n3)

            log_detSigma3 = nc*log(sigma2_c3)+
     +            (nc-1d0)*log(1d0-rho_c3)+log(1d0+(nc-1d0)*rho_c3)+
     +                  nn*log(sigma2_n3)+
     +            (nn-1d0)*log(1d0-rho_n3)+log(1d0+(nn-1d0)*rho_n3)



!           ################

            f1arr(j) = exp(-(log_detSigma1 + delta1)/2d0)
            f2arr(j) = exp(-(log_detSigma2 + delta2)/2d0)
            f3arr(j) = exp(-(log_detSigma3 + delta3)/2d0)


!     end loop over the genes (rows of x)
100   continue


      llkh = 0d0

      do 200 n=1,ngenes

            llkh = llkh + log(pi_1*f1arr(n)+pi_2*f2arr(n)+pi_3*f3arr(n))

200   continue



      return
      end

"getPara" <-
function(X, memSubjects, memGenes, eps=1.0e-6)
{
  Xupp<-X[memGenes==1,, drop=FALSE] 
  Xlow<-X[memGenes==3,, drop=FALSE] 
  Xnon<-X[memGenes==2,, drop=FALSE]

  nCases<-sum(memSubjects==1, na.rm=TRUE) # number of cases
  nControls<-sum(memSubjects==0, na.rm=TRUE) # number of controls
  nc<-nCases
  nn<-nControls
  nSubj<-nCases+nControls
  n<-nSubj
  
  n.upp<-nrow(Xupp)
  n.low<-nrow(Xlow)
  n.non<-nrow(Xnon)

  n<-nrow(X)
  memGenes2<-rep(1, n)
  memGenes2[memGenes==2]<-0

  pi.1<-n.upp/n
  pi.3<-n.low/n
  pi.2<-1-pi.1-pi.3

  Xuppc<-Xupp[,memSubjects==1, drop=FALSE]
  Xuppn<-Xupp[,memSubjects==0, drop=FALSE]

  Xlowc<-Xlow[,memSubjects==1, drop=FALSE]
  Xlown<-Xlow[,memSubjects==0, drop=FALSE]


  # within gene variation
  v.uppc.w<-mean(apply(Xuppc, 1, var, na.rm=TRUE), na.rm=TRUE)
  v.uppn.w<-mean(apply(Xuppn, 1, var, na.rm=TRUE), na.rm=TRUE)

  v.non.w<-mean(apply(Xnon, 1, var, na.rm=TRUE), na.rm=TRUE)

  v.lowc.w<-mean(apply(Xlowc, 1, var, na.rm=TRUE), na.rm=TRUE)
  v.lown.w<-mean(apply(Xlown, 1, var, na.rm=TRUE), na.rm=TRUE)

  # between gene variation
  v.uppc.b<-var(apply(Xuppc, 1, mean, na.rm=TRUE), na.rm=TRUE)
  v.uppn.b<-var(apply(Xuppn, 1, mean, na.rm=TRUE), na.rm=TRUE)

  v.non.b<-var(apply(Xnon, 1, mean, na.rm=TRUE), na.rm=TRUE)

  v.lowc.b<-var(apply(Xlowc, 1, mean, na.rm=TRUE), na.rm=TRUE)
  v.lown.b<-var(apply(Xlown, 1, mean, na.rm=TRUE), na.rm=TRUE)

  # marginal correlation
  rho.c1<-v.uppc.b/(v.uppc.b+v.uppc.w)
  rho.n1<-v.uppn.b/(v.uppn.b+v.uppn.w)
  rho.2<-v.non.b/(v.non.b+v.non.w)
  rho.c3<-v.lowc.b/(v.lowc.b+v.lowc.w)
  rho.n3<-v.lown.b/(v.lown.b+v.lown.w)

  r.c1<-log((1+(nCases-1)*rho.c1)/((1-rho.c1)*(nCases-1)))
  r.n1<-log((1+(nControls-1)*rho.n1)/((1-rho.n1)*(nControls-1)))
  r.2<-log((1+(nSubj-1)*rho.2)/((1-rho.2)*(nSubj-1)))
  r.c3<-log((1+(nCases-1)*rho.c3)/((1-rho.c3)*(nCases-1)))
  r.n3<-log((1+(nControls-1)*rho.n3)/((1-rho.n3)*(nControls-1)))

  mu.c1<-mean(apply(Xuppc, 1, mean, na.rm=TRUE), na.rm=TRUE)
  mu.n1<-mean(apply(Xuppn, 1, mean, na.rm=TRUE), na.rm=TRUE)
  sigma2.c1<-v.uppc.w
  sigma2.n1<-v.uppn.w

  mu.2<-mean(apply(Xnon, 1, mean, na.rm=TRUE), na.rm=TRUE)
  sigma2.2<-v.non.w

  mu.c3<-mean(apply(Xlowc, 1, mean, na.rm=TRUE), na.rm=TRUE)
  mu.n3<-mean(apply(Xlown, 1, mean, na.rm=TRUE), na.rm=TRUE)
  sigma2.c3<-v.lowc.w
  sigma2.n3<-v.lown.w

  if(mu.c1<=mu.n1 || mu.c3>=mu.n3 )
  { llkh<- -Inf }
  else { 
    delta.n1<-log(mu.c1-mu.n1)
    delta.n3<-log(mu.n3-mu.c3)
    paraIni<-c(pi.1, pi.2, 
               mu.c1, log(sigma2.c1), r.c1, 
               delta.n1, log(sigma2.n1), r.n1, 
               mu.2, log(sigma2.2), r.2, 
               mu.c3, log(sigma2.c3), r.c3, 
               delta.n3, log(sigma2.n3), r.n3)
    names(paraIni)<-paraNamesRP
 

#   call the wiFun Fortran routine below 
#    mat<-t(apply(X, 1, wiFun, Psi.m=paraIni, memSubjects=memSubjects, eps=eps))

# initialize the wi arrays
    wi1<-rep(0,nrow(X))
    wi2<-rep(0,nrow(X))
    wi3<-rep(0,nrow(X))

#   checkpara was called within wiFun.R, however this will be handled
#   in Fortran, so checking the dimension of Psi.m may not be included
#   in the future

# replace NAs with zeros and pass this matrix to wifun
    Xnna<-X
    Xnna[is.na(Xnna)] <- 0

    res <- .Fortran("wifun", wi1=as.double(wi1), wi2=as.double(wi2), wi3=as.double(wi3), 
            as.double(Xnna), as.double(paraIni),
            as.integer(memSubjects), as.double(eps),
            as.integer(nrow(X)), as.integer(ncol(X)),
            as.integer(sum(memSubjects==1, na.rm=TRUE)),
            as.integer(sum(memSubjects==0, na.rm=TRUE))     )

    mat<-cbind(res$wi1,res$wi2,res$wi3)


    w1<-as.numeric(mat[,1])
    w2<-as.numeric(mat[,2])
    w3<-as.numeric(mat[,3])
    xcMat<-X[,memSubjects==1,drop=FALSE]
    xnMat<-X[,memSubjects==0,drop=FALSE]
 
    xcTxc<-apply(xcMat, 1, function(x) {sum(x^2, na.rm=TRUE)})
    xcT1<-apply(xcMat, 1, function(x) {sum(x, na.rm=TRUE)})
 
    xnTxn<-apply(xnMat, 1, function(x) {sum(x^2, na.rm=TRUE)})
    xnT1<-apply(xnMat, 1, function(x) {sum(x, na.rm=TRUE)})
 
    xTx<-apply(X, 1, function(x) {sum(x^2, na.rm=TRUE)})
    xT1<-apply(X, 1, function(x) {sum(x, na.rm=TRUE)})

    sumw1<-mean(mat[,1], na.rm=TRUE)
    sumw2<-mean(mat[,2], na.rm=TRUE)
    sumw3<-mean(mat[,3], na.rm=TRUE)
 
    sumw1xcTxc<-sum(w1*xcTxc, na.rm=TRUE)
    sumw1xcT1<-sum(w1*xcT1, na.rm=TRUE)
    sumw1xcT1.sq<-sum(w1*xcT1^2, na.rm=TRUE)
  
    sumw1xnTxn<-sum(w1*xnTxn, na.rm=TRUE)
    sumw1xnT1<-sum(w1*xnT1, na.rm=TRUE)
    sumw1xnT1.sq<-sum(w1*xnT1^2, na.rm=TRUE)
  
    sumw2xTx<-sum(w2*xTx, na.rm=TRUE)
    sumw2xT1<-sum(w2*xT1, na.rm=TRUE)
    sumw2xT1.sq<-sum(w2*xT1^2, na.rm=TRUE)
  
    sumw3xcTxc<-sum(w3*xcTxc, na.rm=TRUE)
    sumw3xcT1<-sum(w3*xcT1, na.rm=TRUE)
    sumw3xcT1.sq<-sum(w3*xcT1^2, na.rm=TRUE)
  
    sumw3xnTxn<-sum(w3*xnTxn, na.rm=TRUE)
    sumw3xnT1<-sum(w3*xnT1, na.rm=TRUE)
    sumw3xnT1.sq<-sum(w3*xnT1^2, na.rm=TRUE)
  

# call the Fortran routine llkh
  llkh <- 0

  res <- .Fortran("llkhfun", llkh=as.double(llkh), 
            as.double(Xnna), as.double(paraIni),
            as.integer(memSubjects), as.double(eps),
            as.integer(nrow(X)), as.integer(ncol(X)),
            as.integer(sum(memSubjects==1, na.rm=TRUE)),
            as.integer(sum(memSubjects==0, na.rm=TRUE))     )

  llkh <- res$llkh
  }

  return(list(para=paraIni, llkh=llkh))
}

### R code from vignette source 'gsMMD.Rnw'

###################################################
### code chunk number 1: gsMMD.Rnw:494-510
###################################################
  library(GeneSelectMMD)
  library(ALL)
  data(ALL)
  eSet1 <- ALL[1:100, ALL$BT == "B3" | ALL$BT == "T2"]
  
  mem.str <- as.character(eSet1$BT)
  nSubjects <- length(mem.str)
  memSubjects <- rep(0,nSubjects)
  # B3 coded as 0, T2 coded as 1
  memSubjects[mem.str == "T2"] <- 1
  
  obj.gsMMD <- gsMMD(eSet1, memSubjects, transformFlag = TRUE, 
    transformMethod = "boxcox", scaleFlag = TRUE, quiet = TRUE)
  para <- obj.gsMMD$para
  print(round(para, 3))



###################################################
### code chunk number 2: gsMMD.Rnw:513-531 (eval = FALSE)
###################################################
##   library(GeneSelectMMD)
##   library(ALL)
##   data(ALL)
##   eSet1 <- ALL[1:100, ALL$BT == "B3" | ALL$BT == "T2"]
##   mat <- exprs(eSet1)
##   
##   mem.str <- as.character(eSet1$BT)
##   nSubjects <- length(mem.str)
##   memSubjects <- rep(0,nSubjects)
##   # B3 coded as 0, T2 coded as 1
##   memSubjects[mem.str == "T2"] <- 1
##  
##   obj.gsMMD <- gsMMD.default(mat, memSubjects, iniGeneMethod = "Ttest",
##           transformFlag = TRUE, transformMethod = "boxcox", scaleFlag = TRUE,
##           quiet=TRUE)
##   para <- obj.gsMMD$para
##   print(round(para, 3))
## 


###################################################
### code chunk number 3: gsMMD.Rnw:537-576 (eval = FALSE)
###################################################
##   library(GeneSelectMMD)
##   library(ALL)
##   data(ALL)
##   eSet1 <- ALL[1:100, ALL$BT == "B3" | ALL$BT == "T2"]
##   
##   mem.str <- as.character(eSet1$BT)
##   nSubjects <- length(mem.str)
##   memSubjects <- rep(0,nSubjects)
##   # B3 coded as 0, T2 coded as 1
##   memSubjects[mem.str == "T2"] <- 1
##   
##   myWilcox <-
##   function(x, memSubjects, alpha = 0.05)
##   {
##     xc <- x[memSubjects == 1]
##     xn <- x[memSubjects == 0]
##   
##     m <- sum(memSubjects == 1)
##     res <- wilcox.test(x = xc, y = xn, conf.level = 1 - alpha)
##     res2 <- c(res$p.value, res$statistic - m * (m + 1) / 2)
##     names(res2) <- c("p.value", "statistic")
##   
##     return(res2)
##   }
##   
##   mat <- exprs(eSet1)
##   tmp <- t(apply(mat, 1, myWilcox, memSubjects = memSubjects))
##   colnames(tmp) <- c("p.value", "statistic")
##   memIni <- rep(2, nrow(mat))
##   memIni[tmp[, 1] < 0.05 & tmp[, 2] > 0] <- 1
##   memIni[tmp[, 1] < 0.05 & tmp[, 2] < 0] <- 3
##   
##   print(table(memIni))
## 
##   obj.gsMMD <- gsMMD2(eSet1, memSubjects, memIni, transformFlag = TRUE, 
##        transformMethod = "boxcox", scaleFlag = TRUE, quiet = TRUE)
##   para <- obj.gsMMD$para
##   print(round(para, 3))
## 


###################################################
### code chunk number 4: gsMMD.Rnw:579-620 (eval = FALSE)
###################################################
##   library(GeneSelectMMD)
##   library(ALL)
##   data(ALL)
##   eSet1 <- ALL[1:100, ALL$BT == "B3" | ALL$BT == "T2"]
##   mat <- exprs(eSet1)
##   
##   mem.str <- as.character(eSet1$BT)
##   nSubjects <- length(mem.str)
##   memSubjects <- rep(0,nSubjects)
##   # B3 coded as 0, T2 coded as 1
##   memSubjects[mem.str == "T2"] <- 1
##  
##   myWilcox <-
##   function(x, memSubjects, alpha = 0.05)
##   {
##     xc <- x[memSubjects == 1]
##     xn <- x[memSubjects == 0]
##   
##     m <- sum(memSubjects == 1)
##     res <- wilcox.test(x = xc, y = xn, conf.level = 1 - alpha)
##     res2 <- c(res$p.value, res$statistic - m * (m + 1) / 2)
##     names(res2) <- c("p.value", "statistic")
##   
##     return(res2)
##   }
##   
##   tmp <- t(apply(mat, 1, myWilcox, memSubjects = memSubjects))
##   colnames(tmp) <- c("p.value", "statistic")
##   memIni <- rep(2, nrow(mat))
##   memIni[tmp[, 1] < 0.05 & tmp[, 2] > 0] <- 1
##   memIni[tmp[, 1] < 0.05 & tmp[, 2] < 0] <- 3
##   
##   print(table(memIni))
##   
##   obj.gsMMD <- gsMMD2.default(mat, memSubjects, memIni = memIni,
##           transformFlag = TRUE, transformMethod = "boxcox", scaleFlag = TRUE,
##           quiet=TRUE)
##   para <- obj.gsMMD$para
##   print(round(para, 3))
## 
## 


###################################################
### code chunk number 5: gsMMD.Rnw:659-661
###################################################
  print(round(errRates(obj.gsMMD), 3))



###################################################
### code chunk number 6: gsMMD.Rnw:677-686
###################################################
  plotHistDensity(obj.gsMMD, plotFlag = "case", 
      mytitle = "Histogram of gene expression levels for T2\nimposed with estimated density (case)", 
      plotComponent = TRUE, 
      x.legend = c(0.8, 3), 
      y.legend = c(0.3, 0.4), 
      numPoints = 500,
      cex.main = 1,
      cex = 1)

const.b1<-2
const.b2<-2
const.b3<-2
const.b.vec<-c(const.b1, const.b2, const.b3)


# para=(pi1, pi2, 
#       muc1, tau.c1, r.c1,
#       delta.n1, tau.n1, r.n1,
#       mu.2, tau.2, r.2,
#       mu.c3, tau.c3, r.c3,
#       delta.n3, tau.n3, r.n3)
# total  17 parameters
 
"paraEst" <-
function(X, paraIni, memSubjects, maxFlag=TRUE, thrshPostProb=0.50, 
         geneNames=NULL, ITMAX=100, eps=1.0e-03, quiet=TRUE)
{
  # check if parameters are in appropriate ranges
  checkPara(paraIni, eps)

  sumb<-sum(const.b.vec, na.rm=TRUE)
  const.b1.b2<-c(const.b1, const.b2)

  n<-ncol(X) # number of patients/subjects
  nc<-sum(memSubjects==1, na.rm=TRUE)  # number of cases
  nn<-sum(memSubjects==0, na.rm=TRUE)  # number of controls
  if(nc>=n){
    stop("Number of cases >= Total number of patients!\n")
  }
  if(nn>=n){
    stop("Number of controls >= Total number of patients!\n")
  }

  # start em algorithm
  Psi.m<-paraIni
  #cat("Psi.m>>\n"); print(Psi.m); cat("\n");
  nGenes<-nrow(X)

  ##############################

  xcMat<-X[,memSubjects==1,drop=FALSE]
  xnMat<-X[,memSubjects==0,drop=FALSE]

  xcTxc<-apply(xcMat, 1, function(x) {sum(x^2, na.rm=TRUE)})
  xcT1<-apply(xcMat, 1, function(x) {sum(x, na.rm=TRUE)})

  xnTxn<-apply(xnMat, 1, function(x) {sum(x^2, na.rm=TRUE)})
  xnT1<-apply(xnMat, 1, function(x) {sum(x, na.rm=TRUE)})

  xTx<-apply(X, 1, function(x) {sum(x^2, na.rm=TRUE)})
  xT1<-apply(X, 1, function(x) {sum(x, na.rm=TRUE)})


# initialize the wi arrays
  wi1<-rep(0,nrow(X))
  wi2<-rep(0,nrow(X))
  wi3<-rep(0,nrow(X))

  loop<-0

# call the paraestloop subroutine in paraEstLoop.f
# paraEstLoop.f includes the loop previously found in paraEst.R

# replace NAs with zeros and pass this matrix to paraestloop
  Xnna<-X
  Xnna[is.na(Xnna)] <- 0

  res <- .Fortran("paraestloop", w1=as.double(wi1), w2=as.double(wi2), w3=as.double(wi3),
          as.double(Xnna), Psi.m=as.double(Psi.m),
          as.integer(memSubjects),
          as.double(xcTxc), as.double(xcT1),
          as.double(xnTxn), as.double(xnT1),
          as.double(xTx), as.double(xT1),
          as.integer(nrow(X)),
          as.integer(ncol(X)),
          as.integer(sum(memSubjects==1, na.rm=TRUE)),
          as.integer(sum(memSubjects==0, na.rm=TRUE)),
          as.integer(ITMAX), as.double(eps),
          as.logical(quiet), loop=as.integer(loop)     )


# extract the parameters, w's and the number of loops required
  Psi.m <- res$Psi.m
  w1<-res$w1
  w2<-res$w2
  w3<-res$w3
  loop<-res$loop
  
  if(!quiet)
  { cat("Total iterations for EM algorithm=", loop, "\n") }


# gene cluster membership
#  memMat was changed to wiMat when Fortran code was included
#  memMat<-t(apply(X, 1, wiFun, Psi.m=Psi.m, memSubjects=memSubjects, eps=eps))

#   probably don't need to create these again because it is created at the start
  wi1<-rep(0,nrow(X))
  wi2<-rep(0,nrow(X))
  wi3<-rep(0,nrow(X))

# checkpara was called within wiFun.R, however this will be handled
# in Fortran, so checking the dimension of Psi.m will not be included
# in the future
#
# eps=1.0e-6 for wiFun

# replace NAs with zeros and pass this matrix to wifun
  Xnna<-X
  Xnna[is.na(Xnna)] <- 0

  res <- .Fortran("wifun", wi1=as.double(wi1), wi2=as.double(wi2), wi3=as.double(wi3), 
            as.double(Xnna), as.double(Psi.m),
            as.integer(memSubjects), as.double(1.0e-6),
            as.integer(nrow(X)), as.integer(ncol(X)),
            as.integer(sum(memSubjects==1, na.rm=TRUE)),
            as.integer(sum(memSubjects==0, na.rm=TRUE))     )


  wiMat<-cbind(res$wi1,res$wi2,res$wi3)

  colnames(wiMat)<-c("cluster1", "cluster2", "cluster3")
  rownames(wiMat)<-geneNames

#  update gene cluster membership
#  using Fortran routine maxposfun (converted maxPosFun in R)
#   memGenes<-apply(memMat, 1, maxPosFun, maxFlag=maxFlag, thrshPostProb=thrshPostProb)
#   memMat was changed to wiMat when Fortran code was included

# initialize memGenes array
  memGenes<-rep(0,as.integer(nrow(X)))

  res <- .Fortran("maxposfun", memGenes=as.integer(memGenes), as.double(wiMat), 
                  as.integer(nrow(X)), as.logical(maxFlag), as.double(thrshPostProb)  )

  memGenes<-res$memGenes


  sumw1<-mean(wiMat[,1], na.rm=TRUE)
  sumw2<-mean(wiMat[,2], na.rm=TRUE)
  sumw3<-mean(wiMat[,3], na.rm=TRUE)

  w1<-as.numeric(wiMat[,1])
  w2<-as.numeric(wiMat[,2])
  w3<-as.numeric(wiMat[,3])

  sumw1xcTxc<-sum(w1*xcTxc, na.rm=TRUE)
  sumw1xcT1<-sum(w1*xcT1, na.rm=TRUE)
  sumw1xcT1.sq<-sum(w1*xcT1^2, na.rm=TRUE)
 
  sumw1xnTxn<-sum(w1*xnTxn, na.rm=TRUE)
  sumw1xnT1<-sum(w1*xnT1, na.rm=TRUE)
  sumw1xnT1.sq<-sum(w1*xnT1^2, na.rm=TRUE)
 
  sumw2xTx<-sum(w2*xTx, na.rm=TRUE)
  sumw2xT1<-sum(w2*xT1, na.rm=TRUE)
  sumw2xT1.sq<-sum(w2*xT1^2, na.rm=TRUE)
 
  sumw3xcTxc<-sum(w3*xcTxc, na.rm=TRUE)
  sumw3xcT1<-sum(w3*xcT1, na.rm=TRUE)
  sumw3xcT1.sq<-sum(w3*xcT1^2, na.rm=TRUE)
 
  sumw3xnTxn<-sum(w3*xnTxn, na.rm=TRUE)
  sumw3xnT1<-sum(w3*xnT1, na.rm=TRUE)
  sumw3xnT1.sq<-sum(w3*xnT1^2, na.rm=TRUE)
 

# call the Fortran subroutine llkh
  llkh <- 0

  res <- .Fortran("llkhfun", llkh=as.double(llkh), 
            as.double(Xnna), as.double(Psi.m),
            as.integer(memSubjects), as.double(eps),
            as.integer(nrow(X)), as.integer(n),
            as.integer(nc),
            as.integer(nn)     )

  llkh <- res$llkh


  memGenes2<-rep(1, nGenes)
  memGenes2[memGenes==2]<-0

  if(sum(is.null(geneNames), na.rm=TRUE))
  {
    geneNames<-paste("gene", 1:nGenes, sep="")
  } 

  names(Psi.m)<-paraNamesRP
  names(memGenes)<-geneNames
  names(memGenes2)<-geneNames

  invisible(list(para=Psi.m, llkh=llkh, memGenes=memGenes, 
                 memGenes2=memGenes2, wiMat=wiMat, loop=loop))
}

c===========   L-BFGS-B (version 3.0.  April 25, 2011  ===================
c
c     This is a modified version of L-BFGS-B. Minor changes in the updated 
c     code appear preceded by a line comment as follows 
c  
c     c-jlm-jn 
c
c     Major changes are described in the accompanying paper:
c
c         Jorge Nocedal and Jose Luis Morales, Remark on "Algorithm 778: 
c         L-BFGS-B: Fortran Subroutines for Large-Scale Bound Constrained 
c         Optimization"  (2011). To appear in  ACM Transactions on 
c         Mathematical Software,
c
c     The paper describes an improvement and a correction to Algorithm 778. 
c     It is shown that the performance of the algorithm can be improved 
c     significantly by making a relatively simple modication to the subspace 
c     minimization phase. The correction concerns an error caused by the use 
c     of routine dpmeps to estimate machine precision. 
c
c     The total work space **wa** required by the new version is 
c 
c                  2*m*n + 11m*m + 5*n + 8*m 
c
c     the old version required 
c
c                  2*m*n + 12m*m + 4*n + 12*m 
c
c
c            J. Nocedal  Department of Electrical Engineering and
c                        Computer Science.
c                        Northwestern University. Evanston, IL. USA
c
c
c           J.L Morales  Departamento de Matematicas, 
c                        Instituto Tecnologico Autonomo de Mexico
c                        Mexico D.F. Mexico.
c
c                        March  2011    
c                                                 
c============================================================================= 
      subroutine setulb(n, m, x, l, u, nbd, f, g, factr, pgtol, wa, iwa,
     +                 task, iprint, csave, lsave, isave, dsave)
 
      character*60     task, csave
      logical          lsave(4)
      integer          n, m, iprint, 
     +                 nbd(n), iwa(3*n), isave(44)
      double precision f, factr, pgtol, x(n), l(n), u(n), g(n),
c
c-jlm-jn
     +                 wa(2*m*n + 5*n + 11*m*m + 8*m), dsave(29)
 
c     ************
c
c     Subroutine setulb
c
c     This subroutine partitions the working arrays wa and iwa, and 
c       then uses the limited memory BFGS method to solve the bound
c       constrained optimization problem by calling mainlb.
c       (The direct method will be used in the subspace minimization.)
c
c     n is an integer variable.
c       On entry n is the dimension of the problem.
c       On exit n is unchanged.
c
c     m is an integer variable.
c       On entry m is the maximum number of variable metric corrections
c         used to define the limited memory matrix.
c       On exit m is unchanged.
c
c     x is a double precision array of dimension n.
c       On entry x is an approximation to the solution.
c       On exit x is the current approximation.
c
c     l is a double precision array of dimension n.
c       On entry l is the lower bound on x.
c       On exit l is unchanged.
c
c     u is a double precision array of dimension n.
c       On entry u is the upper bound on x.
c       On exit u is unchanged.
c
c     nbd is an integer array of dimension n.
c       On entry nbd represents the type of bounds imposed on the
c         variables, and must be specified as follows:
c         nbd(i)=0 if x(i) is unbounded,
c                1 if x(i) has only a lower bound,
c                2 if x(i) has both lower and upper bounds, and
c                3 if x(i) has only an upper bound.
c       On exit nbd is unchanged.
c
c     f is a double precision variable.
c       On first entry f is unspecified.
c       On final exit f is the value of the function at x.
c
c     g is a double precision array of dimension n.
c       On first entry g is unspecified.
c       On final exit g is the value of the gradient at x.
c
c     factr is a double precision variable.
c       On entry factr >= 0 is specified by the user.  The iteration
c         will stop when
c
c         (f^k - f^{k+1})/max{|f^k|,|f^{k+1}|,1} <= factr*epsmch
c
c         where epsmch is the machine precision, which is automatically
c         generated by the code. Typical values for factr: 1.d+12 for
c         low accuracy; 1.d+7 for moderate accuracy; 1.d+1 for extremely
c         high accuracy.
c       On exit factr is unchanged.
c
c     pgtol is a double precision variable.
c       On entry pgtol >= 0 is specified by the user.  The iteration
c         will stop when
c
c                 max{|proj g_i | i = 1, ..., n} <= pgtol
c
c         where pg_i is the ith component of the projected gradient.   
c       On exit pgtol is unchanged.
c
c     wa is a double precision working array of length 
c       (2mmax + 5)nmax + 12mmax^2 + 12mmax.
c
c     iwa is an integer working array of length 3nmax.
c
c     task is a working string of characters of length 60 indicating
c       the current job when entering and quitting this subroutine.
c
c     iprint is an integer variable that must be set by the user.
c       It controls the frequency and type of output generated:
c        iprint<0    no output is generated;
c        iprint=0    print only one line at the last iteration;
c        0<iprint<99 print also f and |proj g| every iprint iterations;
c        iprint=99   print details of every iteration except n-vectors;
c        iprint=100  print also the changes of active set and final x;
c        iprint>100  print details of every iteration including x and g;
c       When iprint > 0, the file iterate.dat will be created to
c                        summarize the iteration.
c
c     csave is a working string of characters of length 60.
c
c     lsave is a logical working array of dimension 4.
c       On exit with 'task' = NEW_X, the following information is 
c                                                             available:
c         If lsave(1) = .true.  then  the initial X has been replaced by
c                                     its projection in the feasible set;
c         If lsave(2) = .true.  then  the problem is constrained;
c         If lsave(3) = .true.  then  each variable has upper and lower
c                                     bounds;
c
c     isave is an integer working array of dimension 44.
c       On exit with 'task' = NEW_X, the following information is 
c                                                             available:
c         isave(22) = the total number of intervals explored in the 
c                         search of Cauchy points;
c         isave(26) = the total number of skipped BFGS updates before 
c                         the current iteration;
c         isave(30) = the number of current iteration;
c         isave(31) = the total number of BFGS updates prior the current
c                         iteration;
c         isave(33) = the number of intervals explored in the search of
c                         Cauchy point in the current iteration;
c         isave(34) = the total number of function and gradient 
c                         evaluations;
c         isave(36) = the number of function value or gradient
c                                  evaluations in the current iteration;
c         if isave(37) = 0  then the subspace argmin is within the box;
c         if isave(37) = 1  then the subspace argmin is beyond the box;
c         isave(38) = the number of free variables in the current
c                         iteration;
c         isave(39) = the number of active constraints in the current
c                         iteration;
c         n + 1 - isave(40) = the number of variables leaving the set of
c                           active constraints in the current iteration;
c         isave(41) = the number of variables entering the set of active
c                         constraints in the current iteration.
c
c     dsave is a double precision working array of dimension 29.
c       On exit with 'task' = NEW_X, the following information is
c                                                             available:
c         dsave(1) = current 'theta' in the BFGS matrix;
c         dsave(2) = f(x) in the previous iteration;
c         dsave(3) = factr*epsmch;
c         dsave(4) = 2-norm of the line search direction vector;
c         dsave(5) = the machine precision epsmch generated by the code;
c         dsave(7) = the accumulated time spent on searching for
c                                                         Cauchy points;
c         dsave(8) = the accumulated time spent on
c                                                 subspace minimization;
c         dsave(9) = the accumulated time spent on line search;
c         dsave(11) = the slope of the line search function at
c                                  the current point of line search;
c         dsave(12) = the maximum relative step length imposed in
c                                                           line search;
c         dsave(13) = the infinity norm of the projected gradient;
c         dsave(14) = the relative step length in the line search;
c         dsave(15) = the slope of the line search function at
c                                 the starting point of the line search;
c         dsave(16) = the square of the 2-norm of the line search
c                                                      direction vector.
c
c     Subprograms called:
c
c       L-BFGS-B Library ... mainlb.    
c
c
c     References:
c
c       [1] R. H. Byrd, P. Lu, J. Nocedal and C. Zhu, ``A limited
c       memory algorithm for bound constrained optimization'',
c       SIAM J. Scientific Computing 16 (1995), no. 5, pp. 1190--1208.
c
c       [2] C. Zhu, R.H. Byrd, P. Lu, J. Nocedal, ``L-BFGS-B: a
c       limited memory FORTRAN code for solving bound constrained
c       optimization problems'', Tech. Report, NAM-11, EECS Department,
c       Northwestern University, 1994.
c
c       (Postscript files of these papers are available via anonymous
c        ftp to eecs.nwu.edu in the directory pub/lbfgs/lbfgs_bcm.)
c
c                           *  *  *
c
c     NEOS, November 1994. (Latest revision June 1996.)
c     Optimization Technology Center.
c     Argonne National Laboratory and Northwestern University.
c     Written by
c                        Ciyou Zhu
c     in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
c
c
c     ************
c-jlm-jn 
      integer   lws,lr,lz,lt,ld,lxp,lwa,
     +          lwy,lsy,lss,lwt,lwn,lsnd

      if (task .eq. 'START') then
         isave(1)  = m*n
         isave(2)  = m**2
         isave(3)  = 4*m**2
         isave(4)  = 1                      ! ws      m*n
         isave(5)  = isave(4)  + isave(1)   ! wy      m*n
         isave(6)  = isave(5)  + isave(1)   ! wsy     m**2
         isave(7)  = isave(6)  + isave(2)   ! wss     m**2
         isave(8)  = isave(7)  + isave(2)   ! wt      m**2
         isave(9)  = isave(8)  + isave(2)   ! wn      4*m**2
         isave(10) = isave(9)  + isave(3)   ! wsnd    4*m**2
         isave(11) = isave(10) + isave(3)   ! wz      n
         isave(12) = isave(11) + n          ! wr      n
         isave(13) = isave(12) + n          ! wd      n
         isave(14) = isave(13) + n          ! wt      n
         isave(15) = isave(14) + n          ! wxp     n
         isave(16) = isave(15) + n          ! wa      8*m
      endif
      lws  = isave(4)
      lwy  = isave(5)
      lsy  = isave(6)
      lss  = isave(7)
      lwt  = isave(8)
      lwn  = isave(9)
      lsnd = isave(10)
      lz   = isave(11)
      lr   = isave(12)
      ld   = isave(13)
      lt   = isave(14)
      lxp  = isave(15)
      lwa  = isave(16)

      call mainlb(n,m,x,l,u,nbd,f,g,factr,pgtol,
     +  wa(lws),wa(lwy),wa(lsy),wa(lss), wa(lwt),
     +  wa(lwn),wa(lsnd),wa(lz),wa(lr),wa(ld),wa(lt),wa(lxp),
     +  wa(lwa),
     +  iwa(1),iwa(n+1),iwa(2*n+1),task,iprint, 
     +  csave,lsave,isave(22),dsave)

      return

      end

c======================= The end of setulb =============================
 
      subroutine mainlb(n, m, x, l, u, nbd, f, g, factr, pgtol, ws, wy,
     +                  sy, ss, wt, wn, snd, z, r, d, t, xp, wa, 
     +                  index, iwhere, indx2, task,
     +                  iprint, csave, lsave, isave, dsave)
      implicit none
      character*60     task, csave
      logical          lsave(4)
      integer          n, m, iprint, nbd(n), index(n),
     +                 iwhere(n), indx2(n), isave(23)
      double precision f, factr, pgtol,
     +                 x(n), l(n), u(n), g(n), z(n), r(n), d(n), t(n), 
c-jlm-jn
     +                 xp(n), 
     +                 wa(8*m), 
     +                 ws(n, m), wy(n, m), sy(m, m), ss(m, m), 
     +                 wt(m, m), wn(2*m, 2*m), snd(2*m, 2*m), dsave(29)

c     ************
c
c     Subroutine mainlb
c
c     This subroutine solves bound constrained optimization problems by
c       using the compact formula of the limited memory BFGS updates.
c       
c     n is an integer variable.
c       On entry n is the number of variables.
c       On exit n is unchanged.
c
c     m is an integer variable.
c       On entry m is the maximum number of variable metric
c          corrections allowed in the limited memory matrix.
c       On exit m is unchanged.
c
c     x is a double precision array of dimension n.
c       On entry x is an approximation to the solution.
c       On exit x is the current approximation.
c
c     l is a double precision array of dimension n.
c       On entry l is the lower bound of x.
c       On exit l is unchanged.
c
c     u is a double precision array of dimension n.
c       On entry u is the upper bound of x.
c       On exit u is unchanged.
c
c     nbd is an integer array of dimension n.
c       On entry nbd represents the type of bounds imposed on the
c         variables, and must be specified as follows:
c         nbd(i)=0 if x(i) is unbounded,
c                1 if x(i) has only a lower bound,
c                2 if x(i) has both lower and upper bounds,
c                3 if x(i) has only an upper bound.
c       On exit nbd is unchanged.
c
c     f is a double precision variable.
c       On first entry f is unspecified.
c       On final exit f is the value of the function at x.
c
c     g is a double precision array of dimension n.
c       On first entry g is unspecified.
c       On final exit g is the value of the gradient at x.
c
c     factr is a double precision variable.
c       On entry factr >= 0 is specified by the user.  The iteration
c         will stop when
c
c         (f^k - f^{k+1})/max{|f^k|,|f^{k+1}|,1} <= factr*epsmch
c
c         where epsmch is the machine precision, which is automatically
c         generated by the code.
c       On exit factr is unchanged.
c
c     pgtol is a double precision variable.
c       On entry pgtol >= 0 is specified by the user.  The iteration
c         will stop when
c
c                 max{|proj g_i | i = 1, ..., n} <= pgtol
c
c         where pg_i is the ith component of the projected gradient.
c       On exit pgtol is unchanged.
c
c     ws, wy, sy, and wt are double precision working arrays used to
c       store the following information defining the limited memory
c          BFGS matrix:
c          ws, of dimension n x m, stores S, the matrix of s-vectors;
c          wy, of dimension n x m, stores Y, the matrix of y-vectors;
c          sy, of dimension m x m, stores S'Y;
c          ss, of dimension m x m, stores S'S;
c          yy, of dimension m x m, stores Y'Y;
c          wt, of dimension m x m, stores the Cholesky factorization
c                                  of (theta*S'S+LD^(-1)L'); see eq.
c                                  (2.26) in [3].
c
c     wn is a double precision working array of dimension 2m x 2m
c       used to store the LEL^T factorization of the indefinite matrix
c                 K = [-D -Y'ZZ'Y/theta     L_a'-R_z'  ]
c                     [L_a -R_z           theta*S'AA'S ]
c
c       where     E = [-I  0]
c                     [ 0  I]
c
c     snd is a double precision working array of dimension 2m x 2m
c       used to store the lower triangular part of
c                 N = [Y' ZZ'Y   L_a'+R_z']
c                     [L_a +R_z  S'AA'S   ]
c            
c     z(n),r(n),d(n),t(n), xp(n),wa(8*m) are double precision working arrays.
c       z  is used at different times to store the Cauchy point and
c          the Newton point.
c       xp is used to safeguard the projected Newton direction
c
c     sg(m),sgo(m),yg(m),ygo(m) are double precision working arrays. 
c
c     index is an integer working array of dimension n.
c       In subroutine freev, index is used to store the free and fixed
c          variables at the Generalized Cauchy Point (GCP).
c
c     iwhere is an integer working array of dimension n used to record
c       the status of the vector x for GCP computation.
c       iwhere(i)=0 or -3 if x(i) is free and has bounds,
c                 1       if x(i) is fixed at l(i), and l(i) .ne. u(i)
c                 2       if x(i) is fixed at u(i), and u(i) .ne. l(i)
c                 3       if x(i) is always fixed, i.e.,  u(i)=x(i)=l(i)
c                -1       if x(i) is always free, i.e., no bounds on it.
c
c     indx2 is an integer working array of dimension n.
c       Within subroutine cauchy, indx2 corresponds to the array iorder.
c       In subroutine freev, a list of variables entering and leaving
c       the free set is stored in indx2, and it is passed on to
c       subroutine formk with this information.
c
c     task is a working string of characters of length 60 indicating
c       the current job when entering and leaving this subroutine.
c
c     iprint is an INTEGER variable that must be set by the user.
c       It controls the frequency and type of output generated:
c        iprint<0    no output is generated;
c        iprint=0    print only one line at the last iteration;
c        0<iprint<99 print also f and |proj g| every iprint iterations;
c        iprint=99   print details of every iteration except n-vectors;
c        iprint=100  print also the changes of active set and final x;
c        iprint>100  print details of every iteration including x and g;
c       When iprint > 0, the file iterate.dat will be created to
c                        summarize the iteration.
c
c     csave is a working string of characters of length 60.
c
c     lsave is a logical working array of dimension 4.
c
c     isave is an integer working array of dimension 23.
c
c     dsave is a double precision working array of dimension 29.
c
c
c     Subprograms called
c
c       L-BFGS-B Library ... cauchy, subsm, lnsrlb, formk, 
c
c        errclb, prn1lb, prn2lb, prn3lb, active, projgr,
c
c        freev, cmprlb, matupd, formt.
c
c       Minpack2 Library ... timer
c
c       Linpack Library ... dcopy, ddot.
c
c
c     References:
c
c       [1] R. H. Byrd, P. Lu, J. Nocedal and C. Zhu, ``A limited
c       memory algorithm for bound constrained optimization'',
c       SIAM J. Scientific Computing 16 (1995), no. 5, pp. 1190--1208.
c
c       [2] C. Zhu, R.H. Byrd, P. Lu, J. Nocedal, ``L-BFGS-B: FORTRAN
c       Subroutines for Large Scale Bound Constrained Optimization''
c       Tech. Report, NAM-11, EECS Department, Northwestern University,
c       1994.
c 
c       [3] R. Byrd, J. Nocedal and R. Schnabel "Representations of
c       Quasi-Newton Matrices and their use in Limited Memory Methods'',
c       Mathematical Programming 63 (1994), no. 4, pp. 129-156.
c
c       (Postscript files of these papers are available via anonymous
c        ftp to eecs.nwu.edu in the directory pub/lbfgs/lbfgs_bcm.)
c
c                           *  *  *
c
c     NEOS, November 1994. (Latest revision June 1996.)
c     Optimization Technology Center.
c     Argonne National Laboratory and Northwestern University.
c     Written by
c                        Ciyou Zhu
c     in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
c
c
c     ************
 
      logical          prjctd,cnstnd,boxed,updatd,wrk
      character*3      word
      integer          i,k,nintol,itfile,iback,nskip,
     +                 head,col,iter,itail,iupdat,
     +                 nseg,nfgv,info,ifun,
     +                 iword,nfree,nact,ileave,nenter
      double precision theta,fold,ddot,dr,rr,tol,
     +                 xstep,sbgnrm,ddum,dnorm,dtd,epsmch,
     +                 cpu1,cpu2,cachyt,sbtime,lnscht,time1,time2,
     +                 gd,gdold,stp,stpmx,time
      double precision one,zero
      parameter        (one=1.0d0,zero=0.0d0)
      
      if (task .eq. 'START') then

         epsmch = epsilon(one)

         call timer(time1)

c        Initialize counters and scalars when task='START'.

c           for the limited memory BFGS matrices:
         col    = 0
         head   = 1
         theta  = one
         iupdat = 0
         updatd = .false.
         iback  = 0
         itail  = 0
         iword  = 0
         nact   = 0
         ileave = 0
         nenter = 0
         fold   = zero
         dnorm  = zero
         cpu1   = zero
         gd     = zero
         stpmx  = zero
         sbgnrm = zero
         stp    = zero
         gdold  = zero
         dtd    = zero

c           for operation counts:
         iter   = 0
         nfgv   = 0
         nseg   = 0
         nintol = 0
         nskip  = 0
         nfree  = n
         ifun   = 0
c           for stopping tolerance:
         tol = factr*epsmch

c           for measuring running time:
         cachyt = 0
         sbtime = 0
         lnscht = 0
 
c           'word' records the status of subspace solutions.
         word = '---'

c           'info' records the termination information.
         info = 0

         itfile = 8
         if (iprint .ge. 1) then
c                                open a summary file 'iterate.dat'
            open (8, file = 'iterate.dat', status = 'unknown')
         endif            

c        Check the input arguments for errors.

         call errclb(n,m,factr,l,u,nbd,task,info,k)
         if (task(1:5) .eq. 'ERROR') then
            call prn3lb(n,x,f,task,iprint,info,itfile,
     +                  iter,nfgv,nintol,nskip,nact,sbgnrm,
     +                  zero,nseg,word,iback,stp,xstep,k,
     +                  cachyt,sbtime,lnscht)
            return
         endif

         call prn1lb(n,m,l,u,x,iprint,itfile,epsmch)
 
c        Initialize iwhere & project x onto the feasible set.
 
         call active(n,l,u,nbd,x,iwhere,iprint,prjctd,cnstnd,boxed) 

c        The end of the initialization.

      else
c          restore local variables.

         prjctd = lsave(1)
         cnstnd = lsave(2)
         boxed  = lsave(3)
         updatd = lsave(4)

         nintol = isave(1)
         itfile = isave(3)
         iback  = isave(4)
         nskip  = isave(5)
         head   = isave(6)
         col    = isave(7)
         itail  = isave(8)
         iter   = isave(9)
         iupdat = isave(10)
         nseg   = isave(12)
         nfgv   = isave(13)
         info   = isave(14)
         ifun   = isave(15)
         iword  = isave(16)
         nfree  = isave(17)
         nact   = isave(18)
         ileave = isave(19)
         nenter = isave(20)

         theta  = dsave(1)
         fold   = dsave(2)
         tol    = dsave(3)
         dnorm  = dsave(4)
         epsmch = dsave(5)
         cpu1   = dsave(6)
         cachyt = dsave(7)
         sbtime = dsave(8)
         lnscht = dsave(9)
         time1  = dsave(10)
         gd     = dsave(11)
         stpmx  = dsave(12)
         sbgnrm = dsave(13)
         stp    = dsave(14)
         gdold  = dsave(15)
         dtd    = dsave(16)
   
c        After returning from the driver go to the point where execution
c        is to resume.

         if (task(1:5) .eq. 'FG_LN') goto 666
         if (task(1:5) .eq. 'NEW_X') goto 777
         if (task(1:5) .eq. 'FG_ST') goto 111
         if (task(1:4) .eq. 'STOP') then
            if (task(7:9) .eq. 'CPU') then
c                                          restore the previous iterate.
               call dcopy(n,t,1,x,1)
               call dcopy(n,r,1,g,1)
               f = fold
            endif
            goto 999
         endif
      endif 

c     Compute f0 and g0.

      task = 'FG_START' 
c          return to the driver to calculate f and g; reenter at 111.
      goto 1000
 111  continue
      nfgv = 1
 
c     Compute the infinity norm of the (-) projected gradient.
 
      call projgr(n,l,u,nbd,x,g,sbgnrm)
  
      if (iprint .ge. 1) then
         write (6,1002) iter,f,sbgnrm
         write (itfile,1003) iter,nfgv,sbgnrm,f
      endif
      if (sbgnrm .le. pgtol) then
c                                terminate the algorithm.
         task = 'CONVERGENCE: NORM_OF_PROJECTED_GRADIENT_<=_PGTOL'
         goto 999
      endif 
 
c ----------------- the beginning of the loop --------------------------
 
 222  continue
      if (iprint .ge. 99) write (6,1001) iter + 1
      iword = -1
c
      if (.not. cnstnd .and. col .gt. 0) then 
c                                            skip the search for GCP.
         call dcopy(n,x,1,z,1)
         wrk = updatd
         nseg = 0
         goto 333
      endif

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c     Compute the Generalized Cauchy Point (GCP).
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      call timer(cpu1) 
      call cauchy(n,x,l,u,nbd,g,indx2,iwhere,t,d,z,
     +            m,wy,ws,sy,wt,theta,col,head,
     +            wa(1),wa(2*m+1),wa(4*m+1),wa(6*m+1),nseg,
     +            iprint, sbgnrm, info, epsmch)
      if (info .ne. 0) then 
c         singular triangular system detected; refresh the lbfgs memory.
         if(iprint .ge. 1) write (6, 1005)
         info   = 0
         col    = 0
         head   = 1
         theta  = one
         iupdat = 0
         updatd = .false.
         call timer(cpu2) 
         cachyt = cachyt + cpu2 - cpu1
         goto 222
      endif
      call timer(cpu2) 
      cachyt = cachyt + cpu2 - cpu1
      nintol = nintol + nseg

c     Count the entering and leaving variables for iter > 0; 
c     find the index set of free and active variables at the GCP.

      call freev(n,nfree,index,nenter,ileave,indx2,
     +           iwhere,wrk,updatd,cnstnd,iprint,iter)
      nact = n - nfree

 333  continue
 
c     If there are no free variables or B=theta*I, then
c                                        skip the subspace minimization.
 
      if (nfree .eq. 0 .or. col .eq. 0) goto 555
 
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c     Subspace minimization.
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      call timer(cpu1) 

c     Form  the LEL^T factorization of the indefinite
c       matrix    K = [-D -Y'ZZ'Y/theta     L_a'-R_z'  ]
c                     [L_a -R_z           theta*S'AA'S ]
c       where     E = [-I  0]
c                     [ 0  I]

      if (wrk) call formk(n,nfree,index,nenter,ileave,indx2,iupdat,
     +                 updatd,wn,snd,m,ws,wy,sy,theta,col,head,info)
      if (info .ne. 0) then
c          nonpositive definiteness in Cholesky factorization;
c          refresh the lbfgs memory and restart the iteration.
         if(iprint .ge. 1) write (6, 1006)
         info   = 0
         col    = 0
         head   = 1
         theta  = one
         iupdat = 0
         updatd = .false.
         call timer(cpu2) 
         sbtime = sbtime + cpu2 - cpu1 
         goto 222
      endif 

c        compute r=-Z'B(xcp-xk)-Z'g (using wa(2m+1)=W'(xcp-x)
c                                                   from 'cauchy').
      call cmprlb(n,m,x,g,ws,wy,sy,wt,z,r,wa,index,
     +           theta,col,head,nfree,cnstnd,info)
      if (info .ne. 0) goto 444

c-jlm-jn   call the direct method. 

      call subsm( n, m, nfree, index, l, u, nbd, z, r, xp, ws, wy,
     +           theta, x, g, col, head, iword, wa, wn, iprint, info)
 444  continue
      if (info .ne. 0) then 
c          singular triangular system detected;
c          refresh the lbfgs memory and restart the iteration.
         if(iprint .ge. 1) write (6, 1005)
         info   = 0
         col    = 0
         head   = 1
         theta  = one
         iupdat = 0
         updatd = .false.
         call timer(cpu2) 
         sbtime = sbtime + cpu2 - cpu1 
         goto 222
      endif
 
      call timer(cpu2) 
      sbtime = sbtime + cpu2 - cpu1 
 555  continue
 
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c     Line search and optimality tests.
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
 
c     Generate the search direction d:=z-x.

      do 40 i = 1, n
         d(i) = z(i) - x(i)
  40  continue
      call timer(cpu1) 
 666  continue
      call lnsrlb(n,l,u,nbd,x,f,fold,gd,gdold,g,d,r,t,z,stp,dnorm,
     +            dtd,xstep,stpmx,iter,ifun,iback,nfgv,info,task,
     +            boxed,cnstnd,csave,isave(22),dsave(17))
      if (info .ne. 0 .or. iback .ge. 20) then
c          restore the previous iterate.
         call dcopy(n,t,1,x,1)
         call dcopy(n,r,1,g,1)
         f = fold
         if (col .eq. 0) then
c             abnormal termination.
            if (info .eq. 0) then
               info = -9
c                restore the actual number of f and g evaluations etc.
               nfgv = nfgv - 1
               ifun = ifun - 1
               iback = iback - 1
            endif
            task = 'ABNORMAL_TERMINATION_IN_LNSRCH'
            iter = iter + 1
            goto 999
         else
c             refresh the lbfgs memory and restart the iteration.
            if(iprint .ge. 1) write (6, 1008)
            if (info .eq. 0) nfgv = nfgv - 1
            info   = 0
            col    = 0
            head   = 1
            theta  = one
            iupdat = 0
            updatd = .false.
            task   = 'RESTART_FROM_LNSRCH'
            call timer(cpu2)
            lnscht = lnscht + cpu2 - cpu1
            goto 222
         endif
      else if (task(1:5) .eq. 'FG_LN') then
c          return to the driver for calculating f and g; reenter at 666.
         goto 1000
      else 
c          calculate and print out the quantities related to the new X.
         call timer(cpu2) 
         lnscht = lnscht + cpu2 - cpu1
         iter = iter + 1
 
c        Compute the infinity norm of the projected (-)gradient.
 
         call projgr(n,l,u,nbd,x,g,sbgnrm)
 
c        Print iteration information.

         call prn2lb(n,x,f,g,iprint,itfile,iter,nfgv,nact,
     +               sbgnrm,nseg,word,iword,iback,stp,xstep)
         goto 1000
      endif
 777  continue

c     Test for termination.

      if (sbgnrm .le. pgtol) then
c                                terminate the algorithm.
         task = 'CONVERGENCE: NORM_OF_PROJECTED_GRADIENT_<=_PGTOL'
         goto 999
      endif 

      ddum = max(abs(fold), abs(f), one)
      if ((fold - f) .le. tol*ddum) then
c                                        terminate the algorithm.
         task = 'CONVERGENCE: REL_REDUCTION_OF_F_<=_FACTR*EPSMCH'
         if (iback .ge. 10) info = -5
c           i.e., to issue a warning if iback>10 in the line search.
         goto 999
      endif 

c     Compute d=newx-oldx, r=newg-oldg, rr=y'y and dr=y's.
 
      do 42 i = 1, n
         r(i) = g(i) - r(i)
  42  continue
      rr = ddot(n,r,1,r,1)
      if (stp .eq. one) then  
         dr = gd - gdold
         ddum = -gdold
      else
         dr = (gd - gdold)*stp
         call dscal(n,stp,d,1)
         ddum = -gdold*stp
      endif
 
      if (dr .le. epsmch*ddum) then
c                            skip the L-BFGS update.
         nskip = nskip + 1
         updatd = .false.
         if (iprint .ge. 1) write (6,1004) dr, ddum
         goto 888
      endif 
 
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c     Update the L-BFGS matrix.
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
 
      updatd = .true.
      iupdat = iupdat + 1

c     Update matrices WS and WY and form the middle matrix in B.

      call matupd(n,m,ws,wy,sy,ss,d,r,itail,
     +            iupdat,col,head,theta,rr,dr,stp,dtd)

c     Form the upper half of the pds T = theta*SS + L*D^(-1)*L';
c        Store T in the upper triangular of the array wt;
c        Cholesky factorize T to J*J' with
c           J' stored in the upper triangular of wt.

      call formt(m,wt,sy,ss,col,theta,info)
 
      if (info .ne. 0) then 
c          nonpositive definiteness in Cholesky factorization;
c          refresh the lbfgs memory and restart the iteration.
         if(iprint .ge. 1) write (6, 1007)
         info = 0
         col = 0
         head = 1
         theta = one
         iupdat = 0
         updatd = .false.
         goto 222
      endif

c     Now the inverse of the middle matrix in B is

c       [  D^(1/2)      O ] [ -D^(1/2)  D^(-1/2)*L' ]
c       [ -L*D^(-1/2)   J ] [  0        J'          ]

 888  continue
 
c -------------------- the end of the loop -----------------------------
 
      goto 222
 999  continue
      call timer(time2)
      time = time2 - time1
      call prn3lb(n,x,f,task,iprint,info,itfile,
     +            iter,nfgv,nintol,nskip,nact,sbgnrm,
     +            time,nseg,word,iback,stp,xstep,k,
     +            cachyt,sbtime,lnscht)
 1000 continue

c     Save local variables.

      lsave(1)  = prjctd
      lsave(2)  = cnstnd
      lsave(3)  = boxed
      lsave(4)  = updatd

      isave(1)  = nintol 
      isave(3)  = itfile 
      isave(4)  = iback 
      isave(5)  = nskip 
      isave(6)  = head 
      isave(7)  = col 
      isave(8)  = itail 
      isave(9)  = iter 
      isave(10) = iupdat 
      isave(12) = nseg
      isave(13) = nfgv 
      isave(14) = info 
      isave(15) = ifun 
      isave(16) = iword 
      isave(17) = nfree 
      isave(18) = nact 
      isave(19) = ileave 
      isave(20) = nenter 

      dsave(1)  = theta 
      dsave(2)  = fold 
      dsave(3)  = tol 
      dsave(4)  = dnorm 
      dsave(5)  = epsmch 
      dsave(6)  = cpu1 
      dsave(7)  = cachyt 
      dsave(8)  = sbtime 
      dsave(9)  = lnscht 
      dsave(10) = time1 
      dsave(11) = gd 
      dsave(12) = stpmx 
      dsave(13) = sbgnrm
      dsave(14) = stp
      dsave(15) = gdold
      dsave(16) = dtd  

 1001 format (//,'ITERATION ',i5)
 1002 format
     +  (/,'At iterate',i5,4x,'f= ',1p,d12.5,4x,'|proj g|= ',1p,d12.5)
 1003 format (2(1x,i4),5x,'-',5x,'-',3x,'-',5x,'-',5x,'-',8x,'-',3x,
     +        1p,2(1x,d10.3))
 1004 format ('  ys=',1p,e10.3,'  -gs=',1p,e10.3,' BFGS update SKIPPED')
 1005 format (/, 
     +' Singular triangular system detected;',/,
     +'   refresh the lbfgs memory and restart the iteration.')
 1006 format (/, 
     +' Nonpositive definiteness in Cholesky factorization in formk;',/,
     +'   refresh the lbfgs memory and restart the iteration.')
 1007 format (/, 
     +' Nonpositive definiteness in Cholesky factorization in formt;',/,
     +'   refresh the lbfgs memory and restart the iteration.')
 1008 format (/, 
     +' Bad direction in the line search;',/,
     +'   refresh the lbfgs memory and restart the iteration.')

      return   

      end
 
c======================= The end of mainlb =============================

      subroutine active(n, l, u, nbd, x, iwhere, iprint,
     +                  prjctd, cnstnd, boxed)

      logical          prjctd, cnstnd, boxed
      integer          n, iprint, nbd(n), iwhere(n)
      double precision x(n), l(n), u(n)

c     ************
c
c     Subroutine active
c
c     This subroutine initializes iwhere and projects the initial x to
c       the feasible set if necessary.
c
c     iwhere is an integer array of dimension n.
c       On entry iwhere is unspecified.
c       On exit iwhere(i)=-1  if x(i) has no bounds
c                         3   if l(i)=u(i)
c                         0   otherwise.
c       In cauchy, iwhere is given finer gradations.
c
c
c                           *  *  *
c
c     NEOS, November 1994. (Latest revision June 1996.)
c     Optimization Technology Center.
c     Argonne National Laboratory and Northwestern University.
c     Written by
c                        Ciyou Zhu
c     in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
c
c
c     ************

      integer          nbdd,i
      double precision zero
      parameter        (zero=0.0d0)

c     Initialize nbdd, prjctd, cnstnd and boxed.

      nbdd = 0
      prjctd = .false.
      cnstnd = .false.
      boxed = .true.

c     Project the initial x to the easible set if necessary.

      do 10 i = 1, n
         if (nbd(i) .gt. 0) then
            if (nbd(i) .le. 2 .and. x(i) .le. l(i)) then
               if (x(i) .lt. l(i)) then
                  prjctd = .true.
                  x(i) = l(i)
               endif
               nbdd = nbdd + 1
            else if (nbd(i) .ge. 2 .and. x(i) .ge. u(i)) then
               if (x(i) .gt. u(i)) then
                  prjctd = .true.
                  x(i) = u(i)
               endif
               nbdd = nbdd + 1
            endif
         endif
  10  continue

c     Initialize iwhere and assign values to cnstnd and boxed.

      do 20 i = 1, n
         if (nbd(i) .ne. 2) boxed = .false.
         if (nbd(i) .eq. 0) then
c                                this variable is always free
            iwhere(i) = -1

c           otherwise set x(i)=mid(x(i), u(i), l(i)).
         else
            cnstnd = .true.
            if (nbd(i) .eq. 2 .and. u(i) - l(i) .le. zero) then
c                   this variable is always fixed
               iwhere(i) = 3
            else 
               iwhere(i) = 0
            endif
         endif
  20  continue

      if (iprint .ge. 0) then
         if (prjctd) write (6,*)
     +   'The initial X is infeasible.  Restart with its projection.'
         if (.not. cnstnd)
     +      write (6,*) 'This problem is unconstrained.'
      endif

      if (iprint .gt. 0) write (6,1001) nbdd

 1001 format (/,'At X0 ',i9,' variables are exactly at the bounds') 

      return

      end

c======================= The end of active =============================
 
      subroutine bmv(m, sy, wt, col, v, p, info)

      integer m, col, info
      double precision sy(m, m), wt(m, m), v(2*col), p(2*col)

c     ************
c
c     Subroutine bmv
c
c     This subroutine computes the product of the 2m x 2m middle matrix 
c       in the compact L-BFGS formula of B and a 2m vector v;  
c       it returns the product in p.
c       
c     m is an integer variable.
c       On entry m is the maximum number of variable metric corrections
c         used to define the limited memory matrix.
c       On exit m is unchanged.
c
c     sy is a double precision array of dimension m x m.
c       On entry sy specifies the matrix S'Y.
c       On exit sy is unchanged.
c
c     wt is a double precision array of dimension m x m.
c       On entry wt specifies the upper triangular matrix J' which is 
c         the Cholesky factor of (thetaS'S+LD^(-1)L').
c       On exit wt is unchanged.
c
c     col is an integer variable.
c       On entry col specifies the number of s-vectors (or y-vectors)
c         stored in the compact L-BFGS formula.
c       On exit col is unchanged.
c
c     v is a double precision array of dimension 2col.
c       On entry v specifies vector v.
c       On exit v is unchanged.
c
c     p is a double precision array of dimension 2col.
c       On entry p is unspecified.
c       On exit p is the product Mv.
c
c     info is an integer variable.
c       On entry info is unspecified.
c       On exit info = 0       for normal return,
c                    = nonzero for abnormal return when the system
c                                to be solved by dtrsl is singular.
c
c     Subprograms called:
c
c       Linpack ... dtrsl.
c
c
c                           *  *  *
c
c     NEOS, November 1994. (Latest revision June 1996.)
c     Optimization Technology Center.
c     Argonne National Laboratory and Northwestern University.
c     Written by
c                        Ciyou Zhu
c     in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
c
c
c     ************
 
      integer          i,k,i2
      double precision sum
 
      if (col .eq. 0) return
 
c     PART I: solve [  D^(1/2)      O ] [ p1 ] = [ v1 ]
c                   [ -L*D^(-1/2)   J ] [ p2 ]   [ v2 ].

c       solve Jp2=v2+LD^(-1)v1.
      p(col + 1) = v(col + 1)
      do 20 i = 2, col
         i2 = col + i
         sum = 0.0d0
         do 10 k = 1, i - 1
            sum = sum + sy(i,k)*v(k)/sy(k,k)
  10     continue
         p(i2) = v(i2) + sum
  20  continue  
c     Solve the triangular system
      call dtrsl(wt,m,col,p(col+1),11,info)
      if (info .ne. 0) return
 
c       solve D^(1/2)p1=v1.
      do 30 i = 1, col
         p(i) = v(i)/sqrt(sy(i,i))
  30  continue 
 
c     PART II: solve [ -D^(1/2)   D^(-1/2)*L'  ] [ p1 ] = [ p1 ]
c                    [  0         J'           ] [ p2 ]   [ p2 ]. 
 
c       solve J^Tp2=p2. 
      call dtrsl(wt,m,col,p(col+1),01,info)
      if (info .ne. 0) return
 
c       compute p1=-D^(-1/2)(p1-D^(-1/2)L'p2)
c                 =-D^(-1/2)p1+D^(-1)L'p2.  
      do 40 i = 1, col
         p(i) = -p(i)/sqrt(sy(i,i))
  40  continue
      do 60 i = 1, col
         sum = 0.d0
         do 50 k = i + 1, col
            sum = sum + sy(k,i)*p(col+k)/sy(i,i)
  50     continue
         p(i) = p(i) + sum
  60  continue

      return

      end

c======================== The end of bmv ===============================

      subroutine cauchy(n, x, l, u, nbd, g, iorder, iwhere, t, d, xcp, 
     +                  m, wy, ws, sy, wt, theta, col, head, p, c, wbp, 
     +                  v, nseg, iprint, sbgnrm, info, epsmch)
      implicit none
      integer          n, m, head, col, nseg, iprint, info, 
     +                 nbd(n), iorder(n), iwhere(n)
      double precision theta, epsmch,
     +                 x(n), l(n), u(n), g(n), t(n), d(n), xcp(n),
     +                 wy(n, col), ws(n, col), sy(m, m),
     +                 wt(m, m), p(2*m), c(2*m), wbp(2*m), v(2*m)

c     ************
c
c     Subroutine cauchy
c
c     For given x, l, u, g (with sbgnrm > 0), and a limited memory
c       BFGS matrix B defined in terms of matrices WY, WS, WT, and
c       scalars head, col, and theta, this subroutine computes the
c       generalized Cauchy point (GCP), defined as the first local
c       minimizer of the quadratic
c
c                  Q(x + s) = g's + 1/2 s'Bs
c
c       along the projected gradient direction P(x-tg,l,u).
c       The routine returns the GCP in xcp. 
c       
c     n is an integer variable.
c       On entry n is the dimension of the problem.
c       On exit n is unchanged.
c
c     x is a double precision array of dimension n.
c       On entry x is the starting point for the GCP computation.
c       On exit x is unchanged.
c
c     l is a double precision array of dimension n.
c       On entry l is the lower bound of x.
c       On exit l is unchanged.
c
c     u is a double precision array of dimension n.
c       On entry u is the upper bound of x.
c       On exit u is unchanged.
c
c     nbd is an integer array of dimension n.
c       On entry nbd represents the type of bounds imposed on the
c         variables, and must be specified as follows:
c         nbd(i)=0 if x(i) is unbounded,
c                1 if x(i) has only a lower bound,
c                2 if x(i) has both lower and upper bounds, and
c                3 if x(i) has only an upper bound. 
c       On exit nbd is unchanged.
c
c     g is a double precision array of dimension n.
c       On entry g is the gradient of f(x).  g must be a nonzero vector.
c       On exit g is unchanged.
c
c     iorder is an integer working array of dimension n.
c       iorder will be used to store the breakpoints in the piecewise
c       linear path and free variables encountered. On exit,
c         iorder(1),...,iorder(nleft) are indices of breakpoints
c                                which have not been encountered; 
c         iorder(nleft+1),...,iorder(nbreak) are indices of
c                                     encountered breakpoints; and
c         iorder(nfree),...,iorder(n) are indices of variables which
c                 have no bound constraits along the search direction.
c
c     iwhere is an integer array of dimension n.
c       On entry iwhere indicates only the permanently fixed (iwhere=3)
c       or free (iwhere= -1) components of x.
c       On exit iwhere records the status of the current x variables.
c       iwhere(i)=-3  if x(i) is free and has bounds, but is not moved
c                 0   if x(i) is free and has bounds, and is moved
c                 1   if x(i) is fixed at l(i), and l(i) .ne. u(i)
c                 2   if x(i) is fixed at u(i), and u(i) .ne. l(i)
c                 3   if x(i) is always fixed, i.e.,  u(i)=x(i)=l(i)
c                 -1  if x(i) is always free, i.e., it has no bounds.
c
c     t is a double precision working array of dimension n. 
c       t will be used to store the break points.
c
c     d is a double precision array of dimension n used to store
c       the Cauchy direction P(x-tg)-x.
c
c     xcp is a double precision array of dimension n used to return the
c       GCP on exit.
c
c     m is an integer variable.
c       On entry m is the maximum number of variable metric corrections 
c         used to define the limited memory matrix.
c       On exit m is unchanged.
c
c     ws, wy, sy, and wt are double precision arrays.
c       On entry they store information that defines the
c                             limited memory BFGS matrix:
c         ws(n,m) stores S, a set of s-vectors;
c         wy(n,m) stores Y, a set of y-vectors;
c         sy(m,m) stores S'Y;
c         wt(m,m) stores the
c                 Cholesky factorization of (theta*S'S+LD^(-1)L').
c       On exit these arrays are unchanged.
c
c     theta is a double precision variable.
c       On entry theta is the scaling factor specifying B_0 = theta I.
c       On exit theta is unchanged.
c
c     col is an integer variable.
c       On entry col is the actual number of variable metric
c         corrections stored so far.
c       On exit col is unchanged.
c
c     head is an integer variable.
c       On entry head is the location of the first s-vector (or y-vector)
c         in S (or Y).
c       On exit col is unchanged.
c
c     p is a double precision working array of dimension 2m.
c       p will be used to store the vector p = W^(T)d.
c
c     c is a double precision working array of dimension 2m.
c       c will be used to store the vector c = W^(T)(xcp-x).
c
c     wbp is a double precision working array of dimension 2m.
c       wbp will be used to store the row of W corresponding
c         to a breakpoint.
c
c     v is a double precision working array of dimension 2m.
c
c     nseg is an integer variable.
c       On exit nseg records the number of quadratic segments explored
c         in searching for the GCP.
c
c     sg and yg are double precision arrays of dimension m.
c       On entry sg  and yg store S'g and Y'g correspondingly.
c       On exit they are unchanged. 
c 
c     iprint is an INTEGER variable that must be set by the user.
c       It controls the frequency and type of output generated:
c        iprint<0    no output is generated;
c        iprint=0    print only one line at the last iteration;
c        0<iprint<99 print also f and |proj g| every iprint iterations;
c        iprint=99   print details of every iteration except n-vectors;
c        iprint=100  print also the changes of active set and final x;
c        iprint>100  print details of every iteration including x and g;
c       When iprint > 0, the file iterate.dat will be created to
c                        summarize the iteration.
c
c     sbgnrm is a double precision variable.
c       On entry sbgnrm is the norm of the projected gradient at x.
c       On exit sbgnrm is unchanged.
c
c     info is an integer variable.
c       On entry info is 0.
c       On exit info = 0       for normal return,
c                    = nonzero for abnormal return when the the system
c                              used in routine bmv is singular.
c
c     Subprograms called:
c 
c       L-BFGS-B Library ... hpsolb, bmv.
c
c       Linpack ... dscal dcopy, daxpy.
c
c
c     References:
c
c       [1] R. H. Byrd, P. Lu, J. Nocedal and C. Zhu, ``A limited
c       memory algorithm for bound constrained optimization'',
c       SIAM J. Scientific Computing 16 (1995), no. 5, pp. 1190--1208.
c
c       [2] C. Zhu, R.H. Byrd, P. Lu, J. Nocedal, ``L-BFGS-B: FORTRAN
c       Subroutines for Large Scale Bound Constrained Optimization''
c       Tech. Report, NAM-11, EECS Department, Northwestern University,
c       1994.
c
c       (Postscript files of these papers are available via anonymous
c        ftp to eecs.nwu.edu in the directory pub/lbfgs/lbfgs_bcm.)
c
c                           *  *  *
c
c     NEOS, November 1994. (Latest revision June 1996.)
c     Optimization Technology Center.
c     Argonne National Laboratory and Northwestern University.
c     Written by
c                        Ciyou Zhu
c     in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
c
c
c     ************

      logical          xlower,xupper,bnded
      integer          i,j,col2,nfree,nbreak,pointr,
     +                 ibp,nleft,ibkmin,iter
      double precision f1,f2,dt,dtm,tsum,dibp,zibp,dibp2,bkmin,
     +                 tu,tl,wmc,wmp,wmw,ddot,tj,tj0,neggi,sbgnrm,
     +                 f2_org
      double precision one,zero
      parameter        (one=1.0d0,zero=0.0d0)
 
c     Check the status of the variables, reset iwhere(i) if necessary;
c       compute the Cauchy direction d and the breakpoints t; initialize
c       the derivative f1 and the vector p = W'd (for theta = 1).
 
      if (sbgnrm .le. zero) then
         if (iprint .ge. 0) write (6,*) 'Subgnorm = 0.  GCP = X.'
         call dcopy(n,x,1,xcp,1)
         return
      endif 
      bnded = .true.
      nfree = n + 1
      nbreak = 0
      ibkmin = 0
      bkmin = zero
      col2 = 2*col
      f1 = zero
      if (iprint .ge. 99) write (6,3010)

c     We set p to zero and build it up as we determine d.

      do 20 i = 1, col2
         p(i) = zero
  20  continue 

c     In the following loop we determine for each variable its bound
c        status and its breakpoint, and update p accordingly.
c        Smallest breakpoint is identified.

      do 50 i = 1, n 
         neggi = -g(i)      
         if (iwhere(i) .ne. 3 .and. iwhere(i) .ne. -1) then
c             if x(i) is not a constant and has bounds,
c             compute the difference between x(i) and its bounds.
            if (nbd(i) .le. 2) tl = x(i) - l(i)
            if (nbd(i) .ge. 2) tu = u(i) - x(i)

c           If a variable is close enough to a bound
c             we treat it as at bound.
            xlower = nbd(i) .le. 2 .and. tl .le. zero
            xupper = nbd(i) .ge. 2 .and. tu .le. zero

c              reset iwhere(i).
            iwhere(i) = 0
            if (xlower) then
               if (neggi .le. zero) iwhere(i) = 1
            else if (xupper) then
               if (neggi .ge. zero) iwhere(i) = 2
            else
               if (abs(neggi) .le. zero) iwhere(i) = -3
            endif
         endif 
         pointr = head
         if (iwhere(i) .ne. 0 .and. iwhere(i) .ne. -1) then
            d(i) = zero
         else
            d(i) = neggi
            f1 = f1 - neggi*neggi
c             calculate p := p - W'e_i* (g_i).
            do 40 j = 1, col
               p(j) = p(j) +  wy(i,pointr)* neggi
               p(col + j) = p(col + j) + ws(i,pointr)*neggi
               pointr = mod(pointr,m) + 1
  40        continue 
            if (nbd(i) .le. 2 .and. nbd(i) .ne. 0
     +                        .and. neggi .lt. zero) then
c                                 x(i) + d(i) is bounded; compute t(i).
               nbreak = nbreak + 1
               iorder(nbreak) = i
               t(nbreak) = tl/(-neggi)
               if (nbreak .eq. 1 .or. t(nbreak) .lt. bkmin) then
                  bkmin = t(nbreak)
                  ibkmin = nbreak
               endif
            else if (nbd(i) .ge. 2 .and. neggi .gt. zero) then
c                                 x(i) + d(i) is bounded; compute t(i).
               nbreak = nbreak + 1
               iorder(nbreak) = i
               t(nbreak) = tu/neggi
               if (nbreak .eq. 1 .or. t(nbreak) .lt. bkmin) then
                  bkmin = t(nbreak)
                  ibkmin = nbreak
               endif
            else
c                x(i) + d(i) is not bounded.
               nfree = nfree - 1
               iorder(nfree) = i
               if (abs(neggi) .gt. zero) bnded = .false.
            endif
         endif
  50  continue 
 
c     The indices of the nonzero components of d are now stored
c       in iorder(1),...,iorder(nbreak) and iorder(nfree),...,iorder(n).
c       The smallest of the nbreak breakpoints is in t(ibkmin)=bkmin.
 
      if (theta .ne. one) then
c                   complete the initialization of p for theta not= one.
         call dscal(col,theta,p(col+1),1)
      endif
 
c     Initialize GCP xcp = x.

      call dcopy(n,x,1,xcp,1)

      if (nbreak .eq. 0 .and. nfree .eq. n + 1) then
c                  is a zero vector, return with the initial xcp as GCP.
         if (iprint .gt. 100) write (6,1010) (xcp(i), i = 1, n)
         return
      endif    
 
c     Initialize c = W'(xcp - x) = 0.
  
      do 60 j = 1, col2
         c(j) = zero
  60  continue 
 
c     Initialize derivative f2.
 
      f2 =  -theta*f1 
      f2_org  =  f2
      if (col .gt. 0) then
         call bmv(m,sy,wt,col,p,v,info)
         if (info .ne. 0) return
         f2 = f2 - ddot(col2,v,1,p,1)
      endif
      dtm = -f1/f2
      tsum = zero
      nseg = 1
      if (iprint .ge. 99) 
     +   write (6,*) 'There are ',nbreak,'  breakpoints '
 
c     If there are no breakpoints, locate the GCP and return. 
 
      if (nbreak .eq. 0) goto 888
             
      nleft = nbreak
      iter = 1
 
 
      tj = zero
 
c------------------- the beginning of the loop -------------------------
 
 777  continue
 
c     Find the next smallest breakpoint;
c       compute dt = t(nleft) - t(nleft + 1).
 
      tj0 = tj
      if (iter .eq. 1) then
c         Since we already have the smallest breakpoint we need not do
c         heapsort yet. Often only one breakpoint is used and the
c         cost of heapsort is avoided.
         tj = bkmin
         ibp = iorder(ibkmin)
      else
         if (iter .eq. 2) then
c             Replace the already used smallest breakpoint with the
c             breakpoint numbered nbreak > nlast, before heapsort call.
            if (ibkmin .ne. nbreak) then
               t(ibkmin) = t(nbreak)
               iorder(ibkmin) = iorder(nbreak)
            endif 
c        Update heap structure of breakpoints
c           (if iter=2, initialize heap).
         endif
         call hpsolb(nleft,t,iorder,iter-2)
         tj = t(nleft)
         ibp = iorder(nleft)  
      endif 
         
      dt = tj - tj0
 
      if (dt .ne. zero .and. iprint .ge. 100) then
         write (6,4011) nseg,f1,f2
         write (6,5010) dt
         write (6,6010) dtm
      endif          
 
c     If a minimizer is within this interval, locate the GCP and return. 
 
      if (dtm .lt. dt) goto 888
 
c     Otherwise fix one variable and
c       reset the corresponding component of d to zero.
    
      tsum = tsum + dt
      nleft = nleft - 1
      iter = iter + 1
      dibp = d(ibp)
      d(ibp) = zero
      if (dibp .gt. zero) then
         zibp = u(ibp) - x(ibp)
         xcp(ibp) = u(ibp)
         iwhere(ibp) = 2
      else
         zibp = l(ibp) - x(ibp)
         xcp(ibp) = l(ibp)
         iwhere(ibp) = 1
      endif
      if (iprint .ge. 100) write (6,*) 'Variable  ',ibp,'  is fixed.'
      if (nleft .eq. 0 .and. nbreak .eq. n) then
c                                             all n variables are fixed,
c                                                return with xcp as GCP.
         dtm = dt
         goto 999
      endif
 
c     Update the derivative information.
 
      nseg = nseg + 1
      dibp2 = dibp**2
 
c     Update f1 and f2.
 
c        temporarily set f1 and f2 for col=0.
      f1 = f1 + dt*f2 + dibp2 - theta*dibp*zibp
      f2 = f2 - theta*dibp2

      if (col .gt. 0) then
c                          update c = c + dt*p.
         call daxpy(col2,dt,p,1,c,1)
 
c           choose wbp,
c           the row of W corresponding to the breakpoint encountered.
         pointr = head
         do 70 j = 1,col
            wbp(j) = wy(ibp,pointr)
            wbp(col + j) = theta*ws(ibp,pointr)
            pointr = mod(pointr,m) + 1
  70     continue 
 
c           compute (wbp)Mc, (wbp)Mp, and (wbp)M(wbp)'.
         call bmv(m,sy,wt,col,wbp,v,info)
         if (info .ne. 0) return
         wmc = ddot(col2,c,1,v,1)
         wmp = ddot(col2,p,1,v,1) 
         wmw = ddot(col2,wbp,1,v,1)
 
c           update p = p - dibp*wbp. 
         call daxpy(col2,-dibp,wbp,1,p,1)
 
c           complete updating f1 and f2 while col > 0.
         f1 = f1 + dibp*wmc
         f2 = f2 + 2.0d0*dibp*wmp - dibp2*wmw
      endif

      f2 = max(epsmch*f2_org,f2)
      if (nleft .gt. 0) then
         dtm = -f1/f2
         goto 777
c                 to repeat the loop for unsearched intervals. 
      else if(bnded) then
         f1 = zero
         f2 = zero
         dtm = zero
      else
         dtm = -f1/f2
      endif 

c------------------- the end of the loop -------------------------------
 
 888  continue
      if (iprint .ge. 99) then
         write (6,*)
         write (6,*) 'GCP found in this segment'
         write (6,4010) nseg,f1,f2
         write (6,6010) dtm
      endif 
      if (dtm .le. zero) dtm = zero
      tsum = tsum + dtm
 
c     Move free variables (i.e., the ones w/o breakpoints) and 
c       the variables whose breakpoints haven't been reached.
 
      call daxpy(n,tsum,d,1,xcp,1)
 
 999  continue
 
c     Update c = c + dtm*p = W'(x^c - x) 
c       which will be used in computing r = Z'(B(x^c - x) + g).
 
      if (col .gt. 0) call daxpy(col2,dtm,p,1,c,1)
      if (iprint .gt. 100) write (6,1010) (xcp(i),i = 1,n)
      if (iprint .ge. 99) write (6,2010)

 1010 format ('Cauchy X =  ',/,(4x,1p,6(1x,d11.4)))
 2010 format (/,'---------------- exit CAUCHY----------------------',/)
 3010 format (/,'---------------- CAUCHY entered-------------------')
 4010 format ('Piece    ',i3,' --f1, f2 at start point ',1p,2(1x,d11.4))
 4011 format (/,'Piece    ',i3,' --f1, f2 at start point ',
     +        1p,2(1x,d11.4))
 5010 format ('Distance to the next break point =  ',1p,d11.4)
 6010 format ('Distance to the stationary point =  ',1p,d11.4) 
 
      return
 
      end

c====================== The end of cauchy ==============================

      subroutine cmprlb(n, m, x, g, ws, wy, sy, wt, z, r, wa, index, 
     +                 theta, col, head, nfree, cnstnd, info)
 
      logical          cnstnd
      integer          n, m, col, head, nfree, info, index(n)
      double precision theta, 
     +                 x(n), g(n), z(n), r(n), wa(4*m), 
     +                 ws(n, m), wy(n, m), sy(m, m), wt(m, m)

c     ************
c
c     Subroutine cmprlb 
c
c       This subroutine computes r=-Z'B(xcp-xk)-Z'g by using 
c         wa(2m+1)=W'(xcp-x) from subroutine cauchy.
c
c     Subprograms called:
c
c       L-BFGS-B Library ... bmv.
c
c
c                           *  *  *
c
c     NEOS, November 1994. (Latest revision June 1996.)
c     Optimization Technology Center.
c     Argonne National Laboratory and Northwestern University.
c     Written by
c                        Ciyou Zhu
c     in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
c
c
c     ************
 
      integer          i,j,k,pointr
      double precision a1,a2

      if (.not. cnstnd .and. col .gt. 0) then 
         do 26 i = 1, n
            r(i) = -g(i)
  26     continue
      else
         do 30 i = 1, nfree
            k = index(i)
            r(i) = -theta*(z(k) - x(k)) - g(k)
  30     continue
         call bmv(m,sy,wt,col,wa(2*m+1),wa(1),info)
         if (info .ne. 0) then
            info = -8
            return
         endif
         pointr = head 
         do 34 j = 1, col
            a1 = wa(j)
            a2 = theta*wa(col + j)
            do 32 i = 1, nfree
               k = index(i)
               r(i) = r(i) + wy(k,pointr)*a1 + ws(k,pointr)*a2
  32        continue
            pointr = mod(pointr,m) + 1
  34     continue
      endif

      return

      end

c======================= The end of cmprlb =============================

      subroutine errclb(n, m, factr, l, u, nbd, task, info, k)
 
      character*60     task
      integer          n, m, info, k, nbd(n)
      double precision factr, l(n), u(n)

c     ************
c
c     Subroutine errclb
c
c     This subroutine checks the validity of the input data.
c
c
c                           *  *  *
c
c     NEOS, November 1994. (Latest revision June 1996.)
c     Optimization Technology Center.
c     Argonne National Laboratory and Northwestern University.
c     Written by
c                        Ciyou Zhu
c     in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
c
c
c     ************

      integer          i
      double precision one,zero
      parameter        (one=1.0d0,zero=0.0d0)

c     Check the input arguments for errors.

      if (n .le. 0) task = 'ERROR: N .LE. 0'
      if (m .le. 0) task = 'ERROR: M .LE. 0'
      if (factr .lt. zero) task = 'ERROR: FACTR .LT. 0'

c     Check the validity of the arrays nbd(i), u(i), and l(i).

      do 10 i = 1, n
         if (nbd(i) .lt. 0 .or. nbd(i) .gt. 3) then
c                                                   return
            task = 'ERROR: INVALID NBD'
            info = -6
            k = i
         endif
         if (nbd(i) .eq. 2) then
            if (l(i) .gt. u(i)) then
c                                    return
               task = 'ERROR: NO FEASIBLE SOLUTION'
               info = -7
               k = i
            endif
         endif
  10  continue

      return

      end

c======================= The end of errclb =============================
 
      subroutine formk(n, nsub, ind, nenter, ileave, indx2, iupdat, 
     +                 updatd, wn, wn1, m, ws, wy, sy, theta, col,
     +                 head, info)

      integer          n, nsub, m, col, head, nenter, ileave, iupdat,
     +                 info, ind(n), indx2(n)
      double precision theta, wn(2*m, 2*m), wn1(2*m, 2*m),
     +                 ws(n, m), wy(n, m), sy(m, m)
      logical          updatd

c     ************
c
c     Subroutine formk 
c
c     This subroutine forms  the LEL^T factorization of the indefinite
c
c       matrix    K = [-D -Y'ZZ'Y/theta     L_a'-R_z'  ]
c                     [L_a -R_z           theta*S'AA'S ]
c                                                    where E = [-I  0]
c                                                              [ 0  I]
c     The matrix K can be shown to be equal to the matrix M^[-1]N
c       occurring in section 5.1 of [1], as well as to the matrix
c       Mbar^[-1] Nbar in section 5.3.
c
c     n is an integer variable.
c       On entry n is the dimension of the problem.
c       On exit n is unchanged.
c
c     nsub is an integer variable
c       On entry nsub is the number of subspace variables in free set.
c       On exit nsub is not changed.
c
c     ind is an integer array of dimension nsub.
c       On entry ind specifies the indices of subspace variables.
c       On exit ind is unchanged. 
c
c     nenter is an integer variable.
c       On entry nenter is the number of variables entering the 
c         free set.
c       On exit nenter is unchanged. 
c
c     ileave is an integer variable.
c       On entry indx2(ileave),...,indx2(n) are the variables leaving
c         the free set.
c       On exit ileave is unchanged. 
c
c     indx2 is an integer array of dimension n.
c       On entry indx2(1),...,indx2(nenter) are the variables entering
c         the free set, while indx2(ileave),...,indx2(n) are the
c         variables leaving the free set.
c       On exit indx2 is unchanged. 
c
c     iupdat is an integer variable.
c       On entry iupdat is the total number of BFGS updates made so far.
c       On exit iupdat is unchanged. 
c
c     updatd is a logical variable.
c       On entry 'updatd' is true if the L-BFGS matrix is updatd.
c       On exit 'updatd' is unchanged. 
c
c     wn is a double precision array of dimension 2m x 2m.
c       On entry wn is unspecified.
c       On exit the upper triangle of wn stores the LEL^T factorization
c         of the 2*col x 2*col indefinite matrix
c                     [-D -Y'ZZ'Y/theta     L_a'-R_z'  ]
c                     [L_a -R_z           theta*S'AA'S ]
c
c     wn1 is a double precision array of dimension 2m x 2m.
c       On entry wn1 stores the lower triangular part of 
c                     [Y' ZZ'Y   L_a'+R_z']
c                     [L_a+R_z   S'AA'S   ]
c         in the previous iteration.
c       On exit wn1 stores the corresponding updated matrices.
c       The purpose of wn1 is just to store these inner products
c       so they can be easily updated and inserted into wn.
c
c     m is an integer variable.
c       On entry m is the maximum number of variable metric corrections
c         used to define the limited memory matrix.
c       On exit m is unchanged.
c
c     ws, wy, sy, and wtyy are double precision arrays;
c     theta is a double precision variable;
c     col is an integer variable;
c     head is an integer variable.
c       On entry they store the information defining the
c                                          limited memory BFGS matrix:
c         ws(n,m) stores S, a set of s-vectors;
c         wy(n,m) stores Y, a set of y-vectors;
c         sy(m,m) stores S'Y;
c         wtyy(m,m) stores the Cholesky factorization
c                                   of (theta*S'S+LD^(-1)L')
c         theta is the scaling factor specifying B_0 = theta I;
c         col is the number of variable metric corrections stored;
c         head is the location of the 1st s- (or y-) vector in S (or Y).
c       On exit they are unchanged.
c
c     info is an integer variable.
c       On entry info is unspecified.
c       On exit info =  0 for normal return;
c                    = -1 when the 1st Cholesky factorization failed;
c                    = -2 when the 2st Cholesky factorization failed.
c
c     Subprograms called:
c
c       Linpack ... dcopy, dpofa, dtrsl.
c
c
c     References:
c       [1] R. H. Byrd, P. Lu, J. Nocedal and C. Zhu, ``A limited
c       memory algorithm for bound constrained optimization'',
c       SIAM J. Scientific Computing 16 (1995), no. 5, pp. 1190--1208.
c
c       [2] C. Zhu, R.H. Byrd, P. Lu, J. Nocedal, ``L-BFGS-B: a
c       limited memory FORTRAN code for solving bound constrained
c       optimization problems'', Tech. Report, NAM-11, EECS Department,
c       Northwestern University, 1994.
c
c       (Postscript files of these papers are available via anonymous
c        ftp to eecs.nwu.edu in the directory pub/lbfgs/lbfgs_bcm.)
c
c                           *  *  *
c
c     NEOS, November 1994. (Latest revision June 1996.)
c     Optimization Technology Center.
c     Argonne National Laboratory and Northwestern University.
c     Written by
c                        Ciyou Zhu
c     in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
c
c
c     ************

      integer          m2,ipntr,jpntr,iy,is,jy,js,is1,js1,k1,i,k,
     +                 col2,pbegin,pend,dbegin,dend,upcl
      double precision ddot,temp1,temp2,temp3,temp4
      double precision one,zero
      parameter        (one=1.0d0,zero=0.0d0)

c     Form the lower triangular part of
c               WN1 = [Y' ZZ'Y   L_a'+R_z'] 
c                     [L_a+R_z   S'AA'S   ]
c        where L_a is the strictly lower triangular part of S'AA'Y
c              R_z is the upper triangular part of S'ZZ'Y.
      
      if (updatd) then
         if (iupdat .gt. m) then 
c                                 shift old part of WN1.
            do 10 jy = 1, m - 1
               js = m + jy
               call dcopy(m-jy,wn1(jy+1,jy+1),1,wn1(jy,jy),1)
               call dcopy(m-jy,wn1(js+1,js+1),1,wn1(js,js),1)
               call dcopy(m-1,wn1(m+2,jy+1),1,wn1(m+1,jy),1)
  10        continue
         endif
 
c          put new rows in blocks (1,1), (2,1) and (2,2).
         pbegin = 1
         pend = nsub
         dbegin = nsub + 1
         dend = n
         iy = col
         is = m + col
         ipntr = head + col - 1
         if (ipntr .gt. m) ipntr = ipntr - m    
         jpntr = head
         do 20 jy = 1, col
            js = m + jy
            temp1 = zero
            temp2 = zero
            temp3 = zero
c             compute element jy of row 'col' of Y'ZZ'Y
            do 15 k = pbegin, pend
               k1 = ind(k)
               temp1 = temp1 + wy(k1,ipntr)*wy(k1,jpntr)
  15        continue
c             compute elements jy of row 'col' of L_a and S'AA'S
            do 16 k = dbegin, dend
               k1 = ind(k)
               temp2 = temp2 + ws(k1,ipntr)*ws(k1,jpntr)
               temp3 = temp3 + ws(k1,ipntr)*wy(k1,jpntr)
  16        continue
            wn1(iy,jy) = temp1
            wn1(is,js) = temp2
            wn1(is,jy) = temp3
            jpntr = mod(jpntr,m) + 1
  20     continue
 
c          put new column in block (2,1).
         jy = col       
         jpntr = head + col - 1
         if (jpntr .gt. m) jpntr = jpntr - m
         ipntr = head
         do 30 i = 1, col
            is = m + i
            temp3 = zero
c             compute element i of column 'col' of R_z
            do 25 k = pbegin, pend
               k1 = ind(k)
               temp3 = temp3 + ws(k1,ipntr)*wy(k1,jpntr)
  25        continue 
            ipntr = mod(ipntr,m) + 1
            wn1(is,jy) = temp3
  30     continue
         upcl = col - 1
      else
         upcl = col
      endif
 
c       modify the old parts in blocks (1,1) and (2,2) due to changes
c       in the set of free variables.
      ipntr = head      
      do 45 iy = 1, upcl
         is = m + iy
         jpntr = head
         do 40 jy = 1, iy
            js = m + jy
            temp1 = zero
            temp2 = zero
            temp3 = zero
            temp4 = zero
            do 35 k = 1, nenter
               k1 = indx2(k)
               temp1 = temp1 + wy(k1,ipntr)*wy(k1,jpntr)
               temp2 = temp2 + ws(k1,ipntr)*ws(k1,jpntr)
  35        continue
            do 36 k = ileave, n
               k1 = indx2(k)
               temp3 = temp3 + wy(k1,ipntr)*wy(k1,jpntr)
               temp4 = temp4 + ws(k1,ipntr)*ws(k1,jpntr)
  36        continue
            wn1(iy,jy) = wn1(iy,jy) + temp1 - temp3 
            wn1(is,js) = wn1(is,js) - temp2 + temp4 
            jpntr = mod(jpntr,m) + 1
  40     continue
         ipntr = mod(ipntr,m) + 1
  45  continue
 
c       modify the old parts in block (2,1).
      ipntr = head      
      do 60 is = m + 1, m + upcl
         jpntr = head 
         do 55 jy = 1, upcl
            temp1 = zero
            temp3 = zero
            do 50 k = 1, nenter
               k1 = indx2(k)
               temp1 = temp1 + ws(k1,ipntr)*wy(k1,jpntr)
  50        continue
            do 51 k = ileave, n
               k1 = indx2(k)
               temp3 = temp3 + ws(k1,ipntr)*wy(k1,jpntr)
  51        continue
         if (is .le. jy + m) then
               wn1(is,jy) = wn1(is,jy) + temp1 - temp3  
            else
               wn1(is,jy) = wn1(is,jy) - temp1 + temp3  
            endif
            jpntr = mod(jpntr,m) + 1
  55     continue
         ipntr = mod(ipntr,m) + 1
  60  continue
 
c     Form the upper triangle of WN = [D+Y' ZZ'Y/theta   -L_a'+R_z' ] 
c                                     [-L_a +R_z        S'AA'S*theta]

      m2 = 2*m
      do 70 iy = 1, col
         is = col + iy
         is1 = m + iy
         do 65 jy = 1, iy
            js = col + jy
            js1 = m + jy
            wn(jy,iy) = wn1(iy,jy)/theta
            wn(js,is) = wn1(is1,js1)*theta
  65     continue
         do 66 jy = 1, iy - 1
            wn(jy,is) = -wn1(is1,jy)
  66     continue
         do 67 jy = iy, col
            wn(jy,is) = wn1(is1,jy)
  67     continue
         wn(iy,iy) = wn(iy,iy) + sy(iy,iy)
  70  continue

c     Form the upper triangle of WN= [  LL'            L^-1(-L_a'+R_z')] 
c                                    [(-L_a +R_z)L'^-1   S'AA'S*theta  ]

c        first Cholesky factor (1,1) block of wn to get LL'
c                          with L' stored in the upper triangle of wn.
      call dpofa(wn,m2,col,info)
      if (info .ne. 0) then
         info = -1
         return
      endif
c        then form L^-1(-L_a'+R_z') in the (1,2) block.
      col2 = 2*col
      do 71 js = col+1 ,col2
         call dtrsl(wn,m2,col,wn(1,js),11,info)
  71  continue

c     Form S'AA'S*theta + (L^-1(-L_a'+R_z'))'L^-1(-L_a'+R_z') in the
c        upper triangle of (2,2) block of wn.
                      

      do 72 is = col+1, col2
         do 74 js = is, col2
               wn(is,js) = wn(is,js) + ddot(col,wn(1,is),1,wn(1,js),1)
  74        continue
  72     continue

c     Cholesky factorization of (2,2) block of wn.

      call dpofa(wn(col+1,col+1),m2,col,info)
      if (info .ne. 0) then
         info = -2
         return
      endif

      return

      end

c======================= The end of formk ==============================

      subroutine formt(m, wt, sy, ss, col, theta, info)
 
      integer          m, col, info
      double precision theta, wt(m, m), sy(m, m), ss(m, m)

c     ************
c
c     Subroutine formt
c
c       This subroutine forms the upper half of the pos. def. and symm.
c         T = theta*SS + L*D^(-1)*L', stores T in the upper triangle
c         of the array wt, and performs the Cholesky factorization of T
c         to produce J*J', with J' stored in the upper triangle of wt.
c
c     Subprograms called:
c
c       Linpack ... dpofa.
c
c
c                           *  *  *
c
c     NEOS, November 1994. (Latest revision June 1996.)
c     Optimization Technology Center.
c     Argonne National Laboratory and Northwestern University.
c     Written by
c                        Ciyou Zhu
c     in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
c
c
c     ************

      integer          i,j,k,k1
      double precision ddum
      double precision zero
      parameter        (zero=0.0d0)


c     Form the upper half of  T = theta*SS + L*D^(-1)*L',
c        store T in the upper triangle of the array wt.
 
      do 52 j = 1, col
         wt(1,j) = theta*ss(1,j)
  52  continue
      do 55 i = 2, col
         do 54 j = i, col
            k1 = min(i,j) - 1
            ddum  = zero
            do 53 k = 1, k1
               ddum  = ddum + sy(i,k)*sy(j,k)/sy(k,k)
  53        continue
            wt(i,j) = ddum + theta*ss(i,j)
  54     continue
  55  continue
 
c     Cholesky factorize T to J*J' with 
c        J' stored in the upper triangle of wt.
 
      call dpofa(wt,m,col,info)
      if (info .ne. 0) then
         info = -3
      endif

      return

      end

c======================= The end of formt ==============================
 
      subroutine freev(n, nfree, index, nenter, ileave, indx2, 
     +                 iwhere, wrk, updatd, cnstnd, iprint, iter)

      integer n, nfree, nenter, ileave, iprint, iter, 
     +        index(n), indx2(n), iwhere(n)
      logical wrk, updatd, cnstnd

c     ************
c
c     Subroutine freev 
c
c     This subroutine counts the entering and leaving variables when
c       iter > 0, and finds the index set of free and active variables
c       at the GCP.
c
c     cnstnd is a logical variable indicating whether bounds are present
c
c     index is an integer array of dimension n
c       for i=1,...,nfree, index(i) are the indices of free variables
c       for i=nfree+1,...,n, index(i) are the indices of bound variables
c       On entry after the first iteration, index gives 
c         the free variables at the previous iteration.
c       On exit it gives the free variables based on the determination
c         in cauchy using the array iwhere.
c
c     indx2 is an integer array of dimension n
c       On entry indx2 is unspecified.
c       On exit with iter>0, indx2 indicates which variables
c          have changed status since the previous iteration.
c       For i= 1,...,nenter, indx2(i) have changed from bound to free.
c       For i= ileave+1,...,n, indx2(i) have changed from free to bound.
c 
c
c                           *  *  *
c
c     NEOS, November 1994. (Latest revision June 1996.)
c     Optimization Technology Center.
c     Argonne National Laboratory and Northwestern University.
c     Written by
c                        Ciyou Zhu
c     in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
c
c
c     ************
 
      integer iact,i,k

      nenter = 0
      ileave = n + 1
      if (iter .gt. 0 .and. cnstnd) then
c                           count the entering and leaving variables.
         do 20 i = 1, nfree
            k = index(i)

c            write(6,*) ' k  = index(i) ', k
c            write(6,*) ' index = ', i

            if (iwhere(k) .gt. 0) then
               ileave = ileave - 1
               indx2(ileave) = k
               if (iprint .ge. 100) write (6,*)
     +             'Variable ',k,' leaves the set of free variables'
            endif
  20     continue
         do 22 i = 1 + nfree, n
            k = index(i)
            if (iwhere(k) .le. 0) then
               nenter = nenter + 1
               indx2(nenter) = k
               if (iprint .ge. 100) write (6,*)
     +             'Variable ',k,' enters the set of free variables'
            endif
  22     continue
         if (iprint .ge. 99) write (6,*)
     +       n+1-ileave,' variables leave; ',nenter,' variables enter'
      endif
      wrk = (ileave .lt. n+1) .or. (nenter .gt. 0) .or. updatd
 
c     Find the index set of free and active variables at the GCP.
 
      nfree = 0 
      iact = n + 1
      do 24 i = 1, n
         if (iwhere(i) .le. 0) then
            nfree = nfree + 1
            index(nfree) = i
         else
            iact = iact - 1
            index(iact) = i
         endif
  24  continue
      if (iprint .ge. 99) write (6,*)
     +      nfree,' variables are free at GCP ',iter + 1  

      return

      end

c======================= The end of freev ==============================

      subroutine hpsolb(n, t, iorder, iheap)
      integer          iheap, n, iorder(n)
      double precision t(n)
  
c     ************
c
c     Subroutine hpsolb 
c
c     This subroutine sorts out the least element of t, and puts the
c       remaining elements of t in a heap.
c 
c     n is an integer variable.
c       On entry n is the dimension of the arrays t and iorder.
c       On exit n is unchanged.
c
c     t is a double precision array of dimension n.
c       On entry t stores the elements to be sorted,
c       On exit t(n) stores the least elements of t, and t(1) to t(n-1)
c         stores the remaining elements in the form of a heap.
c
c     iorder is an integer array of dimension n.
c       On entry iorder(i) is the index of t(i).
c       On exit iorder(i) is still the index of t(i), but iorder may be
c         permuted in accordance with t.
c
c     iheap is an integer variable specifying the task.
c       On entry iheap should be set as follows:
c         iheap .eq. 0 if t(1) to t(n) is not in the form of a heap,
c         iheap .ne. 0 if otherwise.
c       On exit iheap is unchanged.
c
c
c     References:
c       Algorithm 232 of CACM (J. W. J. Williams): HEAPSORT.
c
c                           *  *  *
c
c     NEOS, November 1994. (Latest revision June 1996.)
c     Optimization Technology Center.
c     Argonne National Laboratory and Northwestern University.
c     Written by
c                        Ciyou Zhu
c     in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
c
c     ************
  
      integer          i,j,k,indxin,indxou
      double precision ddum,out

      if (iheap .eq. 0) then

c        Rearrange the elements t(1) to t(n) to form a heap.

         do 20 k = 2, n
            ddum  = t(k)
            indxin = iorder(k)

c           Add ddum to the heap.
            i = k
   10       continue
            if (i.gt.1) then
               j = i/2
               if (ddum .lt. t(j)) then
                  t(i) = t(j)
                  iorder(i) = iorder(j)
                  i = j
                  goto 10 
               endif  
            endif  
            t(i) = ddum
            iorder(i) = indxin
   20    continue
      endif
 
c     Assign to 'out' the value of t(1), the least member of the heap,
c        and rearrange the remaining members to form a heap as
c        elements 1 to n-1 of t.
 
      if (n .gt. 1) then
         i = 1
         out = t(1)
         indxou = iorder(1)
         ddum  = t(n)
         indxin  = iorder(n)

c        Restore the heap 
   30    continue
         j = i+i
         if (j .le. n-1) then
            if (t(j+1) .lt. t(j)) j = j+1
            if (t(j) .lt. ddum ) then
               t(i) = t(j)
               iorder(i) = iorder(j)
               i = j
               goto 30
            endif 
         endif 
         t(i) = ddum
         iorder(i) = indxin
 
c     Put the least member in t(n). 

         t(n) = out
         iorder(n) = indxou
      endif 

      return

      end

c====================== The end of hpsolb ==============================

      subroutine lnsrlb(n, l, u, nbd, x, f, fold, gd, gdold, g, d, r, t,
     +                  z, stp, dnorm, dtd, xstep, stpmx, iter, ifun,
     +                  iback, nfgv, info, task, boxed, cnstnd, csave,
     +                  isave, dsave)

      character*60     task, csave
      logical          boxed, cnstnd
      integer          n, iter, ifun, iback, nfgv, info,
     +                 nbd(n), isave(2)
      double precision f, fold, gd, gdold, stp, dnorm, dtd, xstep,
     +                 stpmx, x(n), l(n), u(n), g(n), d(n), r(n), t(n),
     +                 z(n), dsave(13)
c     **********
c
c     Subroutine lnsrlb
c
c     This subroutine calls subroutine dcsrch from the Minpack2 library
c       to perform the line search.  Subroutine dscrch is safeguarded so
c       that all trial points lie within the feasible region.
c
c     Subprograms called:
c
c       Minpack2 Library ... dcsrch.
c
c       Linpack ... dtrsl, ddot.
c
c
c                           *  *  *
c
c     NEOS, November 1994. (Latest revision June 1996.)
c     Optimization Technology Center.
c     Argonne National Laboratory and Northwestern University.
c     Written by
c                        Ciyou Zhu
c     in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
c
c
c     **********

      integer          i
      double           precision ddot,a1,a2
      double precision one,zero,big
      parameter        (one=1.0d0,zero=0.0d0,big=1.0d+10)
      double precision ftol,gtol,xtol
      parameter        (ftol=1.0d-3,gtol=0.9d0,xtol=0.1d0)

      if (task(1:5) .eq. 'FG_LN') goto 556

      dtd = ddot(n,d,1,d,1)
      dnorm = sqrt(dtd)

c     Determine the maximum step length.

      stpmx = big
      if (cnstnd) then
         if (iter .eq. 0) then
            stpmx = one
         else
            do 43 i = 1, n
               a1 = d(i)
               if (nbd(i) .ne. 0) then
                  if (a1 .lt. zero .and. nbd(i) .le. 2) then
                     a2 = l(i) - x(i)
                     if (a2 .ge. zero) then
                        stpmx = zero
                     else if (a1*stpmx .lt. a2) then
                        stpmx = a2/a1
                     endif
                  else if (a1 .gt. zero .and. nbd(i) .ge. 2) then
                     a2 = u(i) - x(i)
                     if (a2 .le. zero) then
                        stpmx = zero
                     else if (a1*stpmx .gt. a2) then
                        stpmx = a2/a1
                     endif
                  endif
               endif
  43        continue
         endif
      endif
 
      if (iter .eq. 0 .and. .not. boxed) then
         stp = min(one/dnorm, stpmx)
      else
         stp = one
      endif 

      call dcopy(n,x,1,t,1)
      call dcopy(n,g,1,r,1)
      fold = f
      ifun = 0
      iback = 0
      csave = 'START'
 556  continue
      gd = ddot(n,g,1,d,1)
      if (ifun .eq. 0) then
         gdold=gd
         if (gd .ge. zero) then
c                               the directional derivative >=0.
c                               Line search is impossible.
            write(6,*)' ascent direction in projection gd = ', gd
            info = -4
            return
         endif
      endif

      call dcsrch(f,gd,stp,ftol,gtol,xtol,zero,stpmx,csave,isave,dsave)

      xstep = stp*dnorm
      if (csave(1:4) .ne. 'CONV' .and. csave(1:4) .ne. 'WARN') then
         task = 'FG_LNSRCH'
         ifun = ifun + 1
         nfgv = nfgv + 1
         iback = ifun - 1 
         if (stp .eq. one) then
            call dcopy(n,z,1,x,1)
         else
            do 41 i = 1, n
               x(i) = stp*d(i) + t(i)
  41        continue
         endif
      else
         task = 'NEW_X'
      endif

      return

      end

c======================= The end of lnsrlb =============================

      subroutine matupd(n, m, ws, wy, sy, ss, d, r, itail, 
     +                  iupdat, col, head, theta, rr, dr, stp, dtd)
 
      integer          n, m, itail, iupdat, col, head
      double precision theta, rr, dr, stp, dtd, d(n), r(n), 
     +                 ws(n, m), wy(n, m), sy(m, m), ss(m, m)

c     ************
c
c     Subroutine matupd
c
c       This subroutine updates matrices WS and WY, and forms the
c         middle matrix in B.
c
c     Subprograms called:
c
c       Linpack ... dcopy, ddot.
c
c
c                           *  *  *
c
c     NEOS, November 1994. (Latest revision June 1996.)
c     Optimization Technology Center.
c     Argonne National Laboratory and Northwestern University.
c     Written by
c                        Ciyou Zhu
c     in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
c
c
c     ************
 
      integer          j,pointr
      double precision ddot
      double precision one
      parameter        (one=1.0d0)

c     Set pointers for matrices WS and WY.
 
      if (iupdat .le. m) then
         col = iupdat
         itail = mod(head+iupdat-2,m) + 1
      else
         itail = mod(itail,m) + 1
         head = mod(head,m) + 1
      endif
 
c     Update matrices WS and WY.

      call dcopy(n,d,1,ws(1,itail),1)
      call dcopy(n,r,1,wy(1,itail),1)
 
c     Set theta=yy/ys.
 
      theta = rr/dr
 
c     Form the middle matrix in B.
 
c        update the upper triangle of SS,
c                                         and the lower triangle of SY:
      if (iupdat .gt. m) then
c                              move old information
         do 50 j = 1, col - 1
            call dcopy(j,ss(2,j+1),1,ss(1,j),1)
            call dcopy(col-j,sy(j+1,j+1),1,sy(j,j),1)
  50     continue
      endif
c        add new information: the last row of SY
c                                             and the last column of SS:
      pointr = head
      do 51 j = 1, col - 1
         sy(col,j) = ddot(n,d,1,wy(1,pointr),1)
         ss(j,col) = ddot(n,ws(1,pointr),1,d,1)
         pointr = mod(pointr,m) + 1
  51  continue
      if (stp .eq. one) then
         ss(col,col) = dtd
      else
         ss(col,col) = stp*stp*dtd
      endif
      sy(col,col) = dr
 
      return

      end

c======================= The end of matupd =============================

      subroutine prn1lb(n, m, l, u, x, iprint, itfile, epsmch)
 
      integer n, m, iprint, itfile
      double precision epsmch, x(n), l(n), u(n)

c     ************
c
c     Subroutine prn1lb
c
c     This subroutine prints the input data, initial point, upper and
c       lower bounds of each variable, machine precision, as well as 
c       the headings of the output.
c
c
c                           *  *  *
c
c     NEOS, November 1994. (Latest revision June 1996.)
c     Optimization Technology Center.
c     Argonne National Laboratory and Northwestern University.
c     Written by
c                        Ciyou Zhu
c     in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
c
c
c     ************

      integer i

      if (iprint .ge. 0) then
         write (6,7001) epsmch
         write (6,*) 'N = ',n,'    M = ',m
         if (iprint .ge. 1) then
            write (itfile,2001) epsmch
            write (itfile,*)'N = ',n,'    M = ',m
            write (itfile,9001)
            if (iprint .gt. 100) then
               write (6,1004) 'L =',(l(i),i = 1,n)
               write (6,1004) 'X0 =',(x(i),i = 1,n)
               write (6,1004) 'U =',(u(i),i = 1,n)
            endif 
         endif
      endif 

 1004 format (/,a4, 1p, 6(1x,d11.4),/,(4x,1p,6(1x,d11.4)))
 2001 format ('RUNNING THE L-BFGS-B CODE',/,/,
     + 'it    = iteration number',/,
     + 'nf    = number of function evaluations',/,
     + 'nseg  = number of segments explored during the Cauchy search',/,
     + 'nact  = number of active bounds at the generalized Cauchy point'
     + ,/,
     + 'sub   = manner in which the subspace minimization terminated:'
     + ,/,'        con = converged, bnd = a bound was reached',/,
     + 'itls  = number of iterations performed in the line search',/,
     + 'stepl = step length used',/,
     + 'tstep = norm of the displacement (total step)',/,
     + 'projg = norm of the projected gradient',/,
     + 'f     = function value',/,/,
     + '           * * *',/,/,
     + 'Machine precision =',1p,d10.3)
 7001 format ('RUNNING THE L-BFGS-B CODE',/,/,
     + '           * * *',/,/,
     + 'Machine precision =',1p,d10.3)
 9001 format (/,3x,'it',3x,'nf',2x,'nseg',2x,'nact',2x,'sub',2x,'itls',
     +        2x,'stepl',4x,'tstep',5x,'projg',8x,'f')

      return

      end

c======================= The end of prn1lb =============================

      subroutine prn2lb(n, x, f, g, iprint, itfile, iter, nfgv, nact, 
     +                  sbgnrm, nseg, word, iword, iback, stp, xstep)
 
      character*3      word
      integer          n, iprint, itfile, iter, nfgv, nact, nseg,
     +                 iword, iback
      double precision f, sbgnrm, stp, xstep, x(n), g(n)

c     ************
c
c     Subroutine prn2lb
c
c     This subroutine prints out new information after a successful
c       line search. 
c
c
c                           *  *  *
c
c     NEOS, November 1994. (Latest revision June 1996.)
c     Optimization Technology Center.
c     Argonne National Laboratory and Northwestern University.
c     Written by
c                        Ciyou Zhu
c     in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
c
c
c     ************

      integer i,imod

c           'word' records the status of subspace solutions.
      if (iword .eq. 0) then
c                            the subspace minimization converged.
         word = 'con'
      else if (iword .eq. 1) then
c                          the subspace minimization stopped at a bound.
         word = 'bnd'
      else if (iword .eq. 5) then
c                             the truncated Newton step has been used.
         word = 'TNT'
      else
         word = '---'
      endif
      if (iprint .ge. 99) then
         write (6,*) 'LINE SEARCH',iback,' times; norm of step = ',xstep
         write (6,2001) iter,f,sbgnrm
         if (iprint .gt. 100) then      
            write (6,1004) 'X =',(x(i), i = 1, n)
            write (6,1004) 'G =',(g(i), i = 1, n)
         endif
      else if (iprint .gt. 0) then 
         imod = mod(iter,iprint)
         if (imod .eq. 0) write (6,2001) iter,f,sbgnrm
      endif
      if (iprint .ge. 1) write (itfile,3001)
     +          iter,nfgv,nseg,nact,word,iback,stp,xstep,sbgnrm,f

 1004 format (/,a4, 1p, 6(1x,d11.4),/,(4x,1p,6(1x,d11.4)))
 2001 format
     +  (/,'At iterate',i5,4x,'f= ',1p,d12.5,4x,'|proj g|= ',1p,d12.5)
 3001 format(2(1x,i4),2(1x,i5),2x,a3,1x,i4,1p,2(2x,d7.1),1p,2(1x,d10.3))

      return

      end

c======================= The end of prn2lb =============================

      subroutine prn3lb(n, x, f, task, iprint, info, itfile, 
     +                  iter, nfgv, nintol, nskip, nact, sbgnrm, 
     +                  time, nseg, word, iback, stp, xstep, k, 
     +                  cachyt, sbtime, lnscht)
 
      character*60     task
      character*3      word
      integer          n, iprint, info, itfile, iter, nfgv, nintol,
     +                 nskip, nact, nseg, iback, k
      double precision f, sbgnrm, time, stp, xstep, cachyt, sbtime,
     +                 lnscht, x(n)

c     ************
c
c     Subroutine prn3lb
c
c     This subroutine prints out information when either a built-in
c       convergence test is satisfied or when an error message is
c       generated.
c       
c
c                           *  *  *
c
c     NEOS, November 1994. (Latest revision June 1996.)
c     Optimization Technology Center.
c     Argonne National Laboratory and Northwestern University.
c     Written by
c                        Ciyou Zhu
c     in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
c
c
c     ************

      integer i

      if (task(1:5) .eq. 'ERROR') goto 999

      if (iprint .ge. 0) then
         write (6,3003)
         write (6,3004)
         write(6,3005) n,iter,nfgv,nintol,nskip,nact,sbgnrm,f
         if (iprint .ge. 100) then
            write (6,1004) 'X =',(x(i),i = 1,n)
         endif  
         if (iprint .ge. 1) write (6,*) ' F =',f
      endif 
 999  continue
      if (iprint .ge. 0) then
         write (6,3009) task
         if (info .ne. 0) then
            if (info .eq. -1) write (6,9011)
            if (info .eq. -2) write (6,9012)
            if (info .eq. -3) write (6,9013)
            if (info .eq. -4) write (6,9014)
            if (info .eq. -5) write (6,9015)
            if (info .eq. -6) write (6,*)' Input nbd(',k,') is invalid.'
            if (info .eq. -7) 
     +      write (6,*)' l(',k,') > u(',k,').  No feasible solution.'
            if (info .eq. -8) write (6,9018)
            if (info .eq. -9) write (6,9019)
         endif
         if (iprint .ge. 1) write (6,3007) cachyt,sbtime,lnscht
         write (6,3008) time
         if (iprint .ge. 1) then
            if (info .eq. -4 .or. info .eq. -9) then
               write (itfile,3002)
     +             iter,nfgv,nseg,nact,word,iback,stp,xstep
            endif
            write (itfile,3009) task
            if (info .ne. 0) then
               if (info .eq. -1) write (itfile,9011)
               if (info .eq. -2) write (itfile,9012)
               if (info .eq. -3) write (itfile,9013)
               if (info .eq. -4) write (itfile,9014)
               if (info .eq. -5) write (itfile,9015)
               if (info .eq. -8) write (itfile,9018)
               if (info .eq. -9) write (itfile,9019)
            endif
            write (itfile,3008) time
         endif
      endif

 1004 format (/,a4, 1p, 6(1x,d11.4),/,(4x,1p,6(1x,d11.4)))
 3002 format(2(1x,i4),2(1x,i5),2x,a3,1x,i4,1p,2(2x,d7.1),6x,'-',10x,'-')
 3003 format (/,
     + '           * * *',/,/,
     + 'Tit   = total number of iterations',/,
     + 'Tnf   = total number of function evaluations',/,
     + 'Tnint = total number of segments explored during',
     +           ' Cauchy searches',/,
     + 'Skip  = number of BFGS updates skipped',/,
     + 'Nact  = number of active bounds at final generalized',
     +          ' Cauchy point',/,
     + 'Projg = norm of the final projected gradient',/,
     + 'F     = final function value',/,/,
     + '           * * *')
 3004 format (/,3x,'N',4x,'Tit',5x,'Tnf',2x,'Tnint',2x,
     +       'Skip',2x,'Nact',5x,'Projg',8x,'F')
 3005 format (i5,2(1x,i6),(1x,i6),(2x,i4),(1x,i5),1p,2(2x,d10.3))
 3007 format (/,' Cauchy                time',1p,e10.3,' seconds.',/ 
     +        ' Subspace minimization time',1p,e10.3,' seconds.',/
     +        ' Line search           time',1p,e10.3,' seconds.')
 3008 format (/,' Total User time',1p,e10.3,' seconds.',/)
 3009 format (/,a60)
 9011 format (/,
     +' Matrix in 1st Cholesky factorization in formk is not Pos. Def.')
 9012 format (/,
     +' Matrix in 2st Cholesky factorization in formk is not Pos. Def.')
 9013 format (/,
     +' Matrix in the Cholesky factorization in formt is not Pos. Def.')
 9014 format (/,
     +' Derivative >= 0, backtracking line search impossible.',/,
     +'   Previous x, f and g restored.',/,
     +' Possible causes: 1 error in function or gradient evaluation;',/,
     +'                  2 rounding errors dominate computation.')
 9015 format (/,
     +' Warning:  more than 10 function and gradient',/,
     +'   evaluations in the last line search.  Termination',/,
     +'   may possibly be caused by a bad search direction.')
 9018 format (/,' The triangular system is singular.')
 9019 format (/,
     +' Line search cannot locate an adequate point after 20 function',/
     +,'  and gradient evaluations.  Previous x, f and g restored.',/,
     +' Possible causes: 1 error in function or gradient evaluation;',/,
     +'                  2 rounding error dominate computation.')

      return

      end

c======================= The end of prn3lb =============================

      subroutine projgr(n, l, u, nbd, x, g, sbgnrm)

      integer          n, nbd(n)
      double precision sbgnrm, x(n), l(n), u(n), g(n)

c     ************
c
c     Subroutine projgr
c
c     This subroutine computes the infinity norm of the projected
c       gradient.
c
c
c                           *  *  *
c
c     NEOS, November 1994. (Latest revision June 1996.)
c     Optimization Technology Center.
c     Argonne National Laboratory and Northwestern University.
c     Written by
c                        Ciyou Zhu
c     in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
c
c
c     ************

      integer i
      double precision gi
      double precision one,zero
      parameter        (one=1.0d0,zero=0.0d0)

      sbgnrm = zero
      do 15 i = 1, n
        gi = g(i)
        if (nbd(i) .ne. 0) then
           if (gi .lt. zero) then
              if (nbd(i) .ge. 2) gi = max((x(i)-u(i)),gi)
           else
              if (nbd(i) .le. 2) gi = min((x(i)-l(i)),gi)
           endif
        endif
        sbgnrm = max(sbgnrm,abs(gi))
  15  continue

      return

      end

c======================= The end of projgr =============================

      subroutine subsm ( n, m, nsub, ind, l, u, nbd, x, d, xp, ws, wy,
     +                   theta, xx, gg,
     +                   col, head, iword, wv, wn, iprint, info )
      implicit none
      integer          n, m, nsub, col, head, iword, iprint, info, 
     +                 ind(nsub), nbd(n)
      double precision theta, 
     +                 l(n), u(n), x(n), d(n), xp(n), xx(n), gg(n),
     +                 ws(n, m), wy(n, m), 
     +                 wv(2*m), wn(2*m, 2*m)

c     **********************************************************************
c
c     This routine contains the major changes in the updated version.
c     The changes are described in the accompanying paper
c
c      Jose Luis Morales, Jorge Nocedal
c      "Remark On Algorithm 788: L-BFGS-B: Fortran Subroutines for Large-Scale
c       Bound Constrained Optimization". Decemmber 27, 2010.
c
c             J.L. Morales  Departamento de Matematicas, 
c                           Instituto Tecnologico Autonomo de Mexico
c                           Mexico D.F.
c
c             J, Nocedal    Department of Electrical Engineering and
c                           Computer Science.
c                           Northwestern University. Evanston, IL. USA
c
c                           January 17, 2011
c
c      **********************************************************************
c                           
c
c     Subroutine subsm
c
c     Given xcp, l, u, r, an index set that specifies
c       the active set at xcp, and an l-BFGS matrix B 
c       (in terms of WY, WS, SY, WT, head, col, and theta), 
c       this subroutine computes an approximate solution
c       of the subspace problem
c
c       (P)   min Q(x) = r'(x-xcp) + 1/2 (x-xcp)' B (x-xcp)
c
c             subject to l<=x<=u
c                       x_i=xcp_i for all i in A(xcp)
c                     
c       along the subspace unconstrained Newton direction 
c       
c          d = -(Z'BZ)^(-1) r.
c
c       The formula for the Newton direction, given the L-BFGS matrix
c       and the Sherman-Morrison formula, is
c
c          d = (1/theta)r + (1/theta*2) Z'WK^(-1)W'Z r.
c 
c       where
c                 K = [-D -Y'ZZ'Y/theta     L_a'-R_z'  ]
c                     [L_a -R_z           theta*S'AA'S ]
c
c     Note that this procedure for computing d differs 
c     from that described in [1]. One can show that the matrix K is
c     equal to the matrix M^[-1]N in that paper.
c
c     n is an integer variable.
c       On entry n is the dimension of the problem.
c       On exit n is unchanged.
c
c     m is an integer variable.
c       On entry m is the maximum number of variable metric corrections
c         used to define the limited memory matrix.
c       On exit m is unchanged.
c
c     nsub is an integer variable.
c       On entry nsub is the number of free variables.
c       On exit nsub is unchanged.
c
c     ind is an integer array of dimension nsub.
c       On entry ind specifies the coordinate indices of free variables.
c       On exit ind is unchanged.
c
c     l is a double precision array of dimension n.
c       On entry l is the lower bound of x.
c       On exit l is unchanged.
c
c     u is a double precision array of dimension n.
c       On entry u is the upper bound of x.
c       On exit u is unchanged.
c
c     nbd is a integer array of dimension n.
c       On entry nbd represents the type of bounds imposed on the
c         variables, and must be specified as follows:
c         nbd(i)=0 if x(i) is unbounded,
c                1 if x(i) has only a lower bound,
c                2 if x(i) has both lower and upper bounds, and
c                3 if x(i) has only an upper bound.
c       On exit nbd is unchanged.
c
c     x is a double precision array of dimension n.
c       On entry x specifies the Cauchy point xcp. 
c       On exit x(i) is the minimizer of Q over the subspace of
c                                                        free variables. 
c
c     d is a double precision array of dimension n.
c       On entry d is the reduced gradient of Q at xcp.
c       On exit d is the Newton direction of Q. 
c
c    xp is a double precision array of dimension n.
c       used to safeguard the projected Newton direction 
c
c    xx is a double precision array of dimension n
c       On entry it holds the current iterate
c       On output it is unchanged

c    gg is a double precision array of dimension n
c       On entry it holds the gradient at the current iterate
c       On output it is unchanged
c
c     ws and wy are double precision arrays;
c     theta is a double precision variable;
c     col is an integer variable;
c     head is an integer variable.
c       On entry they store the information defining the
c                                          limited memory BFGS matrix:
c         ws(n,m) stores S, a set of s-vectors;
c         wy(n,m) stores Y, a set of y-vectors;
c         theta is the scaling factor specifying B_0 = theta I;
c         col is the number of variable metric corrections stored;
c         head is the location of the 1st s- (or y-) vector in S (or Y).
c       On exit they are unchanged.
c
c     iword is an integer variable.
c       On entry iword is unspecified.
c       On exit iword specifies the status of the subspace solution.
c         iword = 0 if the solution is in the box,
c                 1 if some bound is encountered.
c
c     wv is a double precision working array of dimension 2m.
c
c     wn is a double precision array of dimension 2m x 2m.
c       On entry the upper triangle of wn stores the LEL^T factorization
c         of the indefinite matrix
c
c              K = [-D -Y'ZZ'Y/theta     L_a'-R_z'  ]
c                  [L_a -R_z           theta*S'AA'S ]
c                                                    where E = [-I  0]
c                                                              [ 0  I]
c       On exit wn is unchanged.
c
c     iprint is an INTEGER variable that must be set by the user.
c       It controls the frequency and type of output generated:
c        iprint<0    no output is generated;
c        iprint=0    print only one line at the last iteration;
c        0<iprint<99 print also f and |proj g| every iprint iterations;
c        iprint=99   print details of every iteration except n-vectors;
c        iprint=100  print also the changes of active set and final x;
c        iprint>100  print details of every iteration including x and g;
c       When iprint > 0, the file iterate.dat will be created to
c                        summarize the iteration.
c
c     info is an integer variable.
c       On entry info is unspecified.
c       On exit info = 0       for normal return,
c                    = nonzero for abnormal return 
c                                  when the matrix K is ill-conditioned.
c
c     Subprograms called:
c
c       Linpack dtrsl.
c
c
c     References:
c
c       [1] R. H. Byrd, P. Lu, J. Nocedal and C. Zhu, ``A limited
c       memory algorithm for bound constrained optimization'',
c       SIAM J. Scientific Computing 16 (1995), no. 5, pp. 1190--1208.
c
c
c
c                           *  *  *
c
c     NEOS, November 1994. (Latest revision June 1996.)
c     Optimization Technology Center.
c     Argonne National Laboratory and Northwestern University.
c     Written by
c                        Ciyou Zhu
c     in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.
c
c
c     ************

      integer          pointr,m2,col2,ibd,jy,js,i,j,k
      double precision alpha, xk, dk, temp1, temp2 
      double precision one,zero
      parameter        (one=1.0d0,zero=0.0d0)
c
      double precision dd_p

      if (nsub .le. 0) return
      if (iprint .ge. 99) write (6,1001)

c     Compute wv = W'Zd.

      pointr = head 
      do 20 i = 1, col
         temp1 = zero
         temp2 = zero
         do 10 j = 1, nsub
            k = ind(j)
            temp1 = temp1 + wy(k,pointr)*d(j)
            temp2 = temp2 + ws(k,pointr)*d(j)
  10     continue
         wv(i) = temp1
         wv(col + i) = theta*temp2
         pointr = mod(pointr,m) + 1
  20  continue
 
c     Compute wv:=K^(-1)wv.

      m2 = 2*m
      col2 = 2*col
      call dtrsl(wn,m2,col2,wv,11,info)
      if (info .ne. 0) return
      do 25 i = 1, col
         wv(i) = -wv(i)
  25     continue
      call dtrsl(wn,m2,col2,wv,01,info)
      if (info .ne. 0) return
 
c     Compute d = (1/theta)d + (1/theta**2)Z'W wv.
 
      pointr = head
      do 40 jy = 1, col
         js = col + jy
         do 30 i = 1, nsub
            k = ind(i)
            d(i) = d(i) + wy(k,pointr)*wv(jy)/theta     
     +                  + ws(k,pointr)*wv(js)
  30     continue
         pointr = mod(pointr,m) + 1
  40  continue

      call dscal( nsub, one/theta, d, 1 )
c 
c-----------------------------------------------------------------
c     Let us try the projection, d is the Newton direction

      iword = 0

      call dcopy ( n, x, 1, xp, 1 )
c
      do 50 i=1, nsub
         k  = ind(i)
         dk = d(i)
         xk = x(k)
         if ( nbd(k) .ne. 0 ) then
c
            if ( nbd(k).eq.1 ) then          ! lower bounds only
               x(k) = max( l(k), xk + dk )
               if ( x(k).eq.l(k) ) iword = 1
            else 
c     
               if ( nbd(k).eq.2 ) then       ! upper and lower bounds
                  xk   = max( l(k), xk + dk ) 
                  x(k) = min( u(k), xk )
                  if ( x(k).eq.l(k) .or. x(k).eq.u(k) ) iword = 1
               else
c
                  if ( nbd(k).eq.3 ) then    ! upper bounds only
                     x(k) = min( u(k), xk + dk )
                     if ( x(k).eq.u(k) ) iword = 1
                  end if 
               end if
            end if
c            
         else                                ! free variables
            x(k) = xk + dk
         end if 
 50   continue
c
      if ( iword.eq.0 ) then
         go to 911
      end if
c
c     check sign of the directional derivative
c
      dd_p = zero
      do 55 i=1, n
         dd_p  = dd_p + (x(i) - xx(i))*gg(i)
 55   continue
      if ( dd_p .gt.zero ) then
         call dcopy( n, xp, 1, x, 1 )
         write(6,*) ' Positive dir derivative in projection '
         write(6,*) ' Using the backtracking step '
      else
         go to 911
      endif
c
c-----------------------------------------------------------------
c
      alpha = one
      temp1 = alpha
      ibd   = 0 
      do 60 i = 1, nsub
         k = ind(i)
         dk = d(i)
         if (nbd(k) .ne. 0) then
            if (dk .lt. zero .and. nbd(k) .le. 2) then
               temp2 = l(k) - x(k)
               if (temp2 .ge. zero) then
                  temp1 = zero
               else if (dk*alpha .lt. temp2) then
                  temp1 = temp2/dk
               endif
            else if (dk .gt. zero .and. nbd(k) .ge. 2) then
               temp2 = u(k) - x(k)
               if (temp2 .le. zero) then
                  temp1 = zero
               else if (dk*alpha .gt. temp2) then
                  temp1 = temp2/dk
               endif
            endif
            if (temp1 .lt. alpha) then
               alpha = temp1
               ibd = i
            endif
         endif
 60   continue
      
      if (alpha .lt. one) then
         dk = d(ibd)
         k = ind(ibd)
         if (dk .gt. zero) then
            x(k) = u(k)
            d(ibd) = zero
         else if (dk .lt. zero) then
            x(k) = l(k)
            d(ibd) = zero
         endif
      endif
      do 70 i = 1, nsub
         k    = ind(i)
         x(k) = x(k) + alpha*d(i)
 70   continue
cccccc
 911  continue

      if (iprint .ge. 99) write (6,1004)

 1001 format (/,'----------------SUBSM entered-----------------',/)
 1004 format (/,'----------------exit SUBSM --------------------',/)

      return

      end
c====================== The end of subsm ===============================

      subroutine dcsrch(f,g,stp,ftol,gtol,xtol,stpmin,stpmax,
     +                  task,isave,dsave)
      character*(*) task
      integer isave(2)
      double precision f,g,stp,ftol,gtol,xtol,stpmin,stpmax
      double precision dsave(13)
c     **********
c
c     Subroutine dcsrch
c
c     This subroutine finds a step that satisfies a sufficient
c     decrease condition and a curvature condition.
c
c     Each call of the subroutine updates an interval with 
c     endpoints stx and sty. The interval is initially chosen 
c     so that it contains a minimizer of the modified function
c
c           psi(stp) = f(stp) - f(0) - ftol*stp*f'(0).
c
c     If psi(stp) <= 0 and f'(stp) >= 0 for some step, then the
c     interval is chosen so that it contains a minimizer of f. 
c
c     The algorithm is designed to find a step that satisfies 
c     the sufficient decrease condition 
c
c           f(stp) <= f(0) + ftol*stp*f'(0),
c
c     and the curvature condition
c
c           abs(f'(stp)) <= gtol*abs(f'(0)).
c
c     If ftol is less than gtol and if, for example, the function
c     is bounded below, then there is always a step which satisfies
c     both conditions. 
c
c     If no step can be found that satisfies both conditions, then 
c     the algorithm stops with a warning. In this case stp only 
c     satisfies the sufficient decrease condition.
c
c     A typical invocation of dcsrch has the following outline:
c
c     task = 'START'
c  10 continue
c        call dcsrch( ... )
c        if (task .eq. 'FG') then
c           Evaluate the function and the gradient at stp 
c           goto 10
c           end if
c
c     NOTE: The user must no alter work arrays between calls.
c
c     The subroutine statement is
c
c        subroutine dcsrch(f,g,stp,ftol,gtol,xtol,stpmin,stpmax,
c                          task,isave,dsave)
c     where
c
c       f is a double precision variable.
c         On initial entry f is the value of the function at 0.
c            On subsequent entries f is the value of the 
c            function at stp.
c         On exit f is the value of the function at stp.
c
c       g is a double precision variable.
c         On initial entry g is the derivative of the function at 0.
c            On subsequent entries g is the derivative of the 
c            function at stp.
c         On exit g is the derivative of the function at stp.
c
c       stp is a double precision variable. 
c         On entry stp is the current estimate of a satisfactory 
c            step. On initial entry, a positive initial estimate 
c            must be provided. 
c         On exit stp is the current estimate of a satisfactory step
c            if task = 'FG'. If task = 'CONV' then stp satisfies
c            the sufficient decrease and curvature condition.
c
c       ftol is a double precision variable.
c         On entry ftol specifies a nonnegative tolerance for the 
c            sufficient decrease condition.
c         On exit ftol is unchanged.
c
c       gtol is a double precision variable.
c         On entry gtol specifies a nonnegative tolerance for the 
c            curvature condition. 
c         On exit gtol is unchanged.
c
c       xtol is a double precision variable.
c         On entry xtol specifies a nonnegative relative tolerance
c            for an acceptable step. The subroutine exits with a
c            warning if the relative difference between sty and stx
c            is less than xtol.
c         On exit xtol is unchanged.
c
c       stpmin is a double precision variable.
c         On entry stpmin is a nonnegative lower bound for the step.
c         On exit stpmin is unchanged.
c
c       stpmax is a double precision variable.
c         On entry stpmax is a nonnegative upper bound for the step.
c         On exit stpmax is unchanged.
c
c       task is a character variable of length at least 60.
c         On initial entry task must be set to 'START'.
c         On exit task indicates the required action:
c
c            If task(1:2) = 'FG' then evaluate the function and 
c            derivative at stp and call dcsrch again.
c
c            If task(1:4) = 'CONV' then the search is successful.
c
c            If task(1:4) = 'WARN' then the subroutine is not able
c            to satisfy the convergence conditions. The exit value of
c            stp contains the best point found during the search.
c
c            If task(1:5) = 'ERROR' then there is an error in the
c            input arguments.
c
c         On exit with convergence, a warning or an error, the
c            variable task contains additional information.
c
c       isave is an integer work array of dimension 2.
c         
c       dsave is a double precision work array of dimension 13.
c
c     Subprograms called
c
c       MINPACK-2 ... dcstep
c
c     MINPACK-1 Project. June 1983.
c     Argonne National Laboratory. 
c     Jorge J. More' and David J. Thuente.
c
c     MINPACK-2 Project. October 1993.
c     Argonne National Laboratory and University of Minnesota. 
c     Brett M. Averick, Richard G. Carter, and Jorge J. More'. 
c
c     **********
      double precision zero,p5,p66
      parameter(zero=0.0d0,p5=0.5d0,p66=0.66d0)
      double precision xtrapl,xtrapu
      parameter(xtrapl=1.1d0,xtrapu=4.0d0)

      logical brackt
      integer stage
      double precision finit,ftest,fm,fx,fxm,fy,fym,ginit,gtest,
     +       gm,gx,gxm,gy,gym,stx,sty,stmin,stmax,width,width1

c     Initialization block.

      if (task(1:5) .eq. 'START') then

c        Check the input arguments for errors.

         if (stp .lt. stpmin) task = 'ERROR: STP .LT. STPMIN'
         if (stp .gt. stpmax) task = 'ERROR: STP .GT. STPMAX'
         if (g .ge. zero) task = 'ERROR: INITIAL G .GE. ZERO'
         if (ftol .lt. zero) task = 'ERROR: FTOL .LT. ZERO'
         if (gtol .lt. zero) task = 'ERROR: GTOL .LT. ZERO'
         if (xtol .lt. zero) task = 'ERROR: XTOL .LT. ZERO'
         if (stpmin .lt. zero) task = 'ERROR: STPMIN .LT. ZERO'
         if (stpmax .lt. stpmin) task = 'ERROR: STPMAX .LT. STPMIN'

c        Exit if there are errors on input.

         if (task(1:5) .eq. 'ERROR') return

c        Initialize local variables.

         brackt = .false.
         stage = 1
         finit = f
         ginit = g
         gtest = ftol*ginit
         width = stpmax - stpmin
         width1 = width/p5

c        The variables stx, fx, gx contain the values of the step, 
c        function, and derivative at the best step. 
c        The variables sty, fy, gy contain the value of the step, 
c        function, and derivative at sty.
c        The variables stp, f, g contain the values of the step, 
c        function, and derivative at stp.

         stx = zero
         fx = finit
         gx = ginit
         sty = zero
         fy = finit
         gy = ginit
         stmin = zero
         stmax = stp + xtrapu*stp
         task = 'FG'

         goto 1000

      else

c        Restore local variables.

         if (isave(1) .eq. 1) then
            brackt = .true.
         else
            brackt = .false.
         endif
         stage = isave(2) 
         ginit = dsave(1) 
         gtest = dsave(2) 
         gx = dsave(3) 
         gy = dsave(4) 
         finit = dsave(5) 
         fx = dsave(6) 
         fy = dsave(7) 
         stx = dsave(8) 
         sty = dsave(9) 
         stmin = dsave(10) 
         stmax = dsave(11) 
         width = dsave(12) 
         width1 = dsave(13) 

      endif

c     If psi(stp) <= 0 and f'(stp) >= 0 for some step, then the
c     algorithm enters the second stage.

      ftest = finit + stp*gtest
      if (stage .eq. 1 .and. f .le. ftest .and. g .ge. zero) 
     +   stage = 2

c     Test for warnings.

      if (brackt .and. (stp .le. stmin .or. stp .ge. stmax))
     +   task = 'WARNING: ROUNDING ERRORS PREVENT PROGRESS'
      if (brackt .and. stmax - stmin .le. xtol*stmax) 
     +   task = 'WARNING: XTOL TEST SATISFIED'
      if (stp .eq. stpmax .and. f .le. ftest .and. g .le. gtest) 
     +   task = 'WARNING: STP = STPMAX'
      if (stp .eq. stpmin .and. (f .gt. ftest .or. g .ge. gtest)) 
     +   task = 'WARNING: STP = STPMIN'

c     Test for convergence.

      if (f .le. ftest .and. abs(g) .le. gtol*(-ginit)) 
     +   task = 'CONVERGENCE'

c     Test for termination.

      if (task(1:4) .eq. 'WARN' .or. task(1:4) .eq. 'CONV') goto 1000

c     A modified function is used to predict the step during the
c     first stage if a lower function value has been obtained but 
c     the decrease is not sufficient.

      if (stage .eq. 1 .and. f .le. fx .and. f .gt. ftest) then

c        Define the modified function and derivative values.

         fm = f - stp*gtest
         fxm = fx - stx*gtest
         fym = fy - sty*gtest
         gm = g - gtest
         gxm = gx - gtest
         gym = gy - gtest

c        Call dcstep to update stx, sty, and to compute the new step.

         call dcstep(stx,fxm,gxm,sty,fym,gym,stp,fm,gm,
     +               brackt,stmin,stmax)

c        Reset the function and derivative values for f.

         fx = fxm + stx*gtest
         fy = fym + sty*gtest
         gx = gxm + gtest
         gy = gym + gtest

      else

c       Call dcstep to update stx, sty, and to compute the new step.

        call dcstep(stx,fx,gx,sty,fy,gy,stp,f,g,
     +              brackt,stmin,stmax)

      endif

c     Decide if a bisection step is needed.

      if (brackt) then
         if (abs(sty-stx) .ge. p66*width1) stp = stx + p5*(sty - stx)
         width1 = width
         width = abs(sty-stx)
      endif

c     Set the minimum and maximum steps allowed for stp.

      if (brackt) then
         stmin = min(stx,sty)
         stmax = max(stx,sty)
      else
         stmin = stp + xtrapl*(stp - stx)
         stmax = stp + xtrapu*(stp - stx)
      endif
 
c     Force the step to be within the bounds stpmax and stpmin.
 
      stp = max(stp,stpmin)
      stp = min(stp,stpmax)

c     If further progress is not possible, let stp be the best
c     point obtained during the search.

      if (brackt .and. (stp .le. stmin .or. stp .ge. stmax)
     +   .or. (brackt .and. stmax-stmin .le. xtol*stmax)) stp = stx

c     Obtain another function and derivative.

      task = 'FG'

 1000 continue

c     Save local variables.

      if (brackt) then
         isave(1) = 1
      else
         isave(1) = 0
      endif
      isave(2) = stage
      dsave(1) =  ginit
      dsave(2) =  gtest
      dsave(3) =  gx
      dsave(4) =  gy
      dsave(5) =  finit
      dsave(6) =  fx
      dsave(7) =  fy
      dsave(8) =  stx
      dsave(9) =  sty
      dsave(10) = stmin
      dsave(11) = stmax
      dsave(12) = width
      dsave(13) = width1

      return
      end
      
c====================== The end of dcsrch ==============================

      subroutine dcstep(stx,fx,dx,sty,fy,dy,stp,fp,dp,brackt,
     +                  stpmin,stpmax)
      logical brackt
      double precision stx,fx,dx,sty,fy,dy,stp,fp,dp,stpmin,stpmax
c     **********
c
c     Subroutine dcstep
c
c     This subroutine computes a safeguarded step for a search
c     procedure and updates an interval that contains a step that
c     satisfies a sufficient decrease and a curvature condition.
c
c     The parameter stx contains the step with the least function
c     value. If brackt is set to .true. then a minimizer has
c     been bracketed in an interval with endpoints stx and sty.
c     The parameter stp contains the current step. 
c     The subroutine assumes that if brackt is set to .true. then
c
c           min(stx,sty) < stp < max(stx,sty),
c
c     and that the derivative at stx is negative in the direction 
c     of the step.
c
c     The subroutine statement is
c
c       subroutine dcstep(stx,fx,dx,sty,fy,dy,stp,fp,dp,brackt,
c                         stpmin,stpmax)
c
c     where
c
c       stx is a double precision variable.
c         On entry stx is the best step obtained so far and is an
c            endpoint of the interval that contains the minimizer. 
c         On exit stx is the updated best step.
c
c       fx is a double precision variable.
c         On entry fx is the function at stx.
c         On exit fx is the function at stx.
c
c       dx is a double precision variable.
c         On entry dx is the derivative of the function at 
c            stx. The derivative must be negative in the direction of 
c            the step, that is, dx and stp - stx must have opposite 
c            signs.
c         On exit dx is the derivative of the function at stx.
c
c       sty is a double precision variable.
c         On entry sty is the second endpoint of the interval that 
c            contains the minimizer.
c         On exit sty is the updated endpoint of the interval that 
c            contains the minimizer.
c
c       fy is a double precision variable.
c         On entry fy is the function at sty.
c         On exit fy is the function at sty.
c
c       dy is a double precision variable.
c         On entry dy is the derivative of the function at sty.
c         On exit dy is the derivative of the function at the exit sty.
c
c       stp is a double precision variable.
c         On entry stp is the current step. If brackt is set to .true.
c            then on input stp must be between stx and sty. 
c         On exit stp is a new trial step.
c
c       fp is a double precision variable.
c         On entry fp is the function at stp
c         On exit fp is unchanged.
c
c       dp is a double precision variable.
c         On entry dp is the the derivative of the function at stp.
c         On exit dp is unchanged.
c
c       brackt is an logical variable.
c         On entry brackt specifies if a minimizer has been bracketed.
c            Initially brackt must be set to .false.
c         On exit brackt specifies if a minimizer has been bracketed.
c            When a minimizer is bracketed brackt is set to .true.
c
c       stpmin is a double precision variable.
c         On entry stpmin is a lower bound for the step.
c         On exit stpmin is unchanged.
c
c       stpmax is a double precision variable.
c         On entry stpmax is an upper bound for the step.
c         On exit stpmax is unchanged.
c
c     MINPACK-1 Project. June 1983
c     Argonne National Laboratory. 
c     Jorge J. More' and David J. Thuente.
c
c     MINPACK-2 Project. October 1993.
c     Argonne National Laboratory and University of Minnesota. 
c     Brett M. Averick and Jorge J. More'.
c
c     **********
      double precision zero,p66,two,three
      parameter(zero=0.0d0,p66=0.66d0,two=2.0d0,three=3.0d0)
   
      double precision gamma,p,q,r,s,sgnd,stpc,stpf,stpq,theta

      sgnd = dp*(dx/abs(dx))

c     First case: A higher function value. The minimum is bracketed. 
c     If the cubic step is closer to stx than the quadratic step, the 
c     cubic step is taken, otherwise the average of the cubic and 
c     quadratic steps is taken.

      if (fp .gt. fx) then
         theta = three*(fx - fp)/(stp - stx) + dx + dp
         s = max(abs(theta),abs(dx),abs(dp))
         gamma = s*sqrt((theta/s)**2 - (dx/s)*(dp/s))
         if (stp .lt. stx) gamma = -gamma
         p = (gamma - dx) + theta
         q = ((gamma - dx) + gamma) + dp
         r = p/q
         stpc = stx + r*(stp - stx)
         stpq = stx + ((dx/((fx - fp)/(stp - stx) + dx))/two)*
     +                                                       (stp - stx)
         if (abs(stpc-stx) .lt. abs(stpq-stx)) then
            stpf = stpc
         else
            stpf = stpc + (stpq - stpc)/two
         endif
         brackt = .true.

c     Second case: A lower function value and derivatives of opposite 
c     sign. The minimum is bracketed. If the cubic step is farther from 
c     stp than the secant step, the cubic step is taken, otherwise the 
c     secant step is taken.

      else if (sgnd .lt. zero) then
         theta = three*(fx - fp)/(stp - stx) + dx + dp
         s = max(abs(theta),abs(dx),abs(dp))
         gamma = s*sqrt((theta/s)**2 - (dx/s)*(dp/s))
         if (stp .gt. stx) gamma = -gamma
         p = (gamma - dp) + theta
         q = ((gamma - dp) + gamma) + dx
         r = p/q
         stpc = stp + r*(stx - stp)
         stpq = stp + (dp/(dp - dx))*(stx - stp)
         if (abs(stpc-stp) .gt. abs(stpq-stp)) then
            stpf = stpc
         else
            stpf = stpq
         endif
         brackt = .true.

c     Third case: A lower function value, derivatives of the same sign,
c     and the magnitude of the derivative decreases.

      else if (abs(dp) .lt. abs(dx)) then

c        The cubic step is computed only if the cubic tends to infinity 
c        in the direction of the step or if the minimum of the cubic
c        is beyond stp. Otherwise the cubic step is defined to be the 
c        secant step.

         theta = three*(fx - fp)/(stp - stx) + dx + dp
         s = max(abs(theta),abs(dx),abs(dp))

c        The case gamma = 0 only arises if the cubic does not tend
c        to infinity in the direction of the step.

         gamma = s*sqrt(max(zero,(theta/s)**2-(dx/s)*(dp/s)))
         if (stp .gt. stx) gamma = -gamma
         p = (gamma - dp) + theta
         q = (gamma + (dx - dp)) + gamma
         r = p/q
         if (r .lt. zero .and. gamma .ne. zero) then
            stpc = stp + r*(stx - stp)
         else if (stp .gt. stx) then
            stpc = stpmax
         else
            stpc = stpmin
         endif
         stpq = stp + (dp/(dp - dx))*(stx - stp)

         if (brackt) then

c           A minimizer has been bracketed. If the cubic step is 
c           closer to stp than the secant step, the cubic step is 
c           taken, otherwise the secant step is taken.

            if (abs(stpc-stp) .lt. abs(stpq-stp)) then
               stpf = stpc
            else
               stpf = stpq
            endif
            if (stp .gt. stx) then
               stpf = min(stp+p66*(sty-stp),stpf)
            else
               stpf = max(stp+p66*(sty-stp),stpf)
            endif
         else

c           A minimizer has not been bracketed. If the cubic step is 
c           farther from stp than the secant step, the cubic step is 
c           taken, otherwise the secant step is taken.

            if (abs(stpc-stp) .gt. abs(stpq-stp)) then
               stpf = stpc
            else
               stpf = stpq
            endif
            stpf = min(stpmax,stpf)
            stpf = max(stpmin,stpf)
         endif

c     Fourth case: A lower function value, derivatives of the same sign, 
c     and the magnitude of the derivative does not decrease. If the 
c     minimum is not bracketed, the step is either stpmin or stpmax, 
c     otherwise the cubic step is taken.

      else
         if (brackt) then
            theta = three*(fp - fy)/(sty - stp) + dy + dp
            s = max(abs(theta),abs(dy),abs(dp))
            gamma = s*sqrt((theta/s)**2 - (dy/s)*(dp/s))
            if (stp .gt. sty) gamma = -gamma
            p = (gamma - dp) + theta
            q = ((gamma - dp) + gamma) + dy
            r = p/q
            stpc = stp + r*(sty - stp)
            stpf = stpc
         else if (stp .gt. stx) then
            stpf = stpmax
         else
            stpf = stpmin
         endif
      endif

c     Update the interval which contains a minimizer.

      if (fp .gt. fx) then
         sty = stp
         fy = fp
         dy = dp
      else
         if (sgnd .lt. zero) then
            sty = stx
            fy = fx
            dy = dx
         endif
         stx = stp
         fx = fp
         dx = dp
      endif

c     Compute the new step.

      stp = stpf

      return
      end
      

# replace expression levels by the residuals of regression analysis 
# in which predictor of interest is not in the regression model.
# The purpose is to remove potential confounding factors
obtainResi<-function(es, fmla)
{
 
  dat<-exprs(es)
  pDat<-pData(es)
  pDat2 <- pDat
  rownames(pDat2) <- 1:nrow(pDat2)
  designMat <- model.matrix(fmla, pDat2)
  rn <- as.numeric(rownames(designMat))
  es2 <- es[, rn]
  dat2 <- dat[, rn, drop = FALSE]
  # use pDat instead of pDat2 to keep 
  # original row names 
  pDat3 <- pDat[rn, , drop = FALSE]
  rownames(designMat) <- colnames(dat2)

  fit = lmFit(dat2, designMat)
  ebFit = eBayes(fit)

  # obtain residual matrix
  resMat<-lapply(1:nrow(dat2), function(i) {
      yi<-dat2[i,]
      betai<-ebFit$coefficients[i,]
      ri<-yi - designMat%*%betai
      return(ri)
    }
  )
  
  resMat2<-t(sapply(resMat, function(x) {x}))
  rownames(resMat2)<-rownames(dat2)
  colnames(resMat2)<-colnames(dat2)
  
  exprs(es2)<-resMat2
  pData(es2)<-pDat3

  invisible(es2)
}

"iniMemGenesTestFunc" <-
function(X, 
         memSubjects, 
         testFun, 
         geneNames=NULL, 
         alpha=0.05, eps=1.0e-6)
{

  nGenes<-nrow(X)
  alpha.old<-alpha

  if (identical(testFun,"myTtest")) {

# use the Fortran myTtest routine

    xc<-X[,memSubjects==1]
    xn<-X[,memSubjects==0]
    pvals<-rep(0,nGenes)
    tstats<-rep(0,nGenes)

    res <- .Fortran("myTtest", xc=as.double(xc), xn=as.double(xn),
                  numc=as.integer(ncol(xc)), numn=as.integer(ncol(xn)),
                  nGenes=as.integer(nGenes),
                  pvals=as.double(pvals), tstats=as.double(tstats),
                  NAOK=TRUE      )

    mat<-t(cbind(res$pvals,res$tstats))
  }
  else {
        mat<-apply(X, 1, testFun, memSubjects=memSubjects, alpha=alpha)
  }

  mat<-t(mat)
  colnames(mat)<-c("pvalue", "stat")

  # significant and have positive test statistics
  posUpp<-which(mat[,1]<alpha & mat[,2]>0)
  
  # significant and have negative test statistics
  posLow<-which(mat[,1]<alpha & mat[,2]<0)

  len1<-length(posUpp)
  len3<-length(posLow)
  len2<-nGenes-len1-len3
  if(len1>1 && len2>1 && len3>1 && len2>=len1 && len2>=len3)
  { 
  } else {
    mat2<-cbind(1:nrow(mat), mat) 
    mat3<-mat2[order(mat2[,2]),]
    mat3.p<-mat3[mat3[,3]>0,]
    mat3.n<-mat3[mat3[,3]<0,]
    tmpn<-ceiling(nGenes/5)
    #tmpn<-min(max(tmpn, 2, na.rm=TRUE), floor(nGenes/3), na.rm=TRUE)
    tmpn<-min(c(nrow(mat3.p), nrow(mat3.n),  
      max(tmpn, 2, na.rm=TRUE), floor(nGenes/3)), na.rm=TRUE)
    posUpp<-mat3.p[1:tmpn,1]
    posLow<-mat3.n[1:tmpn,1]
  }
  
  memGenes<-rep(2, nGenes)
  memGenes[posUpp]<-1
  memGenes[posLow]<-3

  memGenes2<-rep(1, nGenes)
  memGenes2[memGenes==2]<-0

  if(sum(is.null(geneNames), na.rm=TRUE))
  { geneNames<-paste("gene", 1:nGenes, sep="") }

  names(memGenes)<-geneNames
  names(memGenes2)<-geneNames
  p.value<-mat[,1]
  names(p.value)<-geneNames
  statistic<-mat[,2]
  names(statistic)<-geneNames

  tmp<-getPara(X, memSubjects, memGenes, eps=eps)
  para<-tmp$para
  llkh<-tmp$llkh

  res<-list(para=para, llkh=llkh, memGenes=memGenes, 
            memGenes2=memGenes2, p.value=p.value, 
            statistic=statistic)

  return(res) 
}

"myWilcox" <-
function(x, memSubjects, alpha =0.05, ...)
{
  xc<-x[memSubjects==1]
  xn<-x[memSubjects==0]

  m<-sum(memSubjects==1)
  res<-wilcox.test(x=xc, y=xn, conf.level = 1-alpha, ...)
  res2<-c(res$p.value, res$statistic-m*(m+1)/2)
  names(res2)<-c("p.value", "statistic")

  return(res2)
}

c      modified for use with GeneSelectMMD
c
c
c                             DRIVER 1 in Fortran 77
c     --------------------------------------------------------------
c                SIMPLE DRIVER FOR L-BFGS-B (version 3.0)
c     --------------------------------------------------------------
c
c        L-BFGS-B is a code for solving large nonlinear optimization
c             problems with simple bounds on the variables.
c
c        The code can also be used for unconstrained problems and is
c        as efficient for these problems as the earlier limited memory
c                          code L-BFGS.
c
c        This is the simplest driver in the package. It uses all the
c                    default settings of the code.
c
c
c     References:
c
c        [1] R. H. Byrd, P. Lu, J. Nocedal and C. Zhu, ``A limited
c        memory algorithm for bound constrained optimization'',
c        SIAM J. Scientific Computing 16 (1995), no. 5, pp. 1190--1208.
c
c        [2] C. Zhu, R.H. Byrd, P. Lu, J. Nocedal, ``L-BFGS-B: FORTRAN
c        Subroutines for Large Scale Bound Constrained Optimization''
c        Tech. Report, NAM-11, EECS Department, Northwestern University,
c        1994.
c
c
c          (Postscript files of these papers are available via anonymous
c           ftp to eecs.nwu.edu in the directory pub/lbfgs/lbfgs_bcm.)
c
c                              *  *  *
c
c         March 2011   (latest revision)
c         Optimization Center at Northwestern University
c         Instituto Tecnologico Autonomo de Mexico
c
c         Jorge Nocedal and Jose Luis Morales, Remark on "Algorithm 778: 
c         L-BFGS-B: Fortran Subroutines for Large-Scale Bound Constrained 
c         Optimization"  (2011). To appear in  ACM Transactions on 
c         Mathematical Software,

c     --------------------------------------------------------------
c             DESCRIPTION OF THE VARIABLES IN L-BFGS-B
c     --------------------------------------------------------------
c
c     n is an INTEGER variable that must be set by the user to the
c       number of variables.  It is not altered by the routine.
c
c     m is an INTEGER variable that must be set by the user to the
c       number of corrections used in the limited memory matrix.
c       It is not altered by the routine.  Values of m < 3  are
c       not recommended, and large values of m can result in excessive
c       computing time. The range  3 <= m <= 20 is recommended. 
c
c     x is a DOUBLE PRECISION array of length n.  On initial entry
c       it must be set by the user to the values of the initial
c       estimate of the solution vector.  Upon successful exit, it
c       contains the values of the variables at the best point
c       found (usually an approximate solution).
c
c     l is a DOUBLE PRECISION array of length n that must be set by
c       the user to the values of the lower bounds on the variables. If
c       the i-th variable has no lower bound, l(i) need not be defined.
c
c     u is a DOUBLE PRECISION array of length n that must be set by
c       the user to the values of the upper bounds on the variables. If
c       the i-th variable has no upper bound, u(i) need not be defined.
c
c     nbd is an INTEGER array of dimension n that must be set by the
c       user to the type of bounds imposed on the variables:
c       nbd(i)=0 if x(i) is unbounded,
c              1 if x(i) has only a lower bound,
c              2 if x(i) has both lower and upper bounds, 
c              3 if x(i) has only an upper bound.
c
c     f is a DOUBLE PRECISION variable.  If the routine setulb returns
c       with task(1:2)= 'FG', then f must be set by the user to
c       contain the value of the function at the point x.
c
c     g is a DOUBLE PRECISION array of length n.  If the routine setulb
c       returns with taskb(1:2)= 'FG', then g must be set by the user to
c       contain the components of the gradient at the point x.
c
c     factr is a DOUBLE PRECISION variable that must be set by the user.
c       It is a tolerance in the termination test for the algorithm.
c       The iteration will stop when
c
c        (f^k - f^{k+1})/max{|f^k|,|f^{k+1}|,1} <= factr*epsmch
c
c       where epsmch is the machine precision which is automatically
c       generated by the code. Typical values for factr on a computer
c       with 15 digits of accuracy in double precision are:
c       factr=1.d+12 for low accuracy;
c             1.d+7  for moderate accuracy; 
c             1.d+1  for extremely high accuracy.
c       The user can suppress this termination test by setting factr=0.
c
c     pgtol is a double precision variable.
c       On entry pgtol >= 0 is specified by the user.  The iteration
c         will stop when
c
c                 max{|proj g_i | i = 1, ..., n} <= pgtol
c
c         where pg_i is the ith component of the projected gradient.
c       The user can suppress this termination test by setting pgtol=0.
c
c     wa is a DOUBLE PRECISION  array of length 
c       (2mmax + 5)nmax + 11mmax^2 + 8mmax used as workspace.
c       This array must not be altered by the user.
c
c     iwa is an INTEGER  array of length 3nmax used as
c       workspace. This array must not be altered by the user.
c
c     task is a CHARACTER string of length 60.
c       On first entry, it must be set to 'START'.
c       On a return with task(1:2)='FG', the user must evaluate the
c         function f and gradient g at the returned value of x.
c       On a return with task(1:5)='NEW_X', an iteration of the
c         algorithm has concluded, and f and g contain f(x) and g(x)
c         respectively.  The user can decide whether to continue or stop
c         the iteration. 
c       When
c         task(1:4)='CONV', the termination test in L-BFGS-B has been 
c           satisfied;
c         task(1:4)='ABNO', the routine has terminated abnormally
c           without being able to satisfy the termination conditions,
c           x contains the best approximation found,
c           f and g contain f(x) and g(x) respectively;
c         task(1:5)='ERROR', the routine has detected an error in the
c           input parameters;
c       On exit with task = 'CONV', 'ABNO' or 'ERROR', the variable task
c         contains additional information that the user can print.
c       This array should not be altered unless the user wants to
c          stop the run for some reason.  See driver2 or driver3
c          for a detailed explanation on how to stop the run 
c          by assigning task(1:4)='STOP' in the driver.
c
c     iprint is an INTEGER variable that must be set by the user.
c       It controls the frequency and type of output generated:
c        iprint<0    no output is generated;
c        iprint=0    print only one line at the last iteration;
c        0<iprint<99 print also f and |proj g| every iprint iterations;
c        iprint=99   print details of every iteration except n-vectors;
c        iprint=100  print also the changes of active set and final x;
c        iprint>100  print details of every iteration including x and g;
c       When iprint > 0, the file iterate.dat will be created to
c                        summarize the iteration.
c
c     csave  is a CHARACTER working array of length 60.
c
c     lsave is a LOGICAL working array of dimension 4.
c       On exit with task = 'NEW_X', the following information is
c         available:
c       lsave(1) = .true.  the initial x did not satisfy the bounds;
c       lsave(2) = .true.  the problem contains bounds;
c       lsave(3) = .true.  each variable has upper and lower bounds.
c
c     isave is an INTEGER working array of dimension 44.
c       On exit with task = 'NEW_X', it contains information that
c       the user may want to access:
c         isave(30) = the current iteration number;
c         isave(34) = the total number of function and gradient
c                         evaluations;
c         isave(36) = the number of function value or gradient
c                                  evaluations in the current iteration;
c         isave(38) = the number of free variables in the current
c                         iteration;
c         isave(39) = the number of active constraints at the current
c                         iteration;
c
c         see the subroutine setulb.f for a description of other 
c         information contained in isave
c
c     dsave is a DOUBLE PRECISION working array of dimension 29.
c       On exit with task = 'NEW_X', it contains information that
c         the user may want to access:
c         dsave(2) = the value of f at the previous iteration;
c         dsave(5) = the machine precision epsmch generated by the code;
c         dsave(13) = the infinity norm of the projected gradient;
c
c         see the subroutine setulb.f for a description of other 
c         information contained in dsave
c
c     --------------------------------------------------------------
c           END OF THE DESCRIPTION OF THE VARIABLES IN L-BFGS-B
c     --------------------------------------------------------------

!      program driver
      subroutine lbfgsbdriver(task, isave, dsave,
     +      par, nc, nn, nobj, 
     +      sumw1, sumw2, sumw3, 
     +      sumw1xcTxc, sumw1xcT1, sumw1xcT1_sq, 
     +      sumw1xnTxn, sumw1xnT1, sumw1xnT1_sq, 
     +      sumw2xTx, sumw2xT1, sumw2xT1_sq, 
     +      sumw3xcTxc, sumw3xcT1, sumw3xcT1_sq, 
     +      sumw3xnTxn, sumw3xnT1, sumw3xnT1_sq  )
 
      implicit none

       integer nc, nn, nobj
       double precision sumw1, sumw2, sumw3
       double precision sumw1xcTxc, sumw1xcT1, sumw1xcT1_sq
       double precision sumw1xnTxn, sumw1xnT1, sumw1xnT1_sq
       double precision sumw2xTx, sumw2xT1, sumw2xT1_sq
       double precision sumw3xcTxc, sumw3xcT1, sumw3xcT1_sq
       double precision sumw3xnTxn, sumw3xnT1, sumw3xnT1_sq

       double precision par(15)
!       double precision lower(2)
!       double precision upper(2)

c      objective function variables
       double precision negq
       double precision dnegq(15)

       double precision ncd, nnd, nobjd


c     This simple driver demonstrates how to call the L-BFGS-B code to
c       solve a sample problem (the extended Rosenbrock function 
c       subject to bounds on the variables). The dimension n of this
c       problem is variable.
 
      integer          nmax, mmax
      parameter        (nmax=1024, mmax=17)
c        nmax is the dimension of the largest problem to be solved.
c        mmax is the maximum number of limited memory corrections.
 
c     Declare the variables needed by the code.
c       A description of all these variables is given at the end of 
c       the driver.
 
      character*60     task, csave
      logical          lsave(4)
      integer          n, m, iprint,
     +                 nbd(nmax), iwa(3*nmax), isave(44)
      double precision f, factr, pgtol, 
     +                 x(nmax), l(nmax), u(nmax), g(nmax), dsave(29), 
     +                 wa(2*mmax*nmax + 5*nmax + 11*mmax*mmax + 8*mmax)

c     Declare a few additional variables for this sample problem.

      integer          i
 
c     We wish to have output at every iteration.

!      iprint = 1
c     No output from Fortran code with iprint = -1
      iprint = -1

c     We specify the tolerances in the stopping criteria.

      factr=1.0d+7
c      factr=0

      pgtol=1.0d-5
c      pgtol=0


c     We specify the dimension n of the sample problem and the number
c        m of limited memory corrections stored.  (n and m should not
c        exceed the limits nmax and mmax respectively.)
 
      n=15
      m=5
c      m=17
 
c     We now provide nbd which defines the bounds on the variables:
c                    l   specifies the lower bounds,
c                    u   specifies the upper bounds. 
 
c     these bounds could be set wider, but that doesn't seem to be required
      do 10 i=1,n
         nbd(i)=2
         l(i)=-10d0
         u(i)=10d0
  10  continue


c     We now define the starting point.

      do 14 i=1,n
         x(i)=par(i)
  14  continue
 
c     cast the array dimension values to double
      ncd = dble(nc)
      nnd = dble(nn)
      nobjd = dble(nobj)


c     We start the iteration by initializing task.
c 
      task = 'START'

c        ------- the beginning of the loop ----------
 
 111  continue
      
c     This is the call to the L-BFGS-B code.
 
      call setulb(n,m,x,l,u,nbd,f,g,factr,pgtol,wa,iwa,task,iprint,
     +            csave,lsave,isave,dsave)
 
      if (task(1:2) .eq. 'FG') then
c        the minimization routine has returned to request the
c        function f and gradient g values at the current x.

c        Compute function value f for the sample problem.

       call negqfunc(negq, x, ncd, nnd, nobjd, 
     +      sumw1, sumw2, sumw3, 
     +      sumw1xcTxc, sumw1xcT1, sumw1xcT1_sq, 
     +      sumw1xnTxn, sumw1xnT1, sumw1xnT1_sq, 
     +      sumw2xTx, sumw2xT1, sumw2xT1_sq, 
     +      sumw3xcTxc, sumw3xcT1, sumw3xcT1_sq, 
     +      sumw3xnTxn, sumw3xnT1, sumw3xnT1_sq  )

       f=negq

c        Compute gradient g for the sample problem.

       call dnegqfunc(dnegq, x, ncd, nnd, nobjd,
     +      sumw1, sumw2, sumw3,
     +      sumw1xcTxc, sumw1xcT1, sumw1xcT1_sq,
     +      sumw1xnTxn, sumw1xnT1, sumw1xnT1_sq,
     +      sumw2xTx, sumw2xT1, sumw2xT1_sq,
     +      sumw3xcTxc, sumw3xcT1, sumw3xcT1_sq,
     +      sumw3xnTxn, sumw3xnT1, sumw3xnT1_sq  )

       do 22 i=1,n
         g(i)=dnegq(i)
  22   continue

c          go back to the minimization routine.
       goto 111
      endif
c
      if (task(1:5) .eq. 'NEW_X')  goto 111
c        the minimization routine has returned with a new iterate,
c         and we have opted to continue the iteration.

c           ---------- the end of the loop -------------
 
c     If task is neither FG nor NEW_X we terminate execution.


c     return the final values

       do 44 i=1,n
         par(i)=x(i)
  44   continue

      return
 
      end

c======================= The end of driver1 ============================

      double precision function dnrm2(n,x,incx)
      integer n,incx
      double precision x(n)
c     **********
c
c     Function dnrm2
c
c     Given a vector x of length n, this function calculates the
c     Euclidean norm of x with stride incx.
c
c     The function statement is
c
c       double precision function dnrm2(n,x,incx)
c
c     where
c
c       n is a positive integer input variable.
c
c       x is an input array of length n.
c
c       incx is a positive integer variable that specifies the 
c         stride of the vector.
c
c     Subprograms called
c
c       FORTRAN-supplied ... abs, max, sqrt
c
c     MINPACK-2 Project. February 1991.
c     Argonne National Laboratory.
c     Brett M. Averick.
c
c     **********
      integer i
      double precision scale

      dnrm2 = 0.0d0
      scale = 0.0d0

      do 10 i = 1, n, incx
         scale = max(scale, abs(x(i)))
   10 continue

      if (scale .eq. 0.0d0) return

      do 20 i = 1, n, incx
         dnrm2 = dnrm2 + (x(i)/scale)**2
   20 continue

      dnrm2 = scale*sqrt(dnrm2)

 
      return

      end
      
c====================== The end of dnrm2 ===============================

      subroutine daxpy(n,da,dx,incx,dy,incy)
c
c     constant times a vector plus a vector.
c     uses unrolled loops for increments equal to one.
c     jack dongarra, linpack, 3/11/78.
c
      double precision dx(*),dy(*),da
      integer i,incx,incy,ix,iy,m,mp1,n
c
      if(n.le.0)return
      if (da .eq. 0.0d0) return
      if(incx.eq.1.and.incy.eq.1)go to 20
c
c        code for unequal increments or equal increments
c          not equal to 1
c
      ix = 1
      iy = 1
      if(incx.lt.0)ix = (-n+1)*incx + 1
      if(incy.lt.0)iy = (-n+1)*incy + 1
      do 10 i = 1,n
        dy(iy) = dy(iy) + da*dx(ix)
        ix = ix + incx
        iy = iy + incy
   10 continue
      return
c
c        code for both increments equal to 1
c
c
c        clean-up loop
c
   20 m = mod(n,4)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dy(i) = dy(i) + da*dx(i)
   30 continue
      if( n .lt. 4 ) return
   40 mp1 = m + 1
      do 50 i = mp1,n,4
        dy(i) = dy(i) + da*dx(i)
        dy(i + 1) = dy(i + 1) + da*dx(i + 1)
        dy(i + 2) = dy(i + 2) + da*dx(i + 2)
        dy(i + 3) = dy(i + 3) + da*dx(i + 3)
   50 continue
      return
      end
      
c====================== The end of daxpy ===============================

      subroutine dcopy(n,dx,incx,dy,incy)
c
c     copies a vector, x, to a vector, y.
c     uses unrolled loops for increments equal to one.
c     jack dongarra, linpack, 3/11/78.
c
      double precision dx(*),dy(*)
      integer i,incx,incy,ix,iy,m,mp1,n
c
      if(n.le.0)return
      if(incx.eq.1.and.incy.eq.1)go to 20
c
c        code for unequal increments or equal increments
c          not equal to 1
c
      ix = 1
      iy = 1
      if(incx.lt.0)ix = (-n+1)*incx + 1
      if(incy.lt.0)iy = (-n+1)*incy + 1
      do 10 i = 1,n
        dy(iy) = dx(ix)
        ix = ix + incx
        iy = iy + incy
   10 continue
      return
c
c        code for both increments equal to 1
c
c
c        clean-up loop
c
   20 m = mod(n,7)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dy(i) = dx(i)
   30 continue
      if( n .lt. 7 ) return
   40 mp1 = m + 1
      do 50 i = mp1,n,7
        dy(i) = dx(i)
        dy(i + 1) = dx(i + 1)
        dy(i + 2) = dx(i + 2)
        dy(i + 3) = dx(i + 3)
        dy(i + 4) = dx(i + 4)
        dy(i + 5) = dx(i + 5)
        dy(i + 6) = dx(i + 6)
   50 continue
      return
      end
      
c====================== The end of dcopy ===============================

      double precision function ddot(n,dx,incx,dy,incy)
c
c     forms the dot product of two vectors.
c     uses unrolled loops for increments equal to one.
c     jack dongarra, linpack, 3/11/78.
c
      double precision dx(*),dy(*),dtemp
      integer i,incx,incy,ix,iy,m,mp1,n
c
      ddot = 0.0d0
      dtemp = 0.0d0
      if(n.le.0)return
      if(incx.eq.1.and.incy.eq.1)go to 20
c
c        code for unequal increments or equal increments
c          not equal to 1
c
      ix = 1
      iy = 1
      if(incx.lt.0)ix = (-n+1)*incx + 1
      if(incy.lt.0)iy = (-n+1)*incy + 1
      do 10 i = 1,n
        dtemp = dtemp + dx(ix)*dy(iy)
        ix = ix + incx
        iy = iy + incy
   10 continue
      ddot = dtemp
      return
c
c        code for both increments equal to 1
c
c
c        clean-up loop
c
   20 m = mod(n,5)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dtemp = dtemp + dx(i)*dy(i)
   30 continue
      if( n .lt. 5 ) go to 60
   40 mp1 = m + 1
      do 50 i = mp1,n,5
        dtemp = dtemp + dx(i)*dy(i) + dx(i + 1)*dy(i + 1) +
     *   dx(i + 2)*dy(i + 2) + dx(i + 3)*dy(i + 3) + dx(i + 4)*dy(i + 4)
   50 continue
   60 ddot = dtemp
      return
      end
      
c====================== The end of ddot ================================

      subroutine  dscal(n,da,dx,incx)
c
c     scales a vector by a constant.
c     uses unrolled loops for increment equal to one.
c     jack dongarra, linpack, 3/11/78.
c     modified 3/93 to return if incx .le. 0.
c
      double precision da,dx(*)
      integer i,incx,m,mp1,n,nincx
c
      if( n.le.0 .or. incx.le.0 )return
      if(incx.eq.1)go to 20
c
c        code for increment not equal to 1
c
      nincx = n*incx
      do 10 i = 1,nincx,incx
        dx(i) = da*dx(i)
   10 continue
      return
c
c        code for increment equal to 1
c
c
c        clean-up loop
c
   20 m = mod(n,5)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dx(i) = da*dx(i)
   30 continue
      if( n .lt. 5 ) return
   40 mp1 = m + 1
      do 50 i = mp1,n,5
        dx(i) = da*dx(i)
        dx(i + 1) = da*dx(i + 1)
        dx(i + 2) = da*dx(i + 2)
        dx(i + 3) = da*dx(i + 3)
        dx(i + 4) = da*dx(i + 4)
   50 continue
      return
      end
      
c====================== The end of dscal ===============================

%\VignetteIndexEntry{Gene Selection based on a mixture of marginal distributions}
%\VignetteDepends{GeneSelectMMD}
%\VignetteKeywords{Microarray experiment}
%\VignettePackage{GeneSelectMMD}

\documentclass[10pt]{article}
\usepackage{url}
\usepackage{amsmath}
\usepackage{natbib}
%\usepackage{isorot}
\usepackage{fullpage} % standard 1 inch margins 

\newcommand{\Robject}[1]{{\texttt{#1}}}
\newcommand{\Rfunction}[1]{{\texttt{#1}}}
\newcommand{\Rpackage}[1]{{\textit{#1}}}
\newcommand{\Rclass}[1]{{\textit{#1}}}
\newcommand{\Rmethod}[1]{{\textit{#1}}}
\newcommand{\code}[1]{\texttt{#1}}
\newcommand{\email}[1]{\texttt{#1}}


\newcommand{\mb}[1]{{\boldsymbol{#1}}}
\newcommand\myC[1]{{\cal #1}\,}
\def\lnit{{\rm lnit}\,}
\def\ln{{\rm ln}\,}
\def\N{{\rm N}\,}
\def\E{{\rm E}\,}
\def\vec{{\rm vec}\,}
\def\vech{{\rm vech}\,}
\def\Var{{\rm Var}\,}
\def\Cov{{\rm Cov}\,}
\def\wCov{\widehat{\rm Cov}\,}
\def\Q{\mb{Q}}
\def\I{\mb{I}}
\def\V{\mb{V}}
\def\B{\mb{B}}
\def\D{\mb{D}}
\def\C{\mb{C}}
\def\W{\mb{W}}
\def\A{\mb{A}}
\def\H{\mb{H}}
\def\Y{{\bf Y}}
\def\y{{\bf y}}
\def\Z{\mb{Z}}
\def\thbf{\mb{\theta}}
\def\mbS{\mb{\Sigma}}
\def\mbmu{\mb{\mu}}
\def\mbI{\mb{I}}
\def\hmbS{\hat{\mb{\Sigma}}}
\def\mbT{\mb{\Theta}}
\def\hmbT{\hat{\mb{\Theta}}}
\def\mbw{\mb{w}}
\newcommand{\oo}{\infty}
\def\diam{\,{\rm diam}\,}
\def\chs{\,{\rm chs}\,}
\def\join{\,{\rm join}\,}
\def\x{{\bf x}}
\def\X{{\bf X}}
\def\mbx{{\bf x}}
\def\mbX{{\bf X}}
\def\diag{{\rm diag}}





%%% Hyperlinks for ``PDF Latex'' :
\ifx\pdfoutput\undefined%%--- usual ``latex'' :
  %% Stuff w/out hyperref
\else%%---------------------- `` pdflatex '' : -- still gives funny errors
  \RequirePackage{hyperref}
  %% The following is R's share/texmf/hyperref.cfg :
  %% Stuff __with__ hyperref :
  \hypersetup{%
    %default: hyperindex,%
    colorlinks,%
    %default: pagebackref,%
    linktocpage,%
    %%plainpages=false,%
    linkcolor=Green,%
    citecolor=Blue,%
    urlcolor=Red,%
    pdfstartview=Fit,%
    pdfview={XYZ null null null}%
    }
  \RequirePackage{color}
  \definecolor{Blue}{rgb}{0,0,0.8}
  \definecolor{Green}{rgb}{0.1,0.75,0.1}
  \definecolor{Red}{rgb}{0.7,0,0}
  %% ESS JCGS v2 :
  %%\hypersetup{backref,colorlinks=true,pagebackref=true,hyperindex=true}
  %%\hypersetup{backref,colorlinks=false,pagebackref=true,hyperindex=true}
\fi


\usepackage{Sweave}
\begin{document}

\title{Gene Selection Using \Rpackage{GeneSelectMMD}}
\author{
        Jarrett Morrow\\
        \email{remdj@channing.harvard.edu},
        \\
        Weilianq Qiu\\
        \email{Weiliang.Qiu@gmail.com},
        \\
        Wenqing He\\
        \email{whe@stats.uwo.ca},
        \\
        Xiaogang Wang\\
        \email{stevenw@mathstat.yorku.ca},
        \\
        Ross Lazarus\\
        \email{ross.lazarus@gmail.com},
        }       

\maketitle

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\section{Introduction}

This document demonstrates how to use the \Rpackage{GeneSelectMMD}
 package to detect significant genes and to 
estimate
false discovery rate (FDR), false non-discovery rate (FNDR),
false positive rate (FPR), and false negative rate (FNR),
for a real microarray data set based
on the method proposed by Qiu et al. (2008).\nocite{QiuEtAl:2008}
It also illustrates how to visualize the fit of the model proposed in
Qiu et al. (2008)
to the real microarray data set when the marginal correlations among subjects
are zero or close to zero.

The \Rpackage{GeneSelectMMD} package is suitable for the case where
there are only two tissue types and for the case where 
tissue samples within a tissue type are conditionally independent given
the gene (c.f. Qiu et al., 2008). 

Note that starting from version 1.7.1, we reparameterize the model parameters.
The input and output are the same as the previous versions.
To improve speed, we use numerical optimization, instead of iteration based on explicit formula, to obtain parameter estimates.
Since verions 2.0.7, the speed has also been improved by using Fortran 77 code for some of the less efficient operations.
Lastly, in version 2.0.8 onward,
the likelihood values provided to the user are based on the observed data, while in previous
versions the likelihood was an expected value.

%\clearpage

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Methods}

MMD assumes that the marginal distribution of a gene profile is a
mixture of 3-component multivariate normal distributions with
special structure for mean vectors and covariance matrices. 
The 3
component distributions correspond to 3 gene clusters: (1) cluster
of genes over-expressed in the case group; (2) cluster of genes
non-differentially expressed; and (3) cluster of genes
under-expressed in the case group. 
Specifically, the marginal density of a gene profile is assumed to be
\begin{equation}\label{Model: mixture of 3-multivariate Normal}
\begin{aligned}
&f(\mb{x}|\thbf_1, \thbf_2, \thbf_3)=\pi_1 f_1(\mb{x}|\thbf_1)+\pi_2 f_2(\mb{x}|\thbf_2)+ \pi_3
f_3(\mb{x}|\thbf_3),\\
&\pi_1+\pi_2+\pi_3=1, \pi_i>0, i=1, 2, 3,
\end{aligned}
\end{equation}
where $\pi_1$, $\pi_2$, $\pi_3$ are mixture proportions. The $m\times 1$
vector $\mb{x}$ is a realization of the random vector
$\mb{X}$ that represents the transformed gene profile for a randomly
selected gene over $m$ tissue samples ($m=m_c+m_n$, where
$m_c$ is the number of case tissue samples and $m_n$ control
tissue samples);
$\thbf_k$,
is the parameter set for the $k$-th component distribution $f_k$,
$k=1, 2, 3$; and $f_1$, $f_2$, and $f_3$ are the density functions for
multivariate Normal distributions with the mean vectors
\begin{equation}\label{Formulae: mean vector structure}
\mb{\mu}_1=\left(\begin{array}{c}
\mu_{c1}\mb{1}_{m_c}\\
\mu_{n1}\mb{1}_{m_n}
\end{array}\right),\quad
\mb{\mu}_2=\mu_2\mb{1}_{m},\quad
\mb{\mu}_3=\left(\begin{array}{c}
\mu_{c3}\mb{1}_{m_c}\\
\mu_{n3}\mb{1}_{m_n}
\end{array}\right).
\end{equation}
and covariance matrices
\begin{equation}\label{Formulae: covariance matrix structure}
\mb{\Sigma}_1=\left(
\begin{array}{cc}
\sigma^2_{c_1}\mb{R}_{c_1}&\mb{0}\\
\mb{0}& \sigma^2_{n_1}\mb{R}_{n_1}
\end{array}
\right),\quad
\mb{\Sigma}_2=\sigma_2^2\mb{R}_{2},\quad
\mb{\Sigma}_3=\left(
\begin{array}{cc}
\sigma^2_{c_3}\mb{R}_{c_3}&\mb{0}\\
\mb{0}& \sigma^2_{n_3}\mb{R}_{n_3}
\end{array}
\right),
\end{equation}
respectively, where correlation matrix
\begin{equation}\label{Formula: Rt}
\mb{R}_{t}=(1-\rho_{t})\left[\mb{I}_{n_t}+\frac{\rho_{t}}{(1-\rho_{t})}\mb{1}_{n_t}\mb{1}_{n_t}^T\right],
\end{equation}
$t=c_1$, $n_1$, $2$, $c_3$, or $n_3$. $n_t=m_c$ if $t=c_1$ or $c_3$;
$n_t=m$ if $t=2$; $n_t=m_n$ if $t=n_1$, or $n_3$. Here we assume,
without loss of generality, that the first $m_c$ elements of the random
vector $\X$ are for the case tissue samples and the remaining
$m_n$ elements are for the control tissue samples. Let
$\thbf_1=(\mu_{c_1}, \sigma^2_{c_1}, \rho_{c_1}, \mu_{n_1},
\sigma^2_{n_1}, \rho_{n_1})^T$, $\thbf_2=(\mu_2, \sigma^2_2,
\rho_{2})^T$, $\thbf_3=(\mu_{c_3}, \sigma^2_{c_3}, \rho_{c_3},
\mu_{n_3}, \sigma^2_{n_3}, \rho_{n_3})^T$. Note that
$\mu_{c_1}>\mu_{n_1}$ for component $1$ in which genes are
over-expressed in case tissue samples, and $\mu_{c_3}<\mu_{n_3}$
for component 3 where genes are under-expressed in case samples.
Our prior belief is that the majority of genes are usually
non-differentially expressed,
so we assume $\pi_2>\pi_1$ and $\pi_2>\pi_3$.


The model parameters can be estimated using the EM
algorithm (Dempster et al., 1977).\nocite{DempsterEtAl:1977}

The cluster membership of a
gene is determined by the posterior probability that the gene
belongs to a cluster given its gene profile. The posterior probability
is a function of the 3 component marginal density functions and the
mixing proportions: 
\begin{equation}
\begin{aligned}
Pr(\mbox{gene $i \in$  cluster $k$}|\mb{x}_i)=&\frac{\pi_k
f_k(\mb{x}_i|\thbf_k)}{\pi_1 f_1(\mb{x}_i|\thbf_1) +\pi_2
f_2(\mb{x}_i|\thbf_2)+\pi_3 f_3(\mb{x}_i|\thbf_3)},\\ k=&1, 2, 3,
\end{aligned}
\end{equation}
Specifically, a gene is assigned to the a gene 
cluster if the
posterior probability that the gene belongs to that gene
cluster given its gene profile is larger than the posterior probability that
the gene belongs to other gene clusters
given its gene profile:
\begin{equation}\label{Formula: criterion for gene selection}
k_0=\arg\max_{k=1,2,3}Pr(\mbox{gene $i \in$ cluster $k$}|\mb{x}_i).
\end{equation}


An important task is to assess the performance of a gene selection method
so that different methods can be objectively compared.
To evaluate the performance of a gene selection method,
investigators usually compare the error rates, such as FDR, FNDR,
FPR, and FNR, via simulation studies. However, when analyzing a real
microarray data set, investigators are more interested in what the
values of FDR, FNDR, FPR, and FNR are for this specific real
microarray set. It is challenging to estimate FDR, FNDR, FPR, and
FNR for real microarray data sets since the true gene cluster
membership is unknown for real data sets. However, model-based gene
clustering methods, such as Bayesian hierarchical models and MMD,
can provide such estimates since these error rates can be
expressed as functions of marginal density functions and mixing
proportions, where the model parameters and mixing proportions can
be estimated from the real microarray data.
It is easy to use MMD to estimate the four error rates
since MMD describes the distributions of gene expression levels directly
via the marginal distributions, while it is usually difficult to
derive the marginal density functions for Bayesian hierarchical
models.

It is of practical importance to evaluate if a model fits a
real microarray data
set well. If a model does not fit well for the real
microarray data set, then it makes no sense to estimate the error
rates based on the model. Although it is quite challenging to asses
the goodness of fit for multivariate data, especially for non-normal
multivariate data, it is possible to do so for some special cases.
For example, when tissue samples are marginally independent, we
could pool the gene expression levels across tissue samples for each
type of tissue samples since they are all independent. We then could
impose the theoretical density curve on the histogram of the pooled
expression levels for each type of tissue samples. The parameters $\rho_{c1}$,
$\rho_{n1}$, $\rho_{2}$, $\rho_{c3}$, and $\rho_{n3}$
indicates the marginal correlations among tissue samples. Assuming
that the marginal correlations are ignorable, we
then could produce such a plot to evaluate the goodness of fit of
MMD to the real microarray data set.

\section{Reparametrization}


To guarantee the constraints $\mu_{c_1}>\mu_{n_1}$ and
$\mu_{c_3}<\mu_{n_3}$ in the iteration of the EM algorithm, we
re-parameterize the model parameters:
\begin{equation*}
\begin{aligned}
\mu_{n_1}=&\mu_{c_1}-\exp(\triangle_{n_1}),\\
\mu_{n_3}=&\mu_{c_3}+\exp(\triangle_{n_3}).
\end{aligned}
\end{equation*}

To make sure the marginal covariance matrices are positive definite,
we set the constraints:
\begin{equation*}
\begin{aligned}
-\frac{1}{n_c-1}<&\rho_{c_1}<1\\
-\frac{1}{n_n-1}<&\rho_{n_1}<1\\
-\frac{1}{n-1}<&\rho_{2}<1\\
-\frac{1}{n_c-1}<&\rho_{c_3}<1\\
-\frac{1}{n_n-1}<&\rho_{n_3}<1
\end{aligned}
\end{equation*}
To make sure the above inequalities hold, we reparametrize $\rho$s as follows:
\begin{equation*}
\begin{aligned}
\rho_{c_1}=&\frac{\exp(r_{c_1})-1/(n_c-1)}{1+\exp(r_{c_1})}\\
\rho_{n_1}=&\frac{\exp(r_{n_1})-1/(n_n-1)}{1+\exp(r_{n_1})}\\
\rho_{2}=&\frac{\exp(r_{2})-1/(n-1)}{1+\exp(r_{2})}\\
\rho_{c_3}=&\frac{\exp(r_{c_3})-1/(n_c-3)}{1+\exp(r_{c_3})}\\
\rho_{n_3}=&\frac{\exp(r_{n_3})-1/(n_n-3)}{1+\exp(r_{n_3})}.
\end{aligned}
\end{equation*}

Another set of constraints are:
\begin{equation*}
\begin{aligned}
0<&\pi_k<1,\quad k=1, 2, 3,\\
\pi_3=&1-\pi_1-\pi_2.
\end{aligned}
\end{equation*}

Then the reparametrized MMD model is
\begin{equation}
f(\mb{x})=\pi_1
f_1(\mb{x}|\mb{\theta}_1)+\pi_2f_2(\mb{x}|\mb{\theta}_2)
+(1-\pi_1-\pi_2) f_3(\mb{x}|\mb{\theta}_3),
\end{equation}
where
\begin{equation}
\begin{aligned}
\mb{\theta}_1=&(\mu_{c_1}, \sigma_{c_1}^2, r_{c_1}, \triangle_{n_1}, \sigma_{n_1}^2, r_{n_1})^T,\\
\mb{\theta}_2=&(\mu_2, \sigma_2^2, r_{2})^T,\\
\mb{\theta}_3=&(\mu_{c_3}, \sigma_{c_3}^2, r_{c_3},
\triangle_{n_3}, \sigma_{n_3}^2, r_{n_3})^T,
\end{aligned}
\end{equation}
and $f_1$, $f_2$, and $f_3$ are the density functions for
multivariate normal distributions with mean vectors
\begin{equation}
\mb{\mu}_1=\left(
\begin{array}{c}
\mu_{c_1}\mb{1}_{n_c}\\
\left[\mu_{c_1}-\exp(\triangle_{n_1})\right]\mb{1}_{n_n}
\end{array}
\right),\; \mb{\mu}_2=\mu_2\mb{1}_{n},\; \mb{\mu}_3=\left(
\begin{array}{c}
\mu_{c_3}\mb{1}_{n_c}\\
\left[\mu_{c_3}+\exp(\triangle_{n_3})\right]\mb{1}_{n_n}
\end{array}
\right)
\end{equation}
and covariance matrices
\begin{equation}
\begin{aligned}
\mb{\Sigma}_1=&\left(
\begin{array}{cc}
\sigma^2_{c_1}\frac{n_c}{(n_c-1)(1+\exp(r_{c_1}))}\mb{R}_{0,c_1}&\mb{0}\\
\mb{0}&\sigma^2_{n_1}\frac{n_n}{(n_n-1)(1+\exp(r_{n_1}))}\mb{R}_{0,n_1}
\end{array}
\right),\\
\mb{\Sigma_2}=&\sigma^2_2\frac{n}{(n-1)(1+\exp(r_{2}))}\mb{R}_{0,2},\\
\mb{\Sigma}_3=&\left(
\begin{array}{cc}
\sigma^2_{c_3}\frac{n_c}{(n_c-1)(1+\exp(r_{c_3}))}\mb{R}_{0,c_3}&\mb{0}\\
\mb{0}&\sigma^2_{n_3}\frac{n_n}{(n_n-1)(1+\exp(r_{n_3}))}\mb{R}_{0,n_3}
\end{array}
\right).
\end{aligned}
\end{equation}
The matrices $\mb{R}_{0, c_1}$, $\mb{R}_{0, n_1}$, $\mb{R}_{0, 2}$,
$\mb{R}_{0, c_3}$, and $\mb{R}_{0, n_3}$ are defined as below:
\begin{equation}
\begin{aligned}
\mb{R}_{0, c_1}=&\mb{I}_{n_c}+\frac{(n_c-1)\exp(r_{c_1})-1}{n_{c_1}}\mb{1}_{n_c}\mb{1}_{n_c}^T,\\
\mb{R}_{0, n_1}=&\mb{I}_{n_n}+\frac{(n_n-1)\exp(r_{n_1})-1}{n_{n_1}}\mb{1}_{n_n}\mb{1}_{n_n}^T,\\
\mb{R}_{0, 2}=&\mb{I}_{n}+\frac{(n-1)\exp(r_{2})-1}{n_{2}}\mb{1}_{n}\mb{1}_{n}^T,\\
\mb{R}_{0, c_3}=&\mb{I}_{n_c}+\frac{(n_c-1)\exp(r_{c_3})-1}{n_{c_3}}\mb{1}_{n_c}\mb{1}_{n_c}^T,\\
\mb{R}_{0,
n_3}=&\mb{I}_{n_n}+\frac{(n_n-1)\exp(r_{n_3})-1}{n_{n_3}}\mb{1}_{n_n}\mb{1}_{n_n}^T.
\end{aligned}
\end{equation}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\section{Gene selection via \Rpackage{GeneSelectMMD}}
\label{Section: Gene Selection Functions}

The \Rpackage{GeneSelectMMD} includes four functions for gene selection:
\code{gsMMD}, \code{gsMMD.default}, \code{gsMMD2}, and \code{gsMMD2.default}.

The functions \code{gsMMD} and \code{gsMMD2} accept the object derived from the class of
 \Rpackage{Bioconductor}'s
\code{ExpressionSet} as data input, while the functions \code{gsMMD.default}
and \code{gsMMD2.default} accept data matrix as data input.

The functions \code{gsMMD} and \code{gsMMD.default} will provide
initial 3-cluster gene partitions (cluster of genes over-expressed in 
case group, cluster of genes non-differentially expressed, and 
cluster of genes under-expressed in case group) based on the 
gene-wise two-sample t-test or two-sample Wilcoxon 
rank-sum test.
In situations where the user would like to provide the initial 3-cluster
partition other than that provided by the gene-wise two-sample t-test or 
two sample Wilcoxon rank-sum test, 
the functions \code{gsMMD2} and \code{gsMMD2.default} could be used. 

The rows of the input data matrix (or the data matrix derived from the object 
derived from the class \code{ExpressionSet}) 
are genes, while the columns are tissue samples. The tissue type of a tissue
sample is indicated by the argument \code{memSubjects}, which is
a $m\times 1$ vector, $m=m_{c}+m_{n}$, $m_{c}$ is the number of
tissue samples in the case group, and $m_{n}$ is the number of
tissue samples in the control group. \code{memSubjects[j]=1} indicates
the $j$-th tissue sample is from the case group. 
\code{memSubjects[j]=0} indicates the $j$-th tissue sample is from
the control group.

The output of gene selection functions include \code{dat}, \code{memGenes},
\code{memGenes2}, and \code{para}. 
\begin{itemize}

  \item \code{dat} is a data matrix with
the same dimensions as the input data matrix. If no data transformation
is performed, \code{dat} is the same as the input data matrix. Otherwise,
it will be the transformed input data matrix. 

  \item \code{memGenes} is a 
$G\times 1$ vector indicating the gene cluster membership, where $G$ is
the number of genes (i.e., the number of rows of the input data matrix).
\code{memGenes[g]=1} indicates the $g$-th gene is assigned to the cluster
of genes over-expressed in case group; 
\code{memGenes[g]=2} indicates the $g$-th gene is assigned to the cluster
of genes non-differentially-expressed; 
\code{memGenes[g]=3} indicates the $g$-th gene is assigned to the cluster
of genes under-expressed in case group. 

  \item \code{memGenes2} is a variant of \code{memGenes}. 
\code{memGenes2[g]=1} means the $g$-th gene is differentially expressed, 
while 
\code{memGenes2[g]=0} means the $g$-th gene is non-differentially expressed,
$g=1,\ldots, G$. 

  \item \code{para} is a $18\times 1$ vector of parameters 
$(\pi_1$, $\pi_2$, $\pi_3$, 
$\mu_{c1}$, $\sigma^2_{c1}$,
$\rho_{c1}$, $\mu_{n1}$, $\sigma^2_{n1}$, $\rho_{n1}$,
$\mu_2$, $\sigma^2_2$, $\rho_2$,
$\mu_{c3}$, $\sigma^2_{c3}$,
$\rho_{c3}$, $\mu_{n3}$, $\sigma^2_{n3}$, $\rho_{n3}$,
$)$ for the model
described in Equations (\ref{Model: mixture of 3-multivariate Normal})-
(\ref{Formula: Rt}), which can be estimated by
the EM algorithm.
\code{para[1]}, \code{para[2]}, and \code{para[3]} are the cluster proportions
for the 3 gene clusters (over-expressed, non-differentially expressed, and
under-expressed).
\code{para[4]}, \code{para[5]}, and \code{para[6]} are 
the marginal mean, variance, and correlation of gene expression levels 
of cluster 1 (over-expressed genes) for case subjects; 
\code{para[7]}, \code{para[8]}, and \code{para[9]} are 
the marginal mean, variance, and correlation of gene expression levels 
of cluster 1 (over-expressed genes) for control subjects; 
\code{para[10]}, \code{para[11]}, and \code{para[12]} 
are the marginal mean, 
variance, and correlation of gene 
expression levels of cluster 2 (non-differentially expressed genes); 
\code{para[13]}, \code{para[14]}, and \code{para[15]} are 
the marginal mean, variance, and correlation of gene expression levels 
of cluster 3 (under-expressed genes) for case subjects; 
\code{para[16]}, \code{para[17]}, and \code{para[18]} are 
the marginal mean, variance, and correlation of gene expression levels 
of cluster 3 (under-expressed genes) for control subjects. 

Note that genes in cluster 2 are non-differentially expressed across
case and control tissue samples. Hence there are only 3 parameters for
cluster 2.
\end{itemize}

For example, to obtain differentially expressed genes for the ALL data 
(Chiaretti et al., 2004),\nocite{ChiarettiEtAl:2004} we can run either
<<echo=TRUE, eval=TRUE,print=FALSE>>=
  library(GeneSelectMMD)
  library(ALL)
  data(ALL)
  eSet1 <- ALL[1:100, ALL$BT == "B3" | ALL$BT == "T2"]
  
  mem.str <- as.character(eSet1$BT)
  nSubjects <- length(mem.str)
  memSubjects <- rep(0,nSubjects)
  # B3 coded as 0, T2 coded as 1
  memSubjects[mem.str == "T2"] <- 1
  
  obj.gsMMD <- gsMMD(eSet1, memSubjects, transformFlag = TRUE, 
    transformMethod = "boxcox", scaleFlag = TRUE, quiet = TRUE)
  para <- obj.gsMMD$para
  print(round(para, 3))

@
or
<<echo=TRUE, eval=FALSE,print=FALSE>>=
  library(GeneSelectMMD)
  library(ALL)
  data(ALL)
  eSet1 <- ALL[1:100, ALL$BT == "B3" | ALL$BT == "T2"]
  mat <- exprs(eSet1)
  
  mem.str <- as.character(eSet1$BT)
  nSubjects <- length(mem.str)
  memSubjects <- rep(0,nSubjects)
  # B3 coded as 0, T2 coded as 1
  memSubjects[mem.str == "T2"] <- 1
 
  obj.gsMMD <- gsMMD.default(mat, memSubjects, iniGeneMethod = "Ttest",
          transformFlag = TRUE, transformMethod = "boxcox", scaleFlag = TRUE,
          quiet=TRUE)
  para <- obj.gsMMD$para
  print(round(para, 3))

@

If we would like to provide the initial 3-cluster partition via the
two sample Wilcoxon rank-sum test, then
we can run either
<<echo=TRUE, eval=FALSE,print=FALSE>>=
  library(GeneSelectMMD)
  library(ALL)
  data(ALL)
  eSet1 <- ALL[1:100, ALL$BT == "B3" | ALL$BT == "T2"]
  
  mem.str <- as.character(eSet1$BT)
  nSubjects <- length(mem.str)
  memSubjects <- rep(0,nSubjects)
  # B3 coded as 0, T2 coded as 1
  memSubjects[mem.str == "T2"] <- 1
  
  myWilcox <-
  function(x, memSubjects, alpha = 0.05)
  {
    xc <- x[memSubjects == 1]
    xn <- x[memSubjects == 0]
  
    m <- sum(memSubjects == 1)
    res <- wilcox.test(x = xc, y = xn, conf.level = 1 - alpha)
    res2 <- c(res$p.value, res$statistic - m * (m + 1) / 2)
    names(res2) <- c("p.value", "statistic")
  
    return(res2)
  }
  
  mat <- exprs(eSet1)
  tmp <- t(apply(mat, 1, myWilcox, memSubjects = memSubjects))
  colnames(tmp) <- c("p.value", "statistic")
  memIni <- rep(2, nrow(mat))
  memIni[tmp[, 1] < 0.05 & tmp[, 2] > 0] <- 1
  memIni[tmp[, 1] < 0.05 & tmp[, 2] < 0] <- 3
  
  print(table(memIni))

  obj.gsMMD <- gsMMD2(eSet1, memSubjects, memIni, transformFlag = TRUE, 
       transformMethod = "boxcox", scaleFlag = TRUE, quiet = TRUE)
  para <- obj.gsMMD$para
  print(round(para, 3))

@
or
<<echo=TRUE, eval=FALSE,print=FALSE>>=
  library(GeneSelectMMD)
  library(ALL)
  data(ALL)
  eSet1 <- ALL[1:100, ALL$BT == "B3" | ALL$BT == "T2"]
  mat <- exprs(eSet1)
  
  mem.str <- as.character(eSet1$BT)
  nSubjects <- length(mem.str)
  memSubjects <- rep(0,nSubjects)
  # B3 coded as 0, T2 coded as 1
  memSubjects[mem.str == "T2"] <- 1
 
  myWilcox <-
  function(x, memSubjects, alpha = 0.05)
  {
    xc <- x[memSubjects == 1]
    xn <- x[memSubjects == 0]
  
    m <- sum(memSubjects == 1)
    res <- wilcox.test(x = xc, y = xn, conf.level = 1 - alpha)
    res2 <- c(res$p.value, res$statistic - m * (m + 1) / 2)
    names(res2) <- c("p.value", "statistic")
  
    return(res2)
  }
  
  tmp <- t(apply(mat, 1, myWilcox, memSubjects = memSubjects))
  colnames(tmp) <- c("p.value", "statistic")
  memIni <- rep(2, nrow(mat))
  memIni[tmp[, 1] < 0.05 & tmp[, 2] > 0] <- 1
  memIni[tmp[, 1] < 0.05 & tmp[, 2] < 0] <- 3
  
  print(table(memIni))
  
  obj.gsMMD <- gsMMD2.default(mat, memSubjects, memIni = memIni,
          transformFlag = TRUE, transformMethod = "boxcox", scaleFlag = TRUE,
          quiet=TRUE)
  para <- obj.gsMMD$para
  print(round(para, 3))


@

Actually, the two sample Wilcoxon rank-sum test is implemented in the 
\Rpackage{GeneSelectMMD}. The above code is used as an illustration
on how to use the functions \code{gsMMD2} and \code{gsMMD2.default}.

Note that the speed of the four functions can be slow for large data sets,
however it has been improved since version 2.0.7 by using Fotran code. 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\section{Error rates estimated for real microarray data sets}

For real microarray data sets, the classical gene selection methods, such as two-sample t-test,
two-sample Wilcoxon rank-sum test, do not directly provide the estimates of
error rates such as false discovery rate (denoted as FDR;
it is the percentage of nondifferentially expressed genes among selected
genes),
false non-discovery rate (denoted as FNDR; it is the percentage of differentially expressed genes among unselected genes), 
false positive rate (denoted as FPR; it is the percentage of selected 
genes among nondifferentially expressed genes), and
false negative rate (denoted as FNR; it is the percentage of
un-selected genes among differentially expressed genes),
since the true gene cluster membership is unknown.


However, model-based gene selection methods (e.g.,
eLNN and eGG proposed by Lo and Gottardo (2007),\nocite{LoGottardo:2007} 
and the method proposed by Qiu et al.'s (2008)) could easily estimate
these error rates, since these error rates are functions of some 
probabilities which are in turn the functions of model parameters.

The function \code{errRates} is used to estimate FDR, FNDR, FPR, and FNR
based on the objects returned by the four gene selection functions
mentioned in the previous section. This function returns a $4\times 1$ vector
with elements FDR, FNDR, FPR, and FNR.

For example,
<<echo=TRUE, eval=TRUE,print=FALSE>>=
  print(round(errRates(obj.gsMMD), 3))

@

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\section{Visualizing the fit of the model to a real microarray data set}
In general, it is difficult to visualize the fit of the model to
a real microarray data set since the data are in high-dimensional space.
However, it is possible to do so for the special case where the tissue
samples within a tissue type are marginally independent. In this case,
we can pool all gene expression levels together since they are all
independent, and regard them as one-dimensional data. Next, we can
impose the density estimate onto the histogram of this pooled data.

The function \code{plotHistDensity} is used for such purpose.
The following R code illustrates the usage of \code{plotHistDensity}:
<<echo=TRUE, eval=TRUE,print=FALSE,fig=TRUE>>=
  plotHistDensity(obj.gsMMD, plotFlag = "case", 
      mytitle = "Histogram of gene expression levels for T2\nimposed with estimated density (case)", 
      plotComponent = TRUE, 
      x.legend = c(0.8, 3), 
      y.legend = c(0.3, 0.4), 
      numPoints = 500,
      cex.main = 1,
      cex = 1)

@
\begin{figure}[h]
\caption{Plot produced using plotHistDensity}
\label{Figure:plotHistDensity}
\end{figure}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Efficiency improvements}

Version 2.0.7 includes efficiency improvements to the EM optimization process.  Specifically, Fortran 77 code for version 3.0 of the L-BFGS-B optimization algorithm from Morales and Nocedal (2011)\nocite{MoralesNocedal:2011} has been included.  The objective functions for the EM optimization have been translated to Fortran 77 in order to facilitate the use of this optimization code.  Additional operations within the EM optimization iterative loop have also been written in Fortran 77 to improve the speed, where the calculation of the weigths has been addressed.  Lastly, although the t-test is an optional operation, and therefore, improvements in its speed may not always be of consequence, the two-sample t-test calculations have increased efficiency in version 2.0.7.

Analysis involving simulated microarray data for 32,000 genes and a total of 2,000 samples is accomplished in under a minute on a notebook computer (Figure ~\ref{Figure:efficiency}), and therefore, the enhanced performance provided by GeneSelectMMD can be attained on larger datasets using typical hardware.

\begin{figure}[ht]
\begin{center}
\includegraphics[width=0.6\textwidth,angle=0]{GS207runTimesSim1k.pdf}
\caption{Plot of running time vs number of genes using an Asus notebook with Intel\textregistered{ }Core\texttrademark{ }i5-2450M CPU and 8GB memory}
\label{Figure:efficiency}
\end{center}
\end{figure}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Discussion}

The speeds of the four gene selection functions described in Section
\ref{Section: Gene Selection Functions} can be slow, however the speed has been
improved by embedding Fortran code in the R code for version 2.0.7 
of the \Rpackage{GeneSelectMMD}.


\bibliographystyle{plain}
\bibliography{GeneSelectMMD}

\end{document}

\name{gsMMD2.default}
\alias{gsMMD2.default}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{Gene selection based on a mixture of marginal distributions}
\description{
Gene selection based on the marginal distributions of gene profiles that characterized by a mixture of three-component multivariate distributions.
Input is a data matrix. The user needs to provide initial gene cluster membership.
}
\usage{
gsMMD2.default(X, 
               memSubjects, 
               memIni,
               maxFlag = TRUE, 
               thrshPostProb = 0.5, 
               geneNames = NULL, 
               alpha = 0.05, 
               transformFlag = FALSE, 
               transformMethod = "boxcox", 
               scaleFlag = TRUE, 
               criterion = c("cor", "skewness", "kurtosis"), 
               minL = -10, 
               maxL = 10, 
               stepL = 0.1, 
               eps = 0.001, 
               ITMAX = 100, 
               plotFlag = FALSE,
               quiet=TRUE)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{X}{a data matrix. The rows of the matrix are genes. The columns of the matrix are subjects.}
  \item{memSubjects}{a vector of membership of subjects. \code{memSubjects[i]=1} means the \eqn{i}-th subject belongs to diseased group, \eqn{0} otherwise. 
  }
  \item{memIni}{a vector of user-provided gene cluster membership.}
  \item{maxFlag}{logical. Indicate how to assign gene class membership. \code{maxFlag}=TRUE means that a gene will be assigned
to a class in which the posterior probability of the gene belongs to this class is maximum. \code{maxFlag}=FALSE means that a gene will be assigned to class 1 if the posterior probability of the gene belongs to class 1 is
greater than \code{thrshPostProb}. Similarly, a gene will be assigned to class 1 if the posterior probability of the gene belongs to class 1 is
greater than \code{thrshPostProb}. If  the posterior probability is less than \code{thrshPostProb}, the gene
will be assigned to class 2 (non-differentially expressed gene group).}
  \item{thrshPostProb}{threshold for posterior probabilities. For example, if the posterior probability that a gene belongs to cluster 1 given its gene expression levels is larger than \code{thrshPostProb}, then this gene will be assigned to cluster 1.}
  \item{geneNames}{an optional character vector of gene names}
  \item{alpha}{significant level which is equal to \code{1-conf.level}, 
\code{conf.level} is the argument for the function \code{t.test}.
  }
  \item{transformFlag}{logical. Indicate if data transformation is needed}
  \item{transformMethod}{method for transforming data. Available methods include "boxcox", "log2", "log10", "log", "none".}
  \item{scaleFlag}{logical. Indicate if gene profiles are to be scaled. If \code{transformFlag=TRUE} and \code{scaleFlag=TRUE}, then scaling is performed after transformation.  
To avoid linear dependence of tissue samples after scaling
    gene profiles, we delete one tissue sample after scaling (c.f. details).
}
  \item{criterion}{if \code{transformFlag=TRUE}, \code{criterion} indicates what criterion to determine if data looks like normal.  \dQuote{cor} means using Pearson's correlation. The idea is that the observed quantiles after transformation should be close to theoretical normal quantiles. So we can use Pearson's correlation to check if the scatter plot of theoretical normal quantiles versus observed quantiles is a straightline.  \dQuote{skewness} means using skewness measure to check if the distribution of the transformed data are close to normal distribution; \dQuote{kurtosis} means using kurtosis measure to check normality.}
  \item{minL}{lower limit for the \code{lambda} parameter used in Box-Cox transformation}
  \item{maxL}{upper limit for the \code{lambda} parameter used in Box-Cox transformation}
  \item{stepL}{tolerance when searching the optimal \code{lambda} parameter used in Box-Cox transformation}
  \item{eps}{a small positive value. If the absolute value of a value is smaller than \code{eps}, this value is regarded as zero.  }
  \item{ITMAX}{maximum iteration allowed for iterations in the EM algorithm}
  \item{plotFlag}{logical. Indicate if the Box-Cox normality plot should be output.}
  \item{quiet}{logical. Indicate if intermediate results should be printed out.}
}
\details{
We assume that the distribution of gene expression profiles is 
a mixture of 3-component multivariate normal distributions 
\eqn{\sum_{k=1}^{3} \pi_k f_k(x|\theta)}. Each component distribution \eqn{f_k} 
corresponds to a gene cluster. The 3 components correspond to 3 gene clusters:
(1) up-regulated gene cluster, (2) non-differentially expressed gene cluster, 
and (3) down-regulated gene cluster. 
The model parameter vector is \eqn{\theta=(\pi_1}, \eqn{\pi_2}, 
\eqn{\pi_3}, \eqn{\mu_{c1}}, 
\eqn{\sigma^2_{c1}}, \eqn{\rho_{c1}}, \eqn{\mu_{n1}}, \eqn{\sigma^2_{n1}}, 
\eqn{\rho_{n1}},
\eqn{\mu_2}, \eqn{\sigma^2_2}, \eqn{\rho_2}, 
\eqn{\mu_{c3}}, 
\eqn{\sigma^2_{c3}}, \eqn{\rho_{c3}}, \eqn{\mu_{n3}}, \eqn{\sigma^2_{n3}}, 
\eqn{\rho_{n3}}.
where \eqn{\pi_1}, \eqn{\pi_2}, and \eqn{\pi_3} are the mixing proportions; 
\eqn{\mu_{c1}}, \eqn{\sigma^2_{c1}}, and \eqn{\rho_{c1}} are 
the marginal mean, variance, and correlation of gene expression levels 
of cluster 1 (up-regulated genes) for diseased subjects; 
\eqn{\mu_{n1}}, \eqn{\sigma^2_{n1}}, and \eqn{\rho_{n1}} are 
the marginal mean, variance, and correlation of gene expression levels 
of cluster 1 (up-regulated genes) for non-diseased subjects; 
\eqn{\mu_2}, \eqn{\sigma^2_2}, and \eqn{\rho_2} are the marginal mean, 
variance, and correlation of gene 
expression levels of cluster 2 (non-differentially expressed genes); 
\eqn{\mu_{c3}}, \eqn{\sigma^2_{c3}}, and \eqn{\rho_{c3}} are 
the marginal mean, variance, and correlation of gene expression levels 
of cluster 3 (up-regulated genes) for diseased subjects; 
\eqn{\mu_{n3}}, \eqn{\sigma^2_{n3}}, and \eqn{\rho_{n3}} are 
the marginal mean, variance, and correlation of gene expression levels 
of cluster 3 (up-regulated genes) for non-diseased subjects. 

Note that genes in cluster 2 are non-differentially expressed across
abnormal and normal tissue samples. Hence there are only 3 parameters for
cluster 2.

To make sure the identifiability, we set the following contraints: 
\eqn{\mu_{c1}>\mu_{n1}} and \eqn{\mu_{c3}<\mu_{n3}}.

To make sure the marginal covariance matrices are poisitive definite,
we set the following contraints: 
\eqn{-1/(n_c-1)<\rho_{c1}<1},
\eqn{-1/(n_n-1)<\rho_{n1}<1},
\eqn{-1/(n-1)<\rho_{2}<1},
\eqn{-1/(n_c-1)<\rho_{c3}<1},
\eqn{-1/(n_n-1)<\rho_{n3}<1}.

We also has the following constraints for the mixing proportion:
\eqn{\pi_3=1-\pi_1-\pi_2}, \eqn{\pi_k>0}, \eqn{k=1,2,3}.

We apply the EM algorithm to estimate the model parameters. 
We regard the cluster membership of genes as missing values.

To facilitate the estimation of the parameters,
we reparametrize the parameter vector as
\eqn{\theta^*=(\pi_1}, \eqn{\pi_2}, 
\eqn{\mu_{c1}}, 
\eqn{\sigma^2_{c1}}, \eqn{r_{c1}}, \eqn{\delta_{n1}}, \eqn{\sigma^2_{n1}}, 
\eqn{r_{n1}},
\eqn{\mu_2}, \eqn{\sigma^2_2}, \eqn{r_2}, 
\eqn{\mu_{c3}}, 
\eqn{\sigma^2_{c3}}, \eqn{r_{c3}}, \eqn{\delta_{n3}}, \eqn{\sigma^2_{n3}}, 
\eqn{r_{n3})},
where
\eqn{\mu_{n1}=\mu_{c1}-\exp(\delta_{n1})},
\eqn{\mu_{n3}=\mu_{c3}+\exp(\delta_{n3})},
\eqn{\rho_{c1}=(\exp(r_{c1})-1/(n_c-1))/(1+\exp(r_{c1}))},
\eqn{\rho_{n1}=(\exp(r_{n1})-1/(n_n-1))/(1+\exp(r_{n1}))},
\eqn{\rho_{2}=(\exp(r_{2})-1/(n-1))/(1+\exp(r_{2}))},
\eqn{\rho_{c3}=(\exp(r_{c3})-1/(n_c-1))/(1+\exp(r_{c3}))},
\eqn{\rho_{n3}=(\exp(r_{n3})-1/(n_n-1))/(1+\exp(r_{n3}))}.

Given a gene, the expression levels of the gene are assumed independent. However, after scaling, the scaled expression levels of the gene are no longer independent and the rank \eqn{r^*=r-1} of the covariance matrix for the scaled gene profile will be one less than the rank \eqn{r} for the un-scaled gene profile Hence the covariance matrix of the
gene profile will no longer be positive-definite. 
To avoid this problem,
we delete a tissue sample after scaling since its information has been incorrporated by other scaled tissue samples. We arbitrarily select the tissue sample, which has the biggest label number, from the tissue sample group that has larger size than the other tissue sample group. For example, if there are 6 cancer tissue samples and 10 normal tissue samples, we delete the 10-th normal tissue sample after scaling.

}
\value{
A list contains 13 elements.
  \item{dat}{the (transformed) microarray data matrix. If tranformation
performed, then \code{dat} will be different from the input 
microarray data matrix.}
  \item{memSubjects}{the same as the input \code{memSubjects}.}
  \item{memGenes}{a vector of cluster membership of genes. \eqn{1} means up-regulated gene; \eqn{2} means non-differentially expressed gene; 
\eqn{3} means down-regulated gene.}
  \item{memGenes2}{an variant of the vector of cluster membership of genes. 
\eqn{1} means differentially expressed gene; \eqn{0} means non-differentially expressed gene.}
  \item{para}{parameter estimates (c.f. details).}
  \item{llkh}{value of the loglikelihood function.}
  \item{wiMat}{posterior probability that a gene belongs to a cluster given the expression levels of this gene. Column i is for cluster i.}
  \item{memIni}{the initial cluster membership of genes.}
  \item{paraIni}{the parameter estimates based on initial gene cluster membership.}
  \item{llkhIni}{the value of loglikelihood function.}
 \item{lambda}{the parameter used to do Box-Cox transformation}
  \item{paraRP}{parameter estimates for reparametrized parameter vector (c.f. details).}
  \item{paraIniRP}{the parameter estimates for reparametrized parameter vector based on initial gene cluster membership.}
}
\references{ Qiu, W.-L., He, W., Wang, X.-G. and Lazarus, R. (2008). 
A Marginal Mixture Model for Selecting Differentially Expressed Genes across Two Types of Tissue Samples. \emph{The International Journal of Biostatistics. 4(1):Article 20.} \url{http://www.bepress.com/ijb/vol4/iss1/20}
}
\author{ 
Jarrett Morrow \email{remdj@channing.harvard.edu},
Weiliang Qiu \email{Weiliang.Qiu@gmail.com},
Wenqing He \email{whe@stats.uwo.ca},
Xiaogang Wang \email{stevenw@mathstat.yorku.ca},
Ross Lazarus \email{ross.lazarus@channing.harvard.edu}
}
\seealso{
\code{\link{gsMMD}},
\code{\link{gsMMD.default}},
\code{\link{gsMMD2}}
}
\note{
The speed of the program can be slow for large data sets, however it has been improved using Fortran code.
}
\examples{

  \dontrun{
    library(ALL)
    data(ALL)
    eSet1 <- ALL[1:100, ALL$BT == "B3" | ALL$BT == "T2"]
    mat <- exprs(eSet1)
    
    mem.str <- as.character(eSet1$BT)
    nSubjects <- length(mem.str)
    memSubjects <- rep(0, nSubjects)
    # B3 coded as 0, T2 coded as 1
    memSubjects[mem.str == "T2"] <- 1
   
    myWilcox <-
    function(x, memSubjects, alpha = 0.05)
    {
      xc <- x[memSubjects == 1]
      xn <- x[memSubjects == 0]
    
      m <- sum(memSubjects == 1)
      res <- wilcox.test(x = xc, y = xn, conf.level = 1 - alpha)
      res2 <- c(res$p.value, res$statistic - m * (m + 1) / 2)
      names(res2) <- c("p.value", "statistic")
    
      return(res2)
    }
    
    tmp <- t(apply(mat, 1, myWilcox, memSubjects = memSubjects))
    colnames(tmp) <- c("p.value", "statistic")
    memIni <- rep(2, nrow(mat))
    memIni[tmp[, 1] < 0.05 & tmp[, 2] > 0] <- 1
    memIni[tmp[, 1] < 0.05 & tmp[,2] < 0] <- 3
    
    cat("initial gene cluster size>>\n"); print(table(memIni)); cat("\n");
    
    obj.gsMMD <- gsMMD2.default(mat, memSubjects, memIni = memIni,
            transformFlag = TRUE, transformMethod = "boxcox", scaleFlag = TRUE)
    round(obj.gsMMD$para, 3)
  }
  
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{ classif }

\name{plotHistDensity}
\alias{plotHistDensity}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{ Plot of histogram and density estimate of the pooled gene expression levels. }
\description{
Plot of histogram of pooled gene expression levels, composited with density estimate based on the mixture of marginal distributions. The density estimate is based on the assumption that the marginal correlations between subjects are zero.
}
\usage{
plotHistDensity(obj.gsMMD,
                plotFlag = "case",
                plotComponent = FALSE,
                myxlab = "expression level",
                myylab = "density",
                mytitle = "Histogram with estimated density (case)",
                x.legend = NULL,
                y.legend = NULL,
                numPoints = 500,
                mycol = 1:4, 
                mylty = 1:4, 
                mylwd = rep(3,4), 
                cex.main = 2, 
                cex.lab = 1.5, 
                cex.axis = 1.5, 
                cex = 2,
                bty = "n")

}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{obj.gsMMD}{an object returned by \code{gsMMD},
\code{gsMMD.default}, \code{gsMMD2},
or \code{gsMMD2.default}
}
  \item{plotFlag}{logical. Indicate the plot will based on which type of subjects.}
  \item{plotComponent}{logical. Indicate if components of the mixture of marginal distribution will be plotted.}
  \item{myxlab}{label for x-axis}
  \item{myylab}{label for y-axis}
  \item{mytitle}{title of the plot}
  \item{x.legend}{the x-corrdiates of the legend}
  \item{y.legend}{the y-corrdiates of the legend}
  \item{numPoints}{logical. Indicate how many genes will be plots.}
  \item{mycol}{color for the density estimates (overall and components)}
  \item{mylty}{line styles for the density estimates (overall and components)}
  \item{mylwd}{line width for the density estimates (overall and components)}
  \item{cex.main}{font for main title}
  \item{cex.lab}{font for x- and y-axis labels}
  \item{cex.axis}{font for x- and y-axis}
  \item{cex}{font for texts}
  \item{bty}{the type of box to be drawn around the legend.  The allowed
          values are '"o"' and '"n"' (the default).}
}
\details{
For a given type of subjects, we pool their expression levels together if
the marginal correlations among subjects are zero. We then draw a histogram
of the pooled expression levels. Next, we composite density estimates of gene
expression levels for the overal distribution and the 3 component distributions.
}
\value{
A list containing coordinates of the density estimates:
  \item{x }{sorted pooled gene expression levels for cases or controls.}
  \item{x2 }{a subset of \code{x} specified by the sequence:
\code{seq(from = 1,to = len.x, by = delta)}, where \code{len.x} is the
length of the vector \code{x}, and \code{delta = floor(len.x/numpoints).}}
  \item{y }{density estimate corresponding to \code{x2}}
  \item{y1 }{weighted density estimate for gene cluster 1}
  \item{y2 }{weighted density estimate for gene cluster 2}
  \item{y3 }{weighted density estimate for gene cluster 3}
}
\references{ Qiu, W.-L., He, W., Wang, X.-G. and Lazarus, R. (2008). 
A Marginal Mixture Model for Selecting Differentially Expressed Genes across Two Types of Tissue Samples. \emph{The International Journal of Biostatistics. 4(1):Article 20.} \url{http://www.bepress.com/ijb/vol4/iss1/20}
}
\author{ 
Jarrett Morrow \email{remdj@channing.harvard.edu},
Weiliang Qiu \email{Weiliang.Qiu@gmail.com},
Wenqing He \email{whe@stats.uwo.ca},
Xiaogang Wang \email{stevenw@mathstat.yorku.ca},
Ross Lazarus \email{ross.lazarus@channing.harvard.edu}
}
\note{ The density estimate is obtained based on the
assumption that the marginal correlation among
subjects is zero. If the estimated marginal correlation obtained by \code{gsMMD} is far from zero, then do not use this plot function. 
}
\examples{
  \dontrun{
library(ALL)
data(ALL)
eSet1 <- ALL[1:100, ALL$BT  ==  "B3" | ALL$BT  ==  "T2"]

mem.str <- as.character(eSet1$BT)
nSubjects <- length(mem.str)
memSubjects <- rep(0,nSubjects)
# B3 coded as 0 (control), T2 coded as 1 (case)
memSubjects[mem.str  ==  "T2"] <- 1

obj.gsMMD <- gsMMD(eSet1, memSubjects, transformFlag  =  TRUE, 
  transformMethod  =  "boxcox", scaleFlag  =  TRUE, quiet  =  FALSE)

plotHistDensity(obj.gsMMD, plotFlag  =  "case", 
    mytitle  =  "Histogram of for T2 imposed with estimated density (case)", 
    plotComponent  =  TRUE, 
    x.legend  =  c(0.8, 3), 
    y.legend  =  c(0.3, 0.4), 
    numPoints  =  500)
  }

}
\keyword{classif }% at least one, from doc/KEYWORDS

\name{obtainResi}
\alias{obtainResi}
\title{
Replace expression levels by the residuals of
regression analysis to remove the confounding effects.
}
\description{
Replace expression levels by the residuals of regression analysis
in which predictor of interest is not in the regression model.
The purpose of this function is to remove potential confounding factors.
}
\usage{
obtainResi(es, fmla)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{es}{
An \code{ExpressionSet} object.
}
  \item{fmla}{
A formula object that specifies the covariates of the linear
regression model. The variable of interest should not be 
included. No response variable should be specified in \code{fmla} since
the response variable is always the expression level. See 
function \code{\link{lmFit}} of R Bioconductor package \code{limma}.
}
}
\details{
To remove confounding effects, we can replace the expression level by
the residuals
of  a linear regression model with
response variable the expression level and covariates the potential
confounders. The functions \code{lmFit}  and \code{eBayes} will
be used to obtain regression coefficients.
}
\value{
An ExpressionSet object with expression levels replaced by
residuals of linear regression analysis.
}
\author{
Jarrett Morrow \email{remdj@channing.harvard.edu},
Weiliang Qiu \email{Weiliang.Qiu@gmail.com},
Wenqing He \email{whe@stats.uwo.ca},
Xiaogang Wang \email{stevenw@mathstat.yorku.ca},
Ross Lazarus \email{ross.lazarus@channing.harvard.edu}
}
\note{
The number of arrays of the returned ExpressionSet object
might be smaller than that of the original ExpressionSet object,
due to missing values in covariates.
}
%\examples{
%}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{ methods }

\name{errRates}
\alias{errRates}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{ Calculating FDR, FNDR, FPR, and FNR for a real microarray data set}
\description{
Calculating FDR, FNDR, FPR, and FNR for a real microarray data set based on the mixture of marginal distributions.
}
\usage{
errRates(obj.gsMMD)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{obj.gsMMD}{an object returned by \code{gsMMD},
\code{gsMMD.default}, \code{gsMMD2},
or \code{gsMMD2.default}
}
}
\details{
We first fit the real microarray data set by the mixture of 
marginal distributions. Then we calculate the error rates
based on the posterior distributions of a gene belonging to
a gene cluster given its gene profiles. Please refer to Formula (7)
on the page 6 of the paper listed in the Reference section.
}
\value{
A vector of 4 elements:
  \item{FDR }{the percentage of nondifferentially expressed genes among selected genes.}
  \item{FNDR }{the percentage of differentially expressed genes among unselected genes.}
  \item{FPR }{the percentage of selected genes among nondifferentially expressed genes}
  \item{FNR }{the percentage of un-selected genes among differentially expressed genes}
}
\references{ Qiu, W.-L., He, W., Wang, X.-G. and Lazarus, R. (2008). 
A Marginal Mixture Model for Selecting Differentially Expressed Genes across Two Types of Tissue Samples. \emph{The International Journal of Biostatistics. 4(1):Article 20.} \url{http://www.bepress.com/ijb/vol4/iss1/20}
}
\author{ 
Jarrett Morrow \email{remdj@channing.harvard.edu},
Weiliang Qiu \email{Weiliang.Qiu@gmail.com},
Wenqing He \email{whe@stats.uwo.ca},
Xiaogang Wang \email{stevenw@mathstat.yorku.ca},
Ross Lazarus \email{ross.lazarus@channing.harvard.edu}
}
\examples{
  \dontrun{
    library(ALL)
    data(ALL)
    eSet1 <- ALL[1:100, ALL$BT == "B3" | ALL$BT == "T2"]
    
    mem.str <- as.character(eSet1$BT)
    nSubjects <- length(mem.str)
    memSubjects <- rep(0,nSubjects)
    # B3 coded as 0, T2 coded as 1
    memSubjects[mem.str == "T2"] <- 1
    
    obj.gsMMD <- gsMMD(eSet1, memSubjects, transformFlag = TRUE, 
      transformMethod = "boxcox", scaleFlag = TRUE, quiet = FALSE)
    round(errRates(obj.gsMMD), 3)
  }
}
\keyword{classif }% at least one, from doc/KEYWORDS
